Package: trtf
Title: Transformation Trees and Forests
Version: 0.4-3
Date: 2025-01-29
Authors@R: person("Torsten", "Hothorn", role = c("aut", "cre"),
                  email = "Torsten.Hothorn@R-project.org", comment = c(ORCID = "0000-0001-8301-0471"))
Description: Recursive partytioning of transformation models with
  corresponding random forest for conditional transformation models 
  as described in 'Transformation Forests' (Hothorn and Zeileis, 2021, <doi:10.1080/10618600.2021.1872581>) 
  and 'Top-Down Transformation Choice' (Hothorn, 2018, <DOI:10.1177/1471082X17748081>).
Depends: mlt (>= 1.4-1), partykit (>= 1.2-1), tram
Imports: Formula, sandwich, grid, stats, variables, libcoin, utils,
        grDevices
Suggests: survival, TH.data, coin, ATR, Matrix, colorspace, inum,
        lattice, latticeExtra, memisc, multcomp, quantregForest,
        AppliedPredictiveModeling, mlbench
URL: http://ctm.R-forge.R-project.org
License: GPL-2
NeedsCompilation: no
Packaged: 2025-01-29 15:01:58 UTC; hothorn
Author: Torsten Hothorn [aut, cre] (<https://orcid.org/0000-0001-8301-0471>)
Maintainer: Torsten Hothorn <Torsten.Hothorn@R-project.org>
Repository: CRAN
Date/Publication: 2025-01-29 15:30:05 UTC
Built: R 4.6.0; ; 2026-04-28 12:37:38 UTC; unix

import("mlt")
import("tram")
import("partykit")
import("Formula")
importFrom("sandwich", "estfun")
importFrom("variables", "mkgrid", "support")
importFrom("libcoin", "ctabs")

importFrom("utils",
  "setTxtProgressBar",
  "txtProgressBar"
)

import("grid")
importFrom("graphics", "plot")
importFrom("grDevices", "gray.colors")
importFrom("stats", "coef", "logLik", "model.frame", "model.weights",
           "na.pass", "predict", "update", "simulate", "variable.names", "resid")

export(trafotree, traforest)

S3method("coef", "trafotree")
S3method("logLik", "trafotree")
S3method("logLik", "traforest")
S3method("predict", "trafotree")
S3method("plot", "trafotree")
S3method("predict", "traforest")
S3method("simulate", "traforest")
S3method("simulate", "trafotree")
S3method("gettree", "traforest")

################################################################################

# Model-based Random Forests for Ordinal Regression

# Muriel Buri and Torsten Hothorn

################################################################################

This folder contains simplified and commented R-scripts used for the empirical
evaluation and applied example of the manuscript "Model-based Random Forests for
Ordinal Regression".

Please refer to the manuscript for details on the functions, models and simulation
procedures.

In each subfolder you will find an instruction on how to run the experiments.

- The subfolder "empeval" is for comparison of prognostic models on the
  simulated data.

- The subfolder "ALS" is for the re-analysis of the "Respiratory" sub-item in
  functioncal rating scales of patients suffering from Amyotrophic Lateral
  Sclerosis (ALS).

This is the folder with R-code implementation of the experiments described in
Korepanova N., Seibold H., Steffen V., Hothorn T. Survival Forests under Test:
Impact of the Proportional Hazards Assumption on Prognostic and Predictive 
Forests for ALS Survival (2019). 

In each subfolder you will find an instruction on how to run the experiments.
The subfolder "prognostic" is for comparison of prognostic models on the
simulated data.
The subfolder "predictive" is for comparison of predictive models on the
simulated data.
The subfolder "ALS" is for experiments in the samples generated from
ALS dataset.

The code in "ALS" subfolder source the code from "prognostic" and
"predictive" subfolders. Therefore, we recommend not to change the
structures of the current folder and its subfolders.

################################################################################

# Model-based Random Forests for Ordinal Regression

# Muriel Buri and Torsten Hothorn

# SUBFOLDER: "ALS"

################################################################################

This folder contains simplified and commented R-scripts used for the re-analysis
of the Respiratory sub-item of the ALS Functional Rating Scale Revised (ALSFRS-R).

Please refer to the manuscript for details on the functions, models and
simulation procedures.

~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Reproducing the re-analysis:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
- As written in the manuscript, the patient-level data is available to registered
  users from https://nctu.partners.org/ProACT.
  The code for data preprocessing is available in the \pkg{TH.data} add-on
  R package.

- The R-script '01_ALS_sample.R' generates the 100 random splits of the data set
  and saves them as 'ALS_samples.rda' to the same directory.

- To compare the seven methods that predict the “Respiratory” ALSFRS-R score at
  half a year after diagnosis based on the 100 random splits are estimated while
  running the file '02_run_1_100.R'. For this, the file 'ALS_analysis.R' is ran
  within a for-loop using the functions of the 'ALS_competitors.R' file.
  The out-of-sample log-likelihoods and the quadratic weighted Cohen's Kappa are
  saved to the file 'ALS_results_combined.rda' in the same directory.

- The R-script '03_summary_plots.R' allows you to plot the results according to
  Figure 2 and 3 of the manuscript.

################################################################################

# Model-based Random Forests for Ordinal Regression

# Muriel Buri and Torsten Hothorn

# SUBFOLDER: "empeval"

################################################################################

This folder contains simplified and commented R-scripts used for the
empirical evaluation of the manuscript.

Please refer to the manuscript for details on the functions, models and
simulation procedures.

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Reproducing the empirical evaluation results:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- To generate the data, run '01_data_simulation.R'. The resulting simulated data
  set called 'simulated_data_nsim100.rda' will be saved in the (automatically
  generated) sub-directory called 'rda'.
  (Note that the whole simulation study with the original experimental conditions
  is quite time-consuming. You may want to run it on a multi-core server.)

- All forest-based prognostic methods with the corresponding predict functions
  as well as the log-likelihood and Kullback-Leibler divergence calculations are
  realized in 'competitors.R'.

- To run the empirical evaluation on the simulated data sets, run '02_empeval.R'.
  At the end of this, the results will again be written and saved to the
  sub-directory 'rda'. The resulting files are called:
    o 'results_empeval_nsim100_ntree250.rda', and
    o 'results_empeval_nsim100_ntree2000.rda'.

- The R-script '03_summary_plots.R' allows you to plot the results according to
  Figure 1, 4, 5, 6, 7, 8, 9, 10, 11, and 12 of the manuscript.

This is an instruction on running experiments on simulated data with prognostic effect.

To generate the data run data_simulation.R. As a result, in the same directory, where
data_simulation.R is located, a new file called simulated_data.rda will be created.
This file will contain the generated data. Approximately 122 Mb of disk space is required.

All forest-based prognostic methods with the corresponding predict functions and
modified weighting scheme are realized in competitors.R.

To learn all prognostic models on the generated data and estimate their log-likelihood
run empeval.R. Runnig this file a folder results_tf will be created, where for each
generated learning-validation pair of datasets a transformation
survival forest model with general splitting score will be saved in the corresponding
.rda file. The log-likelihood estimates will be saved to results_empeval.rda in the
same directory where empeval.R is located. Approximately 80 Kb of disk space is required.

This is an instruction on running experiments on simulated data with predictive effect.

To generate the data run tr_data_simulation.R. As a result, in the same directory, where
tr_data_simulation.R is located, a new file called simulated_data.rda will be created.
This file will contain the generated data. Approximately 115 Mb of disk space is required.

All predictive methods with the corresponding predict functions and
modified weighting scheme are realized in tr_competitors.R.

To learn all predictive models on the generated data and estimate their log-likelihood
run tr_empeval.R. Runnig this file a folder results_tf will be created, where for each
generated learning-validation pair of datasets a transformation
survival forest model with general splitting score will be saved in the corresponding
.rda file. The log-likelihood estimates will be saved to tr_results_empeval.rda in the
same directory where tr_empeval.R is located. Approximately 54 Kb of disk space is required.

library("trtf")
library("randomForest")
library("parallel")
RNGkind("L'Ecuyer-CMRG")
set.seed(29)
library("sandwich")

optim <- mltoptim(auglag = list(itmax = 50,
                                maxtry = 3, kkt2.check = FALSE), 
                  trace = FALSE)[1]

trafotree <- function(...) {
    ret <- try(trtf::trafotree(..., mltargs = list(optim = optim)))
    if (inherits(ret, "try-error")) return(NULL)
    return(ret)
}

traforest <- function(...) {
    ret <- try(trtf::traforest(..., mltargs = list(optim = optim)))
    if (inherits(ret, "try-error")) return(NULL)
    return(ret)
}


Ctree <- function(tt, root_trafo_only = FALSE, ...) {

    if (root_trafo_only) {
        ### NOTE: Setting up ytrafo that way ignores weights and thus we can't 
        ### use it for cforest !!!
        sc <- estfun(tt$trafo(model = TRUE)$object)
        ytrafo <- list(y = function(...) sc)
    } else {
        ytrafo <- NULL
    }
    ct <- ctree(..., ytrafo = ytrafo)
    ct$model <- tt$model
    ct$mltobj <- tt$trafo(model = TRUE, estfun = FALSE)
    ct$trafo <- tt$trafo
    ct$mltargs <- tt$mltargs

    nd <- predict(ct, type = "node")
    ct$models <- tapply(1:length(nd), factor(nd), function(i) 
        ct$trafo(i, estfun = FALSE)) ### note: trafo is (potentially) weighted
    ct$coef <- do.call("rbind", lapply(ct$models, function(x) x$coef))
    ct$logLik <- sapply(ct$models, function(x) x$objfun)

    class(ct) <- c("trafotree", class(ct))
    ct
}

brf <- function(...)
    ### NOTE: nodesize is minsplit NOT minbucket (in rpart terminology)
    randomForest(..., keep.inbag = TRUE, replace = FALSE, nodesize = 25)

myrf <- function(rf, trtf) {
    trtf$rf <- rf
    class(trtf) <- c(class(trtf)[1], "myrf", class(trtf)[-1])
    trtf
}

mycf <- function(cf, trtf) {
    trtf$cf <- cf
    class(trtf) <- c(class(trtf)[1], "mycf", class(trtf)[-1])
    trtf
}

predict.mycf <- function(object, newdata = object$data, OOB = FALSE, type = "weights", ...)
    predict(object$cf, newdata = newdata, OOB = FALSE, type = "weights")

predict.myrf <- function(object, newdata = object$data, OOB = FALSE, type = "weights", ...) {

    type <- match.arg(type)

    lnodes <- attr(predict(object$rf, newdata = object$data, nodes = TRUE), "nodes")

    fdata <- as.list(as.data.frame(lnodes))
    if (OOB) {
        fnewdata <- list()
    } else {
        nnodes <- attr(predict(object$rf, newdata = newdata, nodes = TRUE), "nodes")
        fnewdata <- as.list(as.data.frame(nnodes))
    }

    ### extract weights
    rw <- as.list(as.data.frame(object$rf$inbag))

    w <- partykit:::.rfweights(fdata, fnewdata, rw, scale = TRUE)

    return(w)
}
    
simfun <- function(..., nsim = 100, methods = c("ct", "tt", "ttExSplit", "ttM", "ttSD", "ttBern", 
    "ttBernExSplit", "tfbagg", "tfrf", "tfrfExSplit", "tfM", "tfSD", "tfbaggBern", "tfrfBern", 
    "tfrfBernExSplit", "cfbagg", "cfrf", "bagg", "rf", "rfBern"), prob = c(.1, .5, .9), bounds = c(-Inf, Inf)) {

    methods <- match.arg(methods, several.ok = TRUE)

    ll <- matrix(NA, nrow = nsim, ncol = length(methods) + 1)
    colnames(ll) <- c("gt", methods)
    time <- matrix(NA, nrow = nsim, ncol = length(methods) + 1)
    colnames(time) <- colnames(ll)
    covr <- matrix(NA, nrow = nrow(ll), ncol = ncol(ll))
    colnames(covr) <- colnames(ll)
    clen <- matrix(NA, nrow = nrow(ll), ncol = ncol(ll))
    colnames(clen) <- colnames(ll)
    c_covr <- matrix(NA, nrow = nrow(ll), ncol = ncol(ll))
    colnames(c_covr) <- colnames(ll)
    c_clen <- matrix(NA, nrow = nrow(ll), ncol = ncol(ll))
    colnames(c_clen) <- colnames(ll)
    L1 <- matrix(NA, nrow = nrow(ll), ncol = ncol(ll))
    colnames(L1) <- colnames(ll)
    checkRisk <- array(NA, dim = c(nrow(ll), ncol(ll), length(prob)),
                       dimnames = list(NULL, colnames(ll), as.character(prob)))

for (i in 1:nrow(ll)) {

    print(i)
    l1 <- dgp(...)
    args <- list(...)
    args$n <- args$ntest
    t1 <- do.call("dgp", args)

    fm <- y ~ .
    mtrybagg <- max(1, ncol(l1) - 1)
    mtryrf <- max(1, floor((ncol(l1) - 1) / 3))

    b <- as.basis(~ y, data = l1, ui = matrix(0:1, nrow = 1), ci = 0)
    m <- ctm(b, todistr = "Normal")

    qu <- quantile(l1$y, prob = c(.05, .95))
    add <- c(min(l1$y) - qu[1], max(l1$y) - qu[2])
    if (is.finite(bounds[1])) add[1] <- 0
    yvar <- numeric_var("y", support = qu, bounds = bounds, add = add)
    mBern <- ctm(Bernstein_basis(yvar, ui = "increasing", order = 5), todistr = "Normal")

    models <- vector(mode = "list", length = length(methods))
    names(models) <- methods

    ctrlExSplit <- ctrl
    ctrlExSplit$splitflavour <- "exhaustive"
    ctrlExSplit$restart <- FALSE

    tm <- function(...) system.time(...)[["user.self"]]

    if ("tt" %in% methods)
        time[i, "tt"] <- tm(models[["tt"]] <- trafotree(m, parm = 1:2, formula = fm, data = l1))

    if ("ct" %in% methods) {
       stopifnot(!is.null(models[["tt"]]))
       time[i, "ct"] <- tm(models[["ct"]] <- Ctree(formula = fm, data = l1, tt = models[["tt"]]))
    }

    if ("ttExSplit" %in% methods) {
        ttctrlExSplit <- ctree_control()
        ttctrlExSplit$splitflavour <- "exhaustive"
        ttctrlExSplit$restart <- FALSE
        time[i, "ttExSplit"] <- tm(models[["ttExSplit"]] <- trafotree(m, parm = 1:2, formula = fm, 
            data = l1, control = ttctrlExSplit))
    }

    if ("ttM" %in% methods)
        time[i, "ttM"] <- tm(models[["ttM"]] <- trafotree(m, parm = 1, formula = fm, data = l1))

    if ("ttSD" %in% methods)
        time[i, "ttSD"] <- tm(models[["ttSD"]] <- trafotree(m, parm = 2, formula = fm, data = l1))

    if ("tfbagg" %in% methods)
        time[i, "tfbagg"] <- tm(models[["tfbagg"]] <- traforest(m, parm = 1:2, formula = fm, data = l1, mtry = mtrybagg,
                        ntree = ntree, control = ctrl))

    if ("tfrf" %in% methods)
        time[i, "tfrf"] <- tm(models[["tfrf"]] <- traforest(m, parm = 1:2, formula = fm, data = l1, mtry = mtryrf,
                        ntree = ntree, control = ctrl))

    if ("tfrfExSplit" %in% methods)
        time[i, "tfrfExSplit"] <- tm(models[["tfrfExSplit"]] <- traforest(m, parm = 1:2, formula = fm, data = l1, mtry = mtryrf,
                        ntree = ntree, control = ctrlExSplit))

    if ("cfbagg" %in% methods) {
        stopifnot(!is.null(models[["tfbagg"]]))
        time[i, "cfbagg"] <- tm(models[["cfbagg"]] <- mycf(cforest(formula = fm, data = l1, mtry = mtrybagg,
                        ntree = ntree, control = ctrl), trtf = models[["tfbagg"]]))
    }

    if ("cfrf" %in% methods) {
        stopifnot(!is.null(models[["tfrf"]]))
        time[i, "cfrf"] <- tm(models[["cfrf"]] <- mycf(cforest(formula = fm, data = l1, mtry = mtryrf,
                        ntree = ntree, control = ctrl), trtf = models[["tfrf"]]))
    }

    if ("tfbaggM" %in% methods)
        time[i, "tfbaggM"] <- tm(models[["tfbaggM"]] <- traforest(m, parm = 1, formula = fm, data = l1, mtry = mtrybagg,
                                     ntree = ntree, control = ctrl))

    if ("tfbaggSD" %in% methods)
        time[i, "tfbaggSD"] <- tm(models[["tfbaggSD"]] <- traforest(m, parm = 2, formula = fm, data = l1, mtry = mtrybagg,
                                      ntree = ntree, control = ctrl))

    if ("ttBern" %in% methods)
        time[i, "ttBern"] <- tm(models[["ttBern"]] <- trafotree(mBern, formula = fm, data = l1))

    if ("ttBernExSplit" %in% methods) {
        ttctrlExSplit <- ctree_control()
        ttctrlExSplit$splitflavour <- "exhaustive"
        ttctrlExSplit$restart <- FALSE
        time[i, "ttBernExSplit"] <- tm(models[["ttBernExSplit"]] <- trafotree(mBern, formula = fm, 
            data = l1, control = ttctrlExSplit))
    }

    if ("tfbaggBern" %in% methods)
        time[i, "tfbaggBern"] <- tm(models[["tfbaggBern"]] <- traforest(mBern, formula = fm, data = l1, mtry = mtrybagg,
                            ntree = ntree, control = ctrl))

    if ("tfrfBern" %in% methods)
        time[i, "tfrfBern"] <- tm(models[["tfrfBern"]] <- traforest(mBern, formula = fm, data = l1, mtry = mtryrf,
                                          ntree = ntree, control = ctrl))

    if ("tfrfBernExSplit" %in% methods)
        time[i, "tfrfBernExSplit"] <- tm(models[["tfrfBernExSplit"]] <- traforest(mBern, formula = fm, data = l1, mtry = mtryrf,
                                          ntree = ntree, control = ctrlExSplit))

    if ("bagg" %in% methods) {
        stopifnot(!is.null(models[["tfbagg"]]))
        time[i, "bagg"] <- tm(models[["bagg"]] <- myrf(brf(fm, data = l1, ntree = ntree, mtry = mtrybagg), trtf = models[["tfbagg"]]))
    }

    if ("rf" %in% methods) {
        stopifnot(!is.null(models[["tfrf"]]))
        time[i, "rf"] <- tm(models[["rf"]] <- myrf(brf(fm, data = l1, ntree = ntree, mtry = mtryrf), trtf = models[["tfrf"]]))
    }

    if ("rfBern" %in% methods && !is.null(models[["tfrfBern"]])) {
        stopifnot(!is.null(models[["tfrfBern"]]))
        time[i, "rfBern"] <- tm(models[["rfBern"]] <- myrf(brf(fm, data = l1, ntree = ntree, mtry = mtryrf), trtf = models[["tfrfBern"]]))
    }

    ll[i, "gt"] <- loglik(t1, ...)

    for (m in methods) {
        if (is.null(models[[m]])) next()
        ll[i,m] <- logLik(models[[m]], newdata = t1)
    }

    tmp <- t1
    tmp$y <- NULL

    quant <- vector(mode = "list", length = length(methods) + 1L)
    names(quant) <- c("gt", methods)
    quant[["gt"]] <- qu(t1, prob = prob, ...)

    for (m in methods) {
        if (is.null(models[[m]])) next()
        qtmp <- predict(models[[m]], newdata = tmp, type = "quantile", prob = prob, K = 500)
        if (is.list(qtmp)) {
            qtmp <- lapply(qtmp, as.double)
            qtmp <- do.call("cbind", qtmp)
        }
        if (is.matrix(qtmp)) qtmp <- t(qtmp)
        quant[[m]] <- qtmp
    }

    for (m in c("gt", methods)) {
        if (m != "gt") {
            if (is.null(models[[m]])) next()
        }
        a <- try(u <- t1$y - quant[[m]])
        a <- try(um <- abs(u) * (u < 0) * matrix(1 - prob, ncol = length(prob), nrow = NROW(u), byrow = TRUE))
        a <- try(up <- abs(u) * (u >= 0) * matrix(prob, ncol = length(prob), nrow = NROW(u), byrow = TRUE))
        a <- try(checkRisk[i,m,] <- colSums(um + up))
    }
}

return(list(parm = list(...), ll = ll, time = time, checkRisk = checkRisk))
}

### Reproducibility material
###
###	Predictive Distribution Modelling Using Transformation Forests
###     by Torsten Hothorn & Achim Zeileis
###     doi:10.1080/10618600.2021.1872581
###
### and
###
###     Top-down Transformation Choice
###	by Torsten Hothorn
###	doi:10.1177/1471082X17748081
###

## ----setup, echo = FALSE, results = "hide", message = FALSE, warning = FALSE----
### from CRAN

pkgs <- c("lattice", "latticeExtra", "multcomp", "memisc", "Matrix",
          "colorspace", "grid", "libcoin", "inum", "partykit",
          "ATR", "trtf", "mlt")

ip <- rownames(installed.packages())
if (any(!pkgs %in% ip))
    install.packages(pkgs[!pkgs %in% ip], repos = "https://stat.ethz.ch/CRAN/")

OK <- sapply(pkgs, require, character.only = TRUE)
if (!all(OK)) 
    stop("package(s) ", paste(pkgs[!OK], collapse = ", "), " not available")

options(Matrix.warn = FALSE)


pdf("BMI.pdf", width = 12, height = 8)

col <- diverge_hcl(2, h = c(246, 40), c = 120, l = c(65, 90), alpha = .75)
trellis.par.set(list(plot.symbol = list(col="blue",pch=18, cex=0.5),
                     box.rectangle = list(col=1),
                     box.umbrella = list(lty=1, col=1),
                     strip.background = list(col = "white")))
ltheme <- canonical.theme(color = FALSE)     ## in-built B&W theme
ltheme$strip.background$col <- "transparent" ## change strip bg
lattice.options(default.theme = ltheme)

set.seed(290875)

### original SGB data may NOT be distributed.
### We use a simulated dataset. BMI values were simulated
### from the transformation forest. All explanatory
### variables (except sex and smoking) were permuted.
### Results will differ from what is printed in the manuscript.
frda <- system.file("BMI_artificial.rda", package = "trtf")
load(frda)
xvars <- c("age40", "edu", "swiss", "frveg", "physcat", "agramtag", "regl2")

# learn$age <- as.double(learn$age)
vn <- c(age = "Age", 
        agramtag = "Alcohol intake", 
        frveg = "Fruit and vegetables", 
        physcat = "Physical activity", 
        edu = "Education", 
        swiss = "Nationality", 
        regl2 = "Region",
        sex = "Sex",
        smoking = "Smoking")
learn$wght <- with(learn, wght / sum(wght) * nrow(learn))

order <- 5
d2c <- function(x) formatC(round(x, 2), digits = 2, format = "f")

## ----ss-ecdf, fig = "pdf", echo = FALSE, fig.height = 5------------------
myprepanel <- function (x, y, f.value = NULL, ...) 
{
    ans <- prepanel.default.qqmath(x, f.value = f.value, distribution = qunif)
    with(ans, list(xlim = ylim, ylim = c(0, 1), dx = dy, dy = dx))
}


mypanel <- function (x, y, f.value = NULL, type = "s", groups = NULL, qtype = 7, 
    ref = TRUE, ...) 
{
    if (ref) {
        reference.line <- trellis.par.get("reference.line")
        do.call(panel.abline, c(list(h = c(0, 1)), reference.line))
    }
    x <- as.numeric(x)
    distribution <- qunif
    nobs <- sum(!is.na(x))
    if (!is.null(groups)) {
        panel.superpose(x, y = NULL, f.value = f.value, type = type, 
            distribution = distribution, qtype = qtype, groups = groups, 
            panel.groups = panel.ecdfplot, ...)
    }
    else if (nobs) {
        if (is.null(f.value)) {
            panel.xyplot(x = sort(x), y = cumsum(y[order(x)]) / sum(y),
                type = type, ...)
        }
        else {
            p <- if (is.numeric(f.value)) 
                f.value
            else f.value(nobs)
            panel.xyplot(x = quantile(x, p, names = FALSE, type = qtype, 
                na.rm = TRUE), y = distribution(p), type = type, 
                ...)
        }
    }
}

#library("latticeExtra")
xyplot(wght~bmi | smoking + sex, data = learn, col = col[1], lwd = 3, xlab = "BMI", 
       panel = mypanel, prepanel = myprepanel, ylab = "Empirical CDF")
# ecdfplot(~ bmi | smoking + sex, data = learn, col = col[1], lwd = 3, xlab = "BMI")

## ----ss1, echo = FALSE---------------------------------------------------
### lm; cell-means model
lm_ss1 <- lm(bmi ~ smoking:sex - 1, data = learn, weights = learn$wght)
ll_ss1 <- logLik(lm_ss1)
sd_ss1 <- summary(lm_ss1)$sigma

## ----ss1-tab, echo = FALSE, results = "asis"-----------------------------
ci <- d2c(confint(lm_ss1))
tb <- paste(d2c(coef(lm_ss1)), paste(" (", ci[,1], "--",  ci[,2], sep = ""), ")", sep = "")
tb <- matrix(tb, ncol = 2)
attr(tb, "row.vars") <- list(Smoking = levels(learn$smoking))
attr(tb, "col.vars") <- list(Sex = levels(learn$sex))
class(tb) <- "ftable"
toLatex(tb, digits = 3, useDcolum = FALSE)

## ----ss1-plot, fig = "pdf", echo = FALSE, fig.height = 5-----------------
### set-up grid
b <- 150:400 / 10
sm <- sort(unique(learn$smoking))
se <- sort(unique(learn$sex))
nd <- expand.grid(bmi = b, smoking = sm, sex = se)

### compute normal probabilities
nd$mean <- predict(lm_ss1, newdata = nd)
nd$sd <- summary(lm_ss1)$sigma
nd$prob1 <- pnorm(nd$bmi, mean = nd$mean, sd = nd$sd)


plotfun <- function(prob, data, ...) {
    fm <- as.formula(paste(prob, "~ bmi | smoking + sex"))
    xyplot(fm, data = data, type = "l", 
        panel = function(x, y, subscripts, ...) {
            tmp <- subset(learn, sex == unique(nd[subscripts, "sex"]) & 
                                     smoking == unique(nd[subscripts, "smoking"]))
            mypanel(tmp$bmi, tmp$wght, lwd = 3, col = col[1])
            panel.xyplot(x, y, ...)
    }, col = col[2],  xlab = "BMI", ylab = "CDF", lwd = 2, ...)
}
plotfun("prob1", nd)#, main = "lm constant var")

## ----ss2, echo = FALSE---------------------------------------------------
### varying variances
sd_ss2 <- c(tapply(resid(lm_ss1), iss <- with(learn, interaction(smoking, sex)), sd))
ll_ss2 <- sum(learn$wght * dnorm(learn$bmi, mean = coef(lm_ss1)[iss], sd = sd_ss2[iss], log = TRUE))

## ----ss2-plot, fig = "pdf", echo = FALSE, fig.height = 5-----------------
nd$sd2 <- rep(sd_ss2, each = length(b))
nd$prob2 <- pnorm(nd$bmi, mean = nd$mean, sd = nd$sd2)
plotfun("prob2", nd)# , main = "lm varying var")

## ----ss3, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
bSMK <- as.basis(~ smoking - 1, data = learn)
bSEX <- as.basis(~ sex - 1, data = learn)
vBMI <- numeric_var("bmi", bounds = c(0, Inf),
    support = quantile(learn$bmi, prob = c(.01, .99), na.rm = TRUE),
    add = c(-5, 5))
bBMI <- Bernstein_basis(vBMI, order = order, ui = "increasing")
ctm_ss3 <- ctm(bBMI, interacting = b(sm = bSMK, sex = bSEX),
               data = learn, todistr = "Normal")
mlt_ss3 <- mlt(ctm_ss3, data = learn, scale = TRUE, weights = learn$wght)
ll_ss3 <- logLik(mlt_ss3)

## ----ss3-plot, fig = "pdf", echo = FALSE, message = FALSE, warning = FALSE, results = "hide", fig.height = 5----
nd$prob3 <- predict(mlt_ss3, newdata = nd, type = "distribution")
plotfun("prob3", nd)#, main = "nonlin mlt | smoking + sex")

## ----ss3-trafo-plot, fig = "pdf", echo = FALSE, message = FALSE, warning = FALSE, results = "hide", fig.height = 5----
nd3 <- nd
nd3$trafo3 <- predict(mlt_ss3, newdata = nd, type = "trafo")
bBMIlin <- as.basis(~ bmi, data = learn)
ctm_ss3a <- ctm(bBMIlin, interacting = b(sm = bSMK, sex = bSEX),
               data = learn, todistr = "Normal")
mlt_ss3a <- mlt(ctm_ss3a, data = learn, scale = TRUE, weights = learn$wght)
ll_ss3a <- logLik(mlt_ss3a)
nd3a <- nd
nd3a$trafo3 <- predict(mlt_ss3a, newdata = nd, type = "trafo")
nd3$type <- "Non-normal"
nd3a$type <- "Normal"
ndd <- rbind(nd3, nd3a)
xyplot(trafo3 ~ bmi | smoking + sex, data = ndd, type = "l", groups = type,
       col = col, lwd = 3, xlab = "BMI", ylab = expression(paste("Transformation Function ", h)))
# plotfun("trafo3", nd)#, main = "nonlin mlt | smoking + sex")

## ----ss3-plot-density, fig = "pdf", echo = FALSE, message = FALSE, warning = FALSE, results = "hide", fig.height = 5----
nd$density <- predict(mlt_ss3, newdata = nd, type = "density")
xyplot(density ~ bmi | smoking + sex, data = nd, type = "l", 
       col = col[2],  xlab = "BMI", ylab = "Density", lwd = 2)

## ----ss4, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
bSEXtrt <- as.basis(~ sex, data = learn)
bSMKtrt <- as.basis(~ smoking, data = learn, remove_intercept = TRUE, negative = TRUE)
### non-normal transformation model, conditional on sex with sex-specific smoking shift
ctm_ss4 <- ctm(bBMI, interacting = bSEX,
               shifting = b(smktrt = bSMKtrt, sextrt = bSEXtrt), todistr = "Normal")
mlt_ss4 <- mlt(ctm_ss4, data = learn, scale = TRUE, weights = learn$wght)
ll_ss4 <- logLik(mlt_ss4)

## ----ss4-plot, fig = "pdf", echo = FALSE, message = FALSE, warning = FALSE, results = "hide", fig.height = 5----
nd$prob4 <- predict(mlt_ss4, newdata = nd, type = "distribution")
plotfun("prob4", nd)#, main = "ninlin mlt | sex - smoking(sex)")

## ----ss5, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
### non-normal transformation model, conditional on sex with sex-specific smoking shift
ctm_ss5 <- ctm(bBMI, interacting = bSEX,
               shifting = b(smktrt = bSMKtrt, sextrt = bSEXtrt), todistr = "Logistic")
mlt_ss5 <- mlt(ctm_ss5, data = learn, scale = TRUE, weights = learn$wght)
ll_ss5 <- logLik(mlt_ss5)

## ----ss5-tab, echo = FALSE, results = "asis", warning = FALSE------------
K <- glht(lm(bmi ~ smoking, data = learn), linfct = mcp(smoking = "Dunnett"))$linfct[, -1]
X <- model.matrix(b(smktrt = bSMKtrt, sextrt = bSEXtrt), data = expand.grid(smoking = sm, sex = se))
### bSMKtrt has negative = TRUE, but we need -X %*% beta here!
X <- X * (-1)
K <- contrMat(table(learn$smoking), "Dunnett")
class(K) <- "matrix"
K <- as(K, "Matrix")
K2 <- as(bdiag(K, K), "matrix")
rownames(K2) <- c(paste("Female", rownames(K), sep = ":"), paste("Male", rownames(K), sep = ":"))
cf <- coef(mlt_ss5)
ci <- confint(glht(mlt_ss5, vcov. = vcov, linfct = K2 %*% cbind(matrix(0, ncol = length(cf) - ncol(X), nrow = nrow(X)), X)), 
              calpha = univariate_calpha())
tb <- paste(d2c(exp(ci$confint[,1])), paste(" (", d2c(exp(ci$confint[,2])), "--",
d2c(exp(ci$confint[,3])), sep = ""), ")", sep = "")
tb <- rbind(c(1, 1), matrix(tb, ncol = 2))
attr(tb, "row.vars") <- list(Smoking = levels(learn$smoking))
attr(tb, "col.vars") <- list(Sex = levels(learn$sex))
class(tb) <- "ftable"
toLatex(tb, digits = 3, useDcolum = FALSE)

## ----ss6, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
### non-normal transformation model, conditional on sex with smoking shift
ctm_ss6 <- ctm(bBMI, interacting = bSEX,
               shifting = bSMKtrt, todistr = "Logistic")
mlt_ss6 <- mlt(ctm_ss6, data = learn, scale = TRUE, weights = learn$wght)
ll_ss6 <- logLik(mlt_ss6)

## ----cmpx_tree, echo = FALSE, cache = TRUE-------------------------------
mod0 <- ctm(bBMI, todistr = "Logistic")
mtmp <- mlt(mod0, data = learn)
xvars2 <- xvars
xvars2[xvars2 == "age40"] <- "age"
fm <- as.formula(paste("bmi ~ sex + smoking + ", 
                       paste(xvars2, collapse = "+")))
ctrl <- ctree_control(maxdepth = 4, nmax = 100, 
                      alpha = 1 / 10000)
tlearn <- learn
levels(tlearn$edu) <- c("I", "II", "III")
levels(tlearn$smoking) <- c("N", "F", "L", "M", "H")
tlearn$smoking <- ordered(as.character(tlearn$smoking), 
                          levels = c("N", "F", "L", "M", "H"),
                          labels = c("N", "F", "L", "M", "H"))
levels(tlearn$physcat) <- c(">2", "1-2", "0")
tlearn$physcat <- ordered(as.character(tlearn$physcat),
                          levels = c("0", "1-2", ">2"),
                          labels = c("0", "1-2", ">2"))
iw <- tlearn$wght
iw[iw < .5] <- 1
iw <- as.integer(round(iw))
trtr <- trafotree(mod0, formula = fm, data = tlearn,
                  control = ctrl, weights = iw)
### needs newdata to compute logLik for unbinned data
ll_cmpx_tree <- logLik(trtr, newdata = tlearn)

## ----cmpx-tree-plot, echo = FALSE, fig = "pdf", fig.width = 13, fig.height = 8----
### plot(trtr, tp_args = list(type = "density", id = FALSE, ylines = 0, K = 100, fill = col[1]))

node_BMI <- function(obj, col = "black", bg = "white", fill = "transparent",
                     ylines = 2, id = TRUE, mainlab = NULL, gp = gpar(), K = 20,
                     type = c("trafo", "distribution", "survivor",  
                              "density", "logdensity", "hazard",    
                              "loghazard", "cumhazard", "quantile"),
                     flip = FALSE, axes = TRUE, xaxis = NULL, ...)
{
    mod <- obj$model
    q <- mkgrid(mod, n = K)[[mod$response]]
    type <- match.arg(type)

    if (type %in% c("distribution", "survivor")) {
        yscale <- c(0, 1)
    } else {
        pr <- trtf:::predict.trafotree(obj, q = q, type = type, ...)
        yscale <- range(pr) * c(1, 1.05)
    }
    xscale <- range(q)
    axes <- rep_len(axes, 2)

    ### panel function for ecdf in nodes
    rval <- function(node) {

        nid <- id_node(node)
        dat <- data_party(obj, nid)
        wn <- dat[["(weights)"]]   

        cf <- obj$coef[as.character(nid),]
        coef(mod) <- cf
        y <- as.vector(predict(mod, newdata = data.frame(1), q = q, type = type))

        ## set up plot
        q <- q - xscale[1]
        q <- q / diff(xscale)
        y <- y - yscale[1]
        y <- y / diff(yscale)

        top_vp <- viewport(layout = grid.layout(nrow = 1, ncol = 2,
                           widths = grid::unit(c(1, ylines), c("null", "lines")),
                           heights = grid::unit(1, "null")),
                           width = grid::unit(1, "npc"),
                           height = grid::unit(1, "npc"), # - unit(2, "lines"),
                           name = paste("node_mlt", nid, sep = ""), gp = gp)

        pushViewport(top_vp)
        grid.rect(gp = gpar(fill = bg, col = 0))

        plot <- viewport(layout.pos.col=1, layout.pos.row=1,
                         xscale = if(flip) yscale else xscale,
                         yscale = if(flip) xscale else yscale,
                         name = paste0("node_mlt", nid, "plot"),
                         clip = FALSE)

        pushViewport(plot)
        if(axes[1]) {
            if (is.null(xaxis)) {
                 grid.xaxis()
            } else {
                 if (nid == xaxis)
                     grid.xaxis()
            }
        }
        if(axes[2]) grid.yaxis(main = FALSE)
        grid.rect(gp = gpar(fill = "transparent"))
        grid.clip()
        if(flip) {
          if(fill != "transparent") {
            grid.polygon(c(min(y), y, min(y)), c(q[1], q, q[K]), gp = gpar(col = col, fill = fill))
          } else {
            grid.lines(y, q, gp = gpar(col = col))
          }
        } else {
          if(fill != "transparent") {
            grid.polygon(c(q[1], q, q[K]), c(min(y), y, min(y)), gp = gpar(col = col, fill = fill))
          } else {
            grid.lines(q, y, gp = gpar(col = col))
          }
        }
        upViewport(2)
    }

    return(rval)
}
class(node_BMI) <- "grapcon_generator"

nid <- min(nodeids(trtr, terminal = TRUE))

### weights are caseweights for ctree but not for this application
tmp <- unclass(trtr)
tmp$data[["(weights)"]] <- 1L
tmp$fitted[["(weights)"]] <- 1L
class(tmp) <- class(trtr)
plot(rotate(tmp), terminal_panel = node_BMI, nobs.loc='top',
     tp_args = list(type = "density", id = FALSE, ylines = 3, K = 100, fill = col[1], xaxis = nid))


## ----cmpx_forest, echo = FALSE, cache = TRUE, results = "hide", message = FALSE, warning = FALSE----
trtf <- traforest(mod0, formula = fm, data = learn, weights = learn$wght,
                  control = ctree_control(nmax = 100, mincriterion = 0, minsplit = 500), 
                  ntree = 100, mtry = 9, trace = TRUE)

## ----cmpx_vi, echo = FALSE, cache = TRUE---------------------------------
vi <- varimp(trtf)

## ----cmpx_ll, echo = FALSE, cache = TRUE---------------------------------
set.seed(290875)
n <- nrow(learn)
cf <- predict(trtf, newdata = learn, type = "coef")
i <- which(sapply(cf, function(x) any(abs(x) > 10)))
if (length(i) > 0) 
    cf[i] <- predict(trtf, newdata = learn[i,], type = "coef")
ll_cmpx_forest <- logLik(trtf, coef = cf)

## ----cmpx_predict, echo = FALSE, cache = TRUE----------------------------
### prepare figure
nd2 <- lapply(learn[,xvars2], function(x) {
    if (is.factor(x)) return(sort(x)[1])
    if (is.data.frame(x)) return(NULL)
    median(x)
})
nd2$sex <- sort(unique(learn$sex))
nd2$smoking <- sort(unique(learn$smoking))

nd3 <- nd2
nd3$age <- 18:70
nd4 <- expand.grid(nd3)
bmi <- 15:40

p <- predict(trtf, newdata = nd4, type = "distribution", q = bmi)

## ----cmpx-varimp-plot, echo = FALSE, fig = "pdf", fig.height = 5---------
vi <- vi[c("sex", "smoking", xvars2)]
vi[is.na(vi)] <- 0
# vi <- vi / sum(vi)
names(vi) <- vn[names(vi)]
vi <- vi[order(vi)]
barchart(~ vi, xlim = c(0, max(vi) * 1.1), xlab = "Mean Decrease
Log-Likelihood", col = col[1])

## ----cmpx-age-plot, echo = FALSE, fig = "pdf", fig.height = 5------------
tmp <- expand.grid(c(list(bmi = bmi), nd3))
tmp$p <- unlist(p)

contourplot(p ~ age + bmi | smoking + sex, data = tmp, 
            panel = function(x, y, z, ...) {
                panel.contourplot(x, y, z, ...)
#                 panel.rug(x, col = rgb(.1, .1, .1, .01))
            },
            at = 1:9/10, ylim = c(18, 34), xlab = "Age", ylab = "BMI",
            labels = list(cex = .5))

## ----cmpx-ctm, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
bAGE <- Bernstein_basis(numeric_var("age", support = c(20, 70)), order = order)
xfm <- as.formula(paste("~", paste(xvars[xvars != "age40"], collapse = "+")))
ctm_ctm <- ctm(bBMI, interacting = b(sex = bSEX, age = bAGE),
               shifting = c(sstrt = b(smktrt = bSMKtrt, sextrt = bSEX),
                            lp = as.basis(xfm, data = learn, remove_intercept = TRUE, negative = TRUE)),
             todistr = "Logistic")
mlt_ctm <- mlt(ctm_ctm, data = learn, scale = TRUE, weights = learn$wght)
ll_ctm <- logLik(mlt_ctm)

## ----cmpx-ctm-plot, echo = FALSE, fig = "pdf", fig.height = 5------------
nd3 <- nd2
nd3$age <- 18:70
nd3$bmi <- 15:35
nd3 <- nd3[c("bmi", names(nd3)[names(nd3) != "bmi"])]

p <- predict(mlt_ctm, newdata = nd3, type = "distribution")
tmp <- do.call("expand.grid", nd3)
tmp$p <- c(p)

contourplot(p ~ age + bmi | smoking + sex, data = tmp,
            at = 1:9/10, ylim = c(18, 34), xlab = "Age", ylab = "BMI",
            labels = list(cex = .5))

## ----dr, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
bAGE <- as.basis(~ age, data = learn)

### distribution regression | age 
ctm_dr <- ctm(bBMI, interacting = b(sex = bSEX, age = bAGE),
            shifting = c(sstrt = b(smktrt = bSMKtrt, sextrt = bSEX),
                            lp = as.basis(xfm, data = learn, remove_intercept = TRUE, negative = TRUE)),
            todistr = "Logistic")
mlt_dr <- mlt(ctm_dr, data = learn, scale = TRUE, weights = learn$wght)
ll_dr <- logLik(mlt_dr)

## ----dr-plot, echo = FALSE, fig = "pdf", fig.height = 5------------------
nd3 <- nd2
nd3$age <- 18:70
nd3$bmi <- 15:35
nd3 <- nd3[c("bmi", names(nd3)[names(nd3) != "bmi"])]

p <- predict(mlt_dr, newdata = nd3, type = "distribution")
tmp <- do.call("expand.grid", nd3)
tmp$p <- c(p)

contourplot(p ~ age + bmi | smoking + sex, data = tmp,
            at = 1:9/10, ylim = c(18, 34), xlab = "Age", ylab = "BMI",
            labels = list(cex = .5))

## ----smpl, echo = FALSE, cache = TRUE, message = FALSE, warning = FALSE, results = "hide"----
bAGE <- as.basis(~ age - 1, data = learn, negative = TRUE)
ctm_smpl <- ctm(bBMI, interacting = bSEX,
            shifting = c(sstrt = b(smktrt = bSMKtrt, sextrt = bSEX),
                         sexage = b(sex = bSEX, age = bAGE),
                         lp = as.basis(xfm, data = learn, remove_intercept = TRUE, negative = TRUE)),
            todistr = "Logistic")
mlt_smpl <- mlt(ctm_smpl, data = learn, scale = TRUE, weights = learn$wght)
ll_smpl <- logLik(mlt_smpl)

ciMale <- confint(mlt_smpl)["sexMale:age",]
ciFemale <- confint(mlt_smpl)["sexFemale:age",]

ageMale <- paste("$", d2c(exp(coef(mlt_smpl))["sexMale:age"]), "$ ($", d2c(exp(ciMale)[1]),
  "-", d2c(exp(ciMale)[2]), "$)", sep = "")

ageFemale <- paste("$", d2c(exp(coef(mlt_smpl))["sexFemale:age"]), "$ ($", d2c(exp(ciFemale)[1]),
  "-", d2c(exp(ciFemale)[2]), "$)", sep = "")



## ----summary-tab, echo = FALSE, results = "asis"-------------------------
cf_ctm <- coef(mlt_ctm)
cf_dr <- coef(mlt_dr)
cf_smpl <- coef(mlt_smpl)

xparm <- names(cf_ctm)[-grep("^Bs", names(cf_ctm))]

ci_ctm <- exp(confint(mlt_ctm)[xparm,])
ci_dr <- exp(confint(mlt_dr)[xparm,])
ci_smpl <- exp(confint(mlt_smpl)[xparm,])

tab <- cbind(paste(d2c(exp(cf_ctm[xparm])), " (", d2c(ci_ctm[,1]), "--", d2c(ci_ctm[,2]), ")", sep = ""),
paste(d2c(exp(cf_dr[xparm])), " (", d2c(ci_dr[,1]), "--", d2c(ci_dr[,2]), ")", sep = ""),
paste(d2c(exp(cf_smpl[xparm])), " (", d2c(ci_smpl[,1]), "--", d2c(ci_smpl[,2]), ")", sep = ""))

tab2 <- rbind(" ", rbind("$1$", tab))
rownames(tab2) <- 1:nrow(tab2)
rownames(tab2)[1:2] <- c("empty", "one")
rownames(tab2)[-(1:2)] <- xparm

tab2 <- tab2[c("empty", "one", "smokingFormer:sexFemale",
               "smokingLight:sexFemale",
               "smokingMedium:sexFemale",
               "smokingHeavy:sexFemale",
               "empty", "one", "smokingFormer:sexMale",
               "smokingLight:sexMale",
               "smokingMedium:sexMale",
               "smokingHeavy:sexMale",
               "agramtag", 
               "empty", "one", "frvegLow", 
               "empty", "one", "physcat1-2 d/w", "physcatnone",
               "empty", "one", "eduSecondary II", "eduTertiary",
               "empty", "one", "swissForeign", 
               "empty", "one", "regl2French", "regl2Italian"),,drop = FALSE]
rownames(tab2) <- c("Smoking (Females)", "\\hspace{.4cm}Never", "\\hspace{.4cm}Former", "\\hspace{.4cm}Light", "\\hspace{.4cm}Medium", "\\hspace{.4cm}Heavy",
                    "Smoking (Males)", "\\hspace{.4cm}Never", "\\hspace{.4cm}Former", "\\hspace{.4cm}Light", "\\hspace{.4cm}Medium", "\\hspace{.4cm}Heavy",
                    "Alcohol intake (g/d)",
                    "Fruit and vegetables", "\\hspace{.4cm}High", "\\hspace{.4cm}Low",
                    "Physical activity", "\\hspace{.4cm}High", "\\hspace{.4cm}Moderate", "\\hspace{.4cm}Low",
                    "Education", "\\hspace{.4cm}Mandatory (I)", "\\hspace{.4cm}Secondary (II)", "\\hspace{.4cm}Tertiary (III)",
                    "Nationality", "\\hspace{.4cm}Swiss", "\\hspace{.4cm}Foreign",
                    "Region", "\\hspace{.4cm}German speaking", "\\hspace{.4cm}French speaking", "\\hspace{.4cm}Italian speaking")
colnames(tab2) <- c("Model (\\ref{mod:ctm})", "Model (\\ref{mod:dr})", "Model (\\ref{mod:smpl})")
toLatex(tab2, digits = 3, useDcolum = FALSE)

## ----ss3-high, echo = FALSE, results = "hide", cache = TRUE--------------
bBMIh <- Bernstein_basis(vBMI, order = 2 * order, ui = "increasing")
ctm_ss3h <- ctm(bBMIh, interacting = b(sm = bSMK, sex = bSEX),
               data = learn, todistr = "Normal")
mlt_ss3h <- mlt(ctm_ss3h, data = learn, scale = TRUE, weights = learn$wght)
ll_ss3h <- logLik(mlt_ss3h)

dev.off()

sessionInfo()

load("results_empeval.rda")

i <- 1:nrow(ret)
nm <- c("p", "prod_shift", "prod_scale", "loglik")
m <- colnames(ret)
m <- m[!(m %in% nm)]

meth <- c('rsf', 'rg', 'cf', 'tf_W_alpha', 'tf_B_alpha', 'tf_W', 'tf_B', 'L1')
lab <- c('RSF NP(\U0001D6FC)', 
         'Ranger NP(\U0001D6FC)', 
         'CForest NP(\U0001D6FC)', 
         'TF W(\U0001D6FC)', 
         'TF Bs(\U0001D6FC)', 
         'TF W(\U0001D6DD)',
         'TF Bs(\U0001D6DD)',
         'L1') 


x <- expand.grid(i = i, m = m)
r <- ret[x$i,nm]
r$m <- factor(x$m, levels = meth, label = lab)
r$fll <- do.call("c", ret[,m])
r$id <- factor(i)
for (n in nm[1:3]) r[[n]] <- factor(r[[n]])

r$setup <- with(r, interaction(prod_shift, prod_scale))
levels(r$setup) <- lev <- c("No", "PH", "Non-PH", "Combined")

r$setup <- factor(as.character(r$setup), levels = rev(lev), labels = rev(lev))
levels(r$p) <- c("Low", "High")

r$fll[r$fll == 0] <- NA

library("lattice")

lwd <- 1.5
col <- rgb(.1, .1, .1, .05)
colR <- rgb(.75, 0, 0, .8)
colRlight <- rgb(.75, 0, 0, .1)
colB <- rgb(0, 0, 0.75, .8)
trellis.par.set(list(plot.symbol = list(col="black",pch=18, cex=0.75),
                     box.rectangle = list(col=1),
                     box.umbrella = list(lty=1, col=1),
                     strip.background = list(col = "white")))
ltheme <- canonical.theme(color = FALSE)     ## in-built B&W theme
ltheme$strip.background$col <- "transparent" ## change strip bg
lattice.options(default.theme = ltheme)

mypanel <- function(x, y, groups, subscripts, ...) {
    panel.abline(h = 0, lty = 3)
    panel.bwplot(x = x, y = y, ...)
    tapply(1:length(y), groups[subscripts], function(i) {
        llines(x = 1:nlevels(x), y = y[i][order(x[i])], 
               col = rgb(0.1, 0.1, 0.1, 0.1))
    })
}

bwplot(I(fll - loglik) ~ m | p + setup, data = r, groups = id,
       scales = list(x = list(rot = 45)), layout = c(2, 4), 
       panel = mypanel, ylim = c(-200, 25), 
       ylab = "Log-likelihood Difference")

rm(list=ls())

start_time <- Sys.time()

set.seed(12345)

load("ALS_samples.rda")

idx <- i <- INDEX

source("ALS_competitors.R")

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Define global parameters ------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

## number of trees
ntree <- 250

## maximal depth of trees
nodedepth <- 10

## number of obs in terminal node
minbucket <- 20

### parameters for party/partykit
ctrl_partykit <- partykit:::ctree_control(teststat = "quad", 
                                          testtype = "Univ", 
                                          mincriterion = 0,
                                          minbucket = minbucket,
                                          maxdepth = nodedepth)

if(!dir.exists("rda")){dir.create("rda")}

filename <- paste("rda/ALS_results_combined_", i, ".rda", sep = "")

NSim <- length(learn)

## number of variables to choose from in a node
mtry <- ceiling(sqrt(ncol(tmp) - 2))

res <- data.frame(polrconst =  rep(0, NSim))

res$polrconst <- res$polr <- 0
res$of_eq_ll <- res$of_prop_ll <- 0
res$tf_theta_ll <- res$tf_alpha_ll <- res$cf_ll <- 0

res$tf_of_qk <- res$tf_of_lk <- res$tf_of_ck <- 0

# head(res)

### Models without treatment
### erase Riluzole from the data
riluzole <- tmp$Riluzole
tmp$Riluzole <- NULL

### Unconditional Polr model: intercept only
polrc_fun <- function(i) {
  w <- 1:nrow(tmp) %in% learn[[i]] + 0L
  uncond_polr <- Polr(y ~ 1, data = tmp, weights = w)
  unclass(logLik(uncond_polr, w = 1 - w))
}
polrconst <- lapply(idx, polrc_fun)
res$polrconst[idx] <- unlist(polrconst)

### Prognostic Polr with exp(b)
polr_fun <- function(i) {
  w <- 1:nrow(tmp) %in% learn[[i]] + 0L
  prog_polr <- Polr(y ~ ., data = tmp, weights = w)
  unclass(logLik(prog_polr, w = 1 - w))
}
polrall <- lapply(idx, polr_fun)
res$polr[idx] <- unlist(polrall)

### Ordinal Forest (equally)
of_equal_fun <- function(i) {
  rf <- tryCatch(ordforest_equal(train = tmp[learn[[i]],],
                                 test = tmp[-learn[[i]],],
                                 mtry = mtry),
                 error = function(e) {NA})
  return(as.numeric(unclass(rf$of_eq_ll)))
}
of_eq_all <- lapply(idx, of_equal_fun)
res$of_eq_ll[idx] <- unlist(of_eq_all)
print('of_equal_fun is done!')
save(res, file = filename)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Ordinal Forest (proportional)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

of_prop_fun <- function(i) {
  rf <- tryCatch(ordforest_prop(train = tmp[learn[[i]],],
                                test = tmp[-learn[[i]],],
                                mtry = mtry),
                 error = function(e) {NA})
  return(list(of_prop_ll = rf$of_prop_ll,
              of_prop_y = rf$of_prop_y))
}
of_prop_all <- lapply(idx, of_prop_fun)
res$of_prop_ll[idx] <- as.numeric(of_prop_all[[1]]$of_prop_ll)
of_prop_y <- of_prop_all[[1]]$of_prop_y
print('ordforest_prop is done!')
save(res, file = filename)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Transformation forest OTF(theta)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
trtf_theta_fun <- function(i) {
  rf <- tryCatch(traforest_theta(train = tmp[learn[[i]],],
                                 test = tmp[-learn[[i]],],
                                 mtry = mtry),
                 error = function(e) {NA})
  return(list(tf_theta_ll = rf$tf_theta_ll,
              tf_theta_y = rf$tf_theta_y))
}
tf_theta_all <- lapply(idx, trtf_theta_fun)
res$tf_theta_ll[idx] <- as.numeric(tf_theta_all[[1]]$tf_theta_ll)
tf_theta_y <- tf_theta_all[[1]]$tf_theta_y
print("trtf_theta is done!")
save(res, file = filename)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Cohen's Kappa
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

kappa_tf_of <- ckappa(yobs1 = of_prop_y,
                      yobs2 = tf_theta_y)
res[idx, c("tf_of_ck",                                # unweighted kappa
           "tf_of_lk",                                # linear weighted kappa
           "tf_of_qk")] <- do.call("c", kappa_tf_of)  # quadratic weighted kappa

### save results
print("Cohen's Kappa(s) are done!")
save(res, file = filename)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Transformation forest OTF(alpha)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

trtf_alpha_fun <- function(i) {
  rf <- tryCatch(traforest_alpha(train = tmp[learn[[i]],], 
                                 test = tmp[-learn[[i]],], 
                                 mtry = mtry),
                 error = function(e) {NA})
  return(as.numeric(unclass(rf$tf_alpha_ll)))
}
tf_alpha_all <- lapply(idx, trtf_alpha_fun)
res$tf_alpha_ll[idx] <- unlist(tf_alpha_all)
print("trtf_alpha is done!")
save(res, file = filename)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Conditional Inference Forests
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

cf_fun <- function(i) {
  rf <- tryCatch(mycforest(train = tmp[learn[[i]],],
                           test = tmp[-learn[[i]],],
                           mtry = mtry),
                 error = function(e) {NA})
  return(as.numeric(unclass(rf$cf_ll)))
}
cf_all <- lapply(idx, cf_fun)
res$cf_ll[idx] <- unlist(cf_all)
print('cforest is done!')
save(res, file = filename)

end_time <- Sys.time()

print(end_time - start_time)

sessionInfo()

	
library("trtf")

nsim <- 100
ntree <- 1000

sigma <- function(d, tau = .01, prod_sigma = 1, ...)
    1 + pnorm(d$x1, sd = tau, mean = .5) * prod_sigma

mu <- function(d, tau = .01, prod_mu = 1, ...)
    pnorm(d$x2, sd = tau, mean = .5) * prod_mu

med <- function(d, ...)
    exp(mu(d, ...))

qu <- function(d, prob, prod_mu, prod_sigma, tau, ...)
    do.call("cbind", lapply(prob, function(p) qlnorm(p, meanlog = mu(d, tau = tau, prod_mu = prod_mu), 
                                                     sdlog = sigma(d, tau = tau, prod_sigma = prod_sigma))))

ry <- function(d, tau, prod_mu, prod_sigma, ...)
    exp(rnorm(nrow(d), mean = mu(d, tau = tau, prod_mu = prod_mu), sd = sigma(d, tau = tau, prod_sigma = prod_sigma)))

dgp <- function(n = 250, p = 5, ...) {
    d <- data.frame(x1 = runif(n), 
                    x2 = runif(n),
                    matrix(runif(n * p), nrow = n))
    d$y <- ry(d, ...)
    d	
}

loglik <- function(d, p, tau, prod_mu, prod_sigma, ...)
    sum(dlnorm(d$y, meanlog = mu(d, tau = tau, prod_mu = prod_mu), sdlog = sigma(d, tau = tau, prod_sigma = prod_sigma), log = TRUE))

parm <- expand.grid(nsim = nsim, n = 2500, ntest = 250, p = c(5, 500), tau = 0.01, 
                    prod_sigma = c(0, 1), prod_mu = c(0, 1))
parm <- lapply(1:nrow(parm), function(i) parm[i,])

source("artificial_simfun.R")

ctrl <- ctree_control(teststat = "quad", testtype = "Univ", 
                      mincriterion = 0, saveinfo = FALSE, 
                      minsplit = 25, minbucket = 10)

mc.reset.stream()
sf <- function(...) simfun(..., bounds = c(0, Inf),
                                methods = c("ct", 
                                            "tt", 
#                                            "ttExSplit",
                                            "ttM", 
                                            "ttBern", 
#                                            "ttBernExSplit",
#                                            "tfbagg", 
                                            "tfrf", 
#                                            "tfrfExSplit" 
#                                            "tfbaggBern", 
                                            "tfrfBern", 
#                                            "cfbagg", 
                                            "cfrf", 
#                                            "bagg", 
                                            "rf",
                                            "rfBern"
))
out <- mclapply(parm, function(x) do.call("sf", x), mc.cores = length(parm))

save(out, file = "lognormal_2d.rda")

start_time <- Sys.time()
library("parallel")
ncores <- 30

set.seed(290875)

source("competitors.R")
load("simulated_data.rda")

### number of repetitions
nsim <- length(simdat[[1]])

ret <- args[rep(1:nrow(args), each = nsim),]
simdat <- do.call("c", simdat)
ret$tf_W <- ret$tf_B <- ret$tf_W_alpha <- ret$tf_B_alpha <- 0
ret$cf <- ret$rsf <- ret$L1 <- ret$rg <- 0
ret$loglik <- unlist(do.call("c", loglik))

### filename to save estimated likelihoods
ret_filename <- 'results_empeval.rda'

### TSF Bs(theta)
trtf_fun <- function(i) {
    ## number of randomly preseleted variables to perform a split
    ## for low-dimensional data it is set to the number of variables
    ## for high-dimensional data it is set to the square root of the number of variables
    mtry <- ifelse(ret[i, "p"] <= 5, 
                   ncol(simdat[[i]]$train) - 1, 
                   ceiling(sqrt(ncol(simdat[[i]]$train) - 1)))
    
    ## learning and validation
    tf <- mytraforest(simdat[[i]]$train, simdat[[i]]$test, mtry = mtry)
    ## the learned forest is saved to be used in other methods
    save(tf, file=paste0(tf_dir, '/tf', i, '.rda'))
    unclass(tf$logLik_NN)
}
### the folder to save forest models
tf_folder <- 'results_tf'
dir.create(tf_folder)
tf_dir <- paste(getwd(), tf_folder, sep='/')

### learn and validate for each repetition of the simulated data
tf <- mclapply(1:length(simdat), trtf_fun, mc.cores = ncores)
ret$tf_B <- unlist(tf)
print('Traforest B is done!')
save(ret, file = ret_filename)

### DSF W(theta)
trtf_fun_Seibold <- function(i) {
    ## the corresponding TSF BS(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    tf <- mytraforest_Seibold(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(tf$logLik_NN)                        
}

### learn and validate for each repetition of the simulated data
tf <- mclapply(1:length(simdat), trtf_fun_Seibold, mc.cores = ncores)
ret$tf_W <- unlist(tf)
print('Traforest W is done!')
save(ret, file = ret_filename)

### TSF Bs(alpha)
trtf_fun_B <- function(i) {
    ## the corresponding TSF BS(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    tf <- mytraforest_B(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(tf$logLik_NN)
}

### learn and validate for each repetition of the simulated data
tf <- mclapply(1:length(simdat), trtf_fun_B, mc.cores = ncores)
ret$tf_B_alpha <- unlist(tf)
print('Traforest B_alpha is done!')
save(ret, file = ret_filename)

### DSF W(alpha)
trtf_fun_W <- function(i) {
    ## the corresponding TSF Bs(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    tf <- mytraforest_W(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(tf$logLik_NN)
}

### learn and validate for each repetition of the simulated data
tf <- mclapply(1:length(simdat), trtf_fun_W, mc.cores = ncores)
ret$tf_W_alpha <- unlist(tf)
print('Traforest W_alpha is done!')
save(ret, file = ret_filename)

### Ranger
library("ranger")
rg_fun <- function(i) {
    ## the corresponding TSF Bs(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    rg <- myranger(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(rg$logLik_NN)
}

### learn and validate for each repetition of the simulated data
rg <- mclapply(1:length(simdat), rg_fun, mc.cores = ncores)
ret$rg <- unlist(rg)
print('Ranger is done!')
save(ret, file = ret_filename)

### CForest
cf_fun <- function(i) {
    ## the corresponding TSF Bs(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    cf <- mycforest(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(cf$logLik_NN)
}

### learn and validate for each repetition of the simulated data
cf <- mclapply(1:length(simdat), cf_fun, mc.cores = ncores)
ret$cf <- unlist(cf)
print('Cforest is done!')
save(ret, file = ret_filename)

### RSF
library("randomForestSRC")
rsf_fun <- function(i) {
    ## the corresponding TSF Bs(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    rsf <- myrfsrc(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(rsf$logLik_NN)
}

### learn and validate for each repetition of the simulated data
rsf <- mclapply(1:length(simdat), rsf_fun, mc.cores = ncores)
ret$rsf <- unlist(rsf)
print('RSF is done!')

### L1 is a modified randomForestSRC, therefore we
### detach it to prevent the conflicts
detach(package:randomForestSRC)
save(ret, file = ret_filename)

### L1-splitting forest
### the library is not freely available
if {FALSE}
library("L1SRC")

L1_fun <- function(i) {
    ## the corresponding TSF Bs(theta) forest is loaded
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## learning and validation
    rsf <- myL1(tf, simdat[[i]]$train, simdat[[i]]$test)
    unclass(rsf$logLik_NN)
}

### learn and validate for each repetition of the simulated data
L1 <- mclapply(1:length(simdat), L1_fun, mc.cores = ncores)
ret$L1 <- unlist(L1)
print('L1 is done!')
}

### for each repetition of the simulated data
### the parameters of the corresponding model,
### the true model likelihood and
### the estimated likelihoods of all forest models are saved
save(ret, file = ret_filename)

end_time <- Sys.time()
print(end_time - start_time)

sessionInfo()

## load ALS data
## The data is available to registered users of https://nctu.partners.org/ProACT
## Register, download the data, and run
## R> library("TH.data"); demo("PROACT");
load("data/ALSFRSdata.rda")

set.seed(12345)

tmp <- ALSFRSdata

# <Field Name="10. Respiratory" ID="1214">
tmp$y <- tmp[["v_1214.halfyearafter"]]
tmp <- tmp[, -grep("halfyearafter", colnames(tmp), ignore.case = TRUE)]
tmp <- tmp[complete.cases(tmp),]

# dim(tmp)
# [1] 1013   71

NSim <- 120

learn <- list()
for (i in 1:NSim){
  while(TRUE) {
      learn[[i]] <- sample(1:nrow(tmp), ceiling(nrow(tmp) * .75))
      if (nlevels(tmp$y[learn[[i]], drop = TRUE]) == nlevels(tmp$y))
      break
  }
}

save(tmp, learn, file="ALS_samples.rda")

rm(list=ls())

library("lattice")
library("latticeExtra")

load("ALS_results_combined.rda")

lwd <- 1.5
col <- rgb(.1, .1, .1, .05)
colR <- rgb(.75, 0, 0, .8)
colRlight <- rgb(.75, 0, 0, .1)
colB <- rgb(0, 0, 0.75, .8)
ltheme <- canonical.theme(color = FALSE)     ## in-built B&W theme
ltheme$strip.background$col <- "transparent" ## change strip bg
lattice.options(default.theme = ltheme)

mypanel <- function(x, y, groups, subscripts, ...) {
    # panel.abline(h = 0, lty = 3)
    panel.bwplot(x = x, y = y, ...)
    tapply(1:length(y), groups[subscripts], function(i) {
        llines(x = 1:nlevels(x), y = y[i][order(x[i])], 
               col = rgb(0.1, 0.1, 0.1, 0.1))
    })
}

# only take 100 from 120 random splits into account
res <- res[complete.cases(res),][1:100,]

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Log-Likelihood ----------------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

lab <- c('polrconst' = "Polr()", 
         'polr' = "Polr(\U0001D6FC)",
         'of_eq_ll' = 'Ordinal Forest \n(equal)',
         'of_prop_ll' = 'Ordinal Forest \n(proportional)',
         'cf_ll' = 'CForest',
         'tf_alpha_ll' = 'OTF(\U0001D6FC)',
         'tf_theta_ll' = 'OTF(\U0001D6DD)')

i <- 1:nrow(res)

colnam.ll <- c("polrconst", "polr",
               "of_eq_ll", "of_prop_ll",
               "cf_ll", "tf_alpha_ll", "tf_theta_ll")
res.ll <- res[, colnam.ll]

m <- colnames(res.ll)
r <- expand.grid(i = i, m = m)
r$m <- factor(r$m, levels = names(lab), label = lab)
r$fll <- do.call("c", res.ll)
r$id <- factor(i)

ll.ALS.100.250.plt <- bwplot(fll ~ m, data = r, groups = id, 
                 scales = list(x = list(rot = 45,
                                        label = lab)),
                 main = "ALS SimStudy with ntree = 250",
                 panel = mypanel, ylab = "Log-likelihood")

### Figure 2
print(ll.ALS.100.250.plt)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Quadratic weighted Cohen's Kappa ----------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

lab <- c('tf_of_qk' = "Ordinal Forest (proportional) vs.\n OTF(\U0001D6DD)")

colnam.qwck <- c("tf_of_qk")
res.qwck <- data.frame(qwck = (res[ , colnam.qwck]))
res.qwck$lab <- factor(lab)
qwck.ALS.100.250.plt <- bwplot(qwck ~ lab, data = res.qwck, 
                               main = " ",
                               ylab = "Quadratic weighted\n Cohen's Kappa")

### Figure 3
print(qwck.ALS.100.250.plt)

rm(list=ls())

source("competitors.R")

library("lattice")
library("latticeExtra")

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### GLobal Variables --------------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

lwd <- 1.5
col <- rgb(.1, .1, .1, .05)
colR <- rgb(.75, 0, 0, .8)
colRlight <- rgb(.75, 0, 0, .1)
colB <- rgb(0, 0, 0.75, .8)
ltheme <- canonical.theme(color = FALSE)     ## in-built B&W theme

ltheme$layout.heights$strip <- 2.5
ltheme$layout.heights$axis.panel <- 2.5
ltheme$layout.heights$panel <- 2.5
ltheme$par.sub.text$lineheight <- 2
ltheme$par.main.text$cex <- 1
ltheme$par.main.text$lineheight <- 2.5

ltheme$strip.background$col <- "transparent" ## change strip bg
lattice.options(default.theme = ltheme)

LL.KL.mypanel <- function(x, y, groups, subscripts, ...){
    panel.abline(h = 0, lty = 3)
    panel.bwplot(x = x, y = y, ...)
    tapply(1:length(y), groups[subscripts], function(i) {
        llines(x = 1:nlevels(x), y = y[i][order(x[i])], 
               col = rgb(0.1, 0.1, 0.1, 0.1))
    })
}

plot.bwplot <- function(data.tmp, lab.tmp, ylab.tmp,
                        layout.tmp, mypanel,
                        title.tmp = rownames(sim_para)[p]){
    
    i <- 1:nrow(data.tmp)
    nm <- colnames(data.tmp)[param.colnam]
    m <- colnames(data.tmp)[-param.colnam]
    # m <- m[!(m %in% nm)]
    
    x <- expand.grid(i = i, m = m)
    r <- data.tmp[x$i, nm]
    
    r$m <- factor(x$m, levels = names(lab.tmp), label = lab.tmp)
    r$y <- do.call("c", data.tmp[,m])
    
    r$id <- factor(i)
    for (n in nm) r[[n]] <- factor(r[[n]])
    
    r$setup <- with(r, interaction(prod_shift, prod_scale))
    levels(r$setup) <- lev <- c("No", "PO", "Non-PO", "Combined")
    r$setup <- factor(as.character(r$setup), levels = lev, labels = lev)
    levels(r$p) <- c("Low-dimensional", "High-dimensional")
    
    plt <- bwplot(y ~ m | p + setup, data = r, groups = id,
                  between = (list(x = 0.25)),
                  par.settings = list(layout.heights = list(strip = 2.5)), 
                  scales = list(x = list(rot = 45, label = lab.tmp)),
                  as.table = TRUE,
                  layout = layout.tmp,
                  main = title.tmp,
                  panel = mypanel,
                  ylab = ylab.tmp)
    
    return(plt)
}

plot.ll.diff.bwplot <- function(data.tmp, lab.tmp, ylab.tmp,
                                layout.tmp, mypanel,
                                title.tmp = rownames(sim_para)[p]){
    
    i <- 1:nrow(data.tmp)
    nm <- c("p", "prod_shift", "prod_scale", "loglik")
    m <- names(lab.tmp)
    
    x <- expand.grid(i = i, m = m)
    r <- data.tmp[x$i, nm]
    
    r$m <- factor(x$m, levels = names(lab.tmp), label = lab.tmp)
    r$y <- do.call("c", data.tmp[,m])
    
    r$id <- factor(i)
    for (n in  c("p", "prod_shift", "prod_scale")) r[[n]] <- factor(r[[n]])
    
    r$setup <- with(r, interaction(prod_shift, prod_scale))
    levels(r$setup) <- lev <- c("No", "PO", "Non-PO", "Combined")
    r$setup <- factor(as.character(r$setup), levels = lev, labels = lev)
    levels(r$p) <- c("Low-dimensional", "High-dimensional")
    
    r$y[r$y == 0] <- NA
    
    plt <- bwplot(I(y - loglik) ~ m | p + setup, data = r, groups = id,
                  between = (list(x = 0.25)), #as.table = TRUE,
                  par.settings = list(layout.heights = list(strip = 2.5)), 
                  scales = list(x = list(rot = 45, label = lab.tmp)),
                  as.table = TRUE,
                  layout = layout.tmp,
                  main = title.tmp,
                  panel = mypanel,
                  ylab = ylab.tmp)
    
    return(plt)
}

load("../simulation/rda/results_empeval_nsim100_ntree250.rda")
res.100.250 <- res
load("../simulation/rda/results_empeval_nsim100_ntree2000.rda")
res.100.2000 <- res

param.colnam <- which(colnames(res) %in% c("p",
                                           "prod_shift",
                                           "prod_scale"))
ll.colnam <- grep("loglik", colnames(res))
ll.colnam <- c(ll.colnam, grep("_ll", colnames(res)))
ck.colnam <- grep("_ck", colnames(res))
lk.colnam <- grep("_lk", colnames(res))
qk.colnam <- grep("_qk", colnames(res))
KLI.colnam <- which(grepl("_KL_I\\b", colnames(res)))
KLII.colnam <- which(grepl("_KL_II\\b", colnames(res)))
KLIII.colnam <- which(grepl("_KL_III\\b", colnames(res)))

### nsim = 100, ntree = 250
res.100.250.ll <- res.100.250[, c(param.colnam, ll.colnam)]
res.100.250.ck <- res.100.250[, c(param.colnam, ck.colnam, lk.colnam, qk.colnam)]
res.100.250.KLI <- res.100.250[, c(param.colnam, KLI.colnam)]
res.100.250.KLII <- res.100.250[, c(param.colnam, KLII.colnam)]
res.100.250.KLIII <- res.100.250[, c(param.colnam, KLIII.colnam)]

### nsim = 100, ntree = 2000
res.100.2000.ll <- res.100.2000[, c(param.colnam, ll.colnam)]
res.100.2000.ck <- res.100.2000[, c(param.colnam, ck.colnam, lk.colnam, qk.colnam)]
res.100.2000.KLI <- res.100.2000[, c(param.colnam, KLI.colnam)]
res.100.2000.KLII <- res.100.2000[, c(param.colnam, KLII.colnam)]
res.100.2000.KLIII <- res.100.2000[, c(param.colnam, KLIII.colnam)]

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Log-Likelihood Differences ----------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ll.diff.100.250.plt <- plot.ll.diff.bwplot(data.tmp = res.100.250.ll,
                                           lab.tmp = c('of_eq_ll' = 'Ordinal Forest \n(equal)',
                                                       'of_prop_ll' = 'Ordinal Forest \n(proportional)',
                                                       'cf_ll' = 'CForest',
                                                       'tf_alpha_ll' = 'OTF(\U0001D6FC)', 
                                                       'tf_theta_ll' = 'OTF(\U0001D6DD)'),
                                           ylab.tmp = "Log-Likelihood Differences",
                                           layout.tmp = c(2, 4),
                                           mypanel = LL.KL.mypanel,
                                           title.tmp = NULL)
### Figure 1
print(useOuterStrips(ll.diff.100.250.plt))


ll.diff.100.2000.plt <- plot.ll.diff.bwplot(data.tmp = res.100.2000.ll,
                                            lab.tmp = c('of_eq_ll' = 'Ordinal Forest \n(equal)',
                                                        'of_prop_ll' = 'Ordinal Forest \n(proportional)',
                                                        'cf_ll' = 'CForest',
                                                        'tf_alpha_ll' = 'OTF(\U0001D6FC)', 
                                                        'tf_theta_ll' = 'OTF(\U0001D6DD)'),
                                            ylab.tmp = "Log-Likelihood Differences",
                                            layout.tmp = c(2, 4),
                                            mypanel = LL.KL.mypanel,
                                            title.tmp = NULL)
### Figure 4
print(useOuterStrips(ll.diff.100.2000.plt))

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Log-Likelihood ----------------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ll.100.250.plt <- plot.bwplot(data.tmp = res.100.250.ll,
                              lab.tmp = c('loglik' = 'True Log-Likelihood',
                                          'of_eq_ll' = 'Ordinal Forest \n(equal)',
                                          'of_prop_ll' = 'Ordinal Forest \n(proportional)',
                                          'cf_ll' = 'CForest',
                                          'tf_alpha_ll' = 'OTF(\U0001D6FC)', 
                                          'tf_theta_ll' = 'OTF(\U0001D6DD)'),
                              ylab.tmp = "Log-Likelihood",
                              layout.tmp = c(2, 4),
                              mypanel = LL.KL.mypanel,
                              title.tmp = NULL)
### Figure 5
print(useOuterStrips(ll.100.250.plt))

ll.100.2000.plt <- plot.bwplot(data.tmp = res.100.2000.ll,
                               lab.tmp = c('loglik' = 'True Log-Likelihood',
                                           'of_eq_ll' = 'Ordinal Forest \n(equal)',
                                           'of_prop_ll' = 'Ordinal Forest \n(proportional)',
                                           'cf_ll' = 'CForest',
                                           'tf_alpha_ll' = 'OTF(\U0001D6FC)', 
                                           'tf_theta_ll' = 'OTF(\U0001D6DD)'),
                               ylab.tmp = "Log-Likelihood",
                               layout.tmp = c(2, 4),
                               mypanel = LL.KL.mypanel,
                               title.tmp = NULL)
### Figure 6
print(useOuterStrips(ll.100.2000.plt))

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Kullback-Leibler Divergence ---------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Kullback-Leibler divergence KL comparing the true and estimated conditional
# probability densities according to Equation (7)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KLIII.100.250.plt <- plot.bwplot(data.tmp = res.100.250.KLIII * 500,
                                 lab.tmp = c('of_eq_KL_III' = 'Ordinal Forest \n(equal)',
                                             'of_prop_KL_III' = 'Ordinal Forest \n(proportional)',
                                             'cf_KL_III' = 'CForest',
                                             'tf_alpha_KL_III' = 'OTF(\U0001D6FC)', 
                                             'tf_theta_KL_III' = 'OTF(\U0001D6DD)'),
                                 ylab.tmp = "Kullback-Leibler Divergence KL",
                                 layout.tmp = c(2, 4),
                                 mypanel = LL.KL.mypanel,
                                 title.tmp = NULL)
### Figure 7
print(useOuterStrips(KLIII.100.250.plt))

KLIII.100.2000.plt <- plot.bwplot(data.tmp = res.100.2000.KLIII * 500,
                                  lab.tmp = c('of_eq_KL_III' = 'Ordinal Forest \n(equal)',
                                              'of_prop_KL_III' = 'Ordinal Forest \n(proportional)',
                                              'cf_KL_III' = 'CForest',
                                              'tf_alpha_KL_III' = 'OTF(\U0001D6FC)', 
                                              'tf_theta_KL_III' = 'OTF(\U0001D6DD)'),
                                  ylab.tmp = "Kullback-Leibler Divergence KL",
                                  layout.tmp = c(2, 4),
                                  mypanel = LL.KL.mypanel,
                                  title.tmp = NULL)
### Figure 8
print(useOuterStrips(KLIII.100.2000.plt))

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Results of the Kullback-Leibler divergence KL_1 according to Equation (8).
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KLI.100.250.plt <- plot.bwplot(data.tmp = res.100.250.KLI * 500,
                               lab.tmp = c('of_eq_KL_I' = 'Ordinal Forest \n(equal)',
                                           'of_prop_KL_I' = 'Ordinal Forest \n(proportional)',
                                           'cf_KL_I' = 'CForest',
                                           'tf_alpha_KL_I' = 'OTF(\U0001D6FC)', 
                                           'tf_theta_KL_I' = 'OTF(\U0001D6DD)'),
                               ylab.tmp = expression(paste("Kullback-Leibler Divergence ", KL[1])),
                               layout.tmp = c(2, 4),
                               mypanel = LL.KL.mypanel,
                               title.tmp = NULL)
### Figure 9
print(useOuterStrips(KLI.100.250.plt))

KLI.100.2000.plt <- plot.bwplot(data.tmp = res.100.2000.KLI * 500,
                                lab.tmp = c('of_eq_KL_I' = 'Ordinal Forest \n(equal)',
                                            'of_prop_KL_I' = 'Ordinal Forest \n(proportional)',
                                            'cf_KL_I' = 'CForest',
                                            'tf_alpha_KL_I' = 'OTF(\U0001D6FC)', 
                                            'tf_theta_KL_I' = 'OTF(\U0001D6DD)'),
                                ylab.tmp = expression(paste("Kullback-Leibler Divergence ", KL[1])),
                                layout.tmp = c(2, 4),
                                mypanel = LL.KL.mypanel,
                                title.tmp = NULL)
### Figure 10
print(useOuterStrips(KLI.100.2000.plt))

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Results of the Kullback-Leibler divergence KL_2 according to Equation (9).
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KLII.100.250.plt <- plot.bwplot(data.tmp = res.100.250.KLII * 500,
                                lab.tmp = c('of_eq_KL_II' = 'Ordinal Forest \n(equal)',
                                            'of_prop_KL_II' = 'Ordinal Forest \n(proportional)',
                                            'cf_KL_II' = 'CForest',
                                            'tf_alpha_KL_II' = 'OTF(\U0001D6FC)', 
                                            'tf_theta_KL_II' = 'OTF(\U0001D6DD)'),
                                ylab.tmp = expression(paste("Kullback-Leibler Divergence ", KL[2])),
                                layout.tmp = c(2, 4),
                                mypanel = LL.KL.mypanel,
                                title.tmp = NULL)
### Figure 11
print(useOuterStrips(KLII.100.250.plt)) 

KLII.100.2000.plt <- plot.bwplot(data.tmp = res.100.2000.KLII * 500,
                                 lab.tmp = c('of_eq_KL_II' = 'Ordinal Forest \n(equal)',
                                             'of_prop_KL_II' = 'Ordinal Forest \n(proportional)',
                                             'cf_KL_II' = 'CForest',
                                             'tf_alpha_KL_II' = 'OTF(\U0001D6FC)', 
                                             'tf_theta_KL_II' = 'OTF(\U0001D6DD)'),
                                 ylab.tmp = expression(paste("Kullback-Leibler Divergence ", KL[2])),
                                 layout.tmp = c(2, 4),
                                 mypanel = LL.KL.mypanel,
                                 title.tmp = NULL)
### Figure 12
print(useOuterStrips(KLII.100.2000.plt))

load('tr_results_empeval.rda')

meth <- c('w_alpha_beta', 'bs_alpha_beta', 'w_theta_beta', 'bs_theta_beta', 'w_theta_theta', 'bs_theta_theta')

tick_meth <- c('TF W(\U0001D6FC, \U0001D6FD)', 'TF Bs(\U0001D6FC, \U0001D6FD)', 
               'TF W(\U0001D6DD, \U0001D6FD)', 'TF Bs(\U0001D6DD, \U0001D6FD)',
               expression('TF W(\U0001D6DD, \U0001D6DD'['tr']*')'), expression('TF Bs(\U0001D6DD, \U0001D6DD'['tr']*')'))

#args <- data.frame(p = rep(sort(unique(ret$p)), times = 3),
#                   prod_shift_int = rep(c(1, 0, 1), each = 2),
#                   prod_scale_int = rep(c(0, 1, 1), each = 2))

#args$prod_shift_trt <- args$prod_shift_int
#args$prod_scale_trt <- args$prod_scale_int

args <- unique(ret[, c('p', 'prod_shift_int', 'prod_shift_trt', 'prod_scale_int', 'prod_scale_trt')])

LogLikDiff <- ret[, meth] -ret$loglik


cairo_pdf(file = 'tr_boxplots.pdf')
par(mfrow=c(3,2), oma = c(8, 6, 4, 8),
    mar=c(0, 0, 0, 0))


for (i in 1:nrow(args)){
  arg <- args[i,]
  idx <- (ret$prod_shift_int == arg$prod_shift_int) & 
         (ret$prod_shift_trt == arg$prod_shift_trt) & 
         (ret$prod_scale_int == arg$prod_scale_int) &
         (ret$prod_scale_trt == arg$prod_scale_trt) &
         (ret$p == arg$p)
  
  boxplot(LogLikDiff[idx, ], axes=FALSE, frame.plot=TRUE, ylim = c(-420, -20), at = c(1, 2, 3.5, 4.5, 6, 7))
  if (i > 4) {
    axis(1, col = "grey40", col.axis = "grey20", at = c(1, 2, 3.5, 4.5, 6, 7), labels=tick_meth, las=2, cex.axis = 1.5)
  }
  if (i %% 2 == 1){
    axis(2, col = "grey40", col.axis = "grey20", at=c(-50, -100, -150, -200, -250, -300, -350, -400))
  }
  if (i == 1){ mtext("low", side=3, line=1, cex = 1.2)}
  if (i == 2) {
    mtext('high', side=3, line=1, cex = 1.2)
    mtext('PH', side=4, line=1, las=2, cex = 1.2)}
  if (i == 4) {
    mtext('Non-PH', side=4, line=1, las=2, cex = 1.2)
  }
  if (i == 6){
    mtext('Combined', side=4, line=1, las=2, cex = 1.2)
  }
  if(i == 3){
    mtext('log-likelihood difference', outer=TRUE, side=2, line=3, cex = 1.2)
  }
}
dev.off()

library("lattice")

plotfun <- function(file, doplot = TRUE) {

load(file)

parm <- lapply(out, function(x) x$parm)

df <- NULL

for (i in 1:length(parm)) {

    tmp <- c()

    chkRsk <- out[[i]][["checkRisk"]]
    out[[i]][["L.1"]] <- chkRsk[,,1]
    out[[i]][["L.5"]] <- chkRsk[,,2]
    out[[i]][["L.9"]] <- chkRsk[,,3]
    nm <- names(out[[i]])
    out[[i]] <- out[[i]][nm[nm != "checkRisk"]]

    for (j in 2:length(out[[i]])) {
        dummy <- as.data.frame(out[[i]][[j]])
        vars <- 1:ncol(dummy)
        vnames <- colnames(dummy)[vars]
        dummy$id <- factor(1:nrow(dummy))

        dummy <- reshape(dummy, varying = list(vars), direction = "long", 
                         idvar = "id", timevar = "model", 
                         v.names = names(out[[i]])[j])
        dummy$model <- factor(dummy$model, levels = vars, labels = vnames)

        if (j == 2) { 
            tmp <- dummy
        } else {
            tmp <- merge(tmp, dummy, by = c("id", "model"))
        }
    }
    tmp <- cbind(tmp, as.data.frame(parm[i])[rep(1, nrow(tmp)),,drop = FALSE])
    df <- rbind(df, tmp)
}

save(df, file = paste(file, "_out.rda", sep = ""))

if (doplot) {
df <- subset(df, model != "ttBern")
df <- subset(df, model != "ttBernExSplit")

pdf(paste(file, ".pdf", sep = ""))
print(bwplot(ll ~ model | p + tau + prod_mu + prod_sigma, data = df, scales = list(y =
"free", x = list(rot = 45)), main = file))#, ylim = c(-800, -350)))
print(bwplot(time ~ model | p + tau + prod_mu + prod_sigma, data = df, scales = list(y =
"free", x = list(rot = 45)), main = file))#, ylim = c(-800, -350)))
print(bwplot(L.1 ~ model | p + tau + prod_mu + prod_sigma,, data = df,
scales = list(y = "free", x = list(rot = 45)), main = file))
print(bwplot(L.5 ~ model | p + tau + prod_mu + prod_sigma,, data = df,
scales = list(y = "free", x = list(rot = 45)), main = file))
print(bwplot(L.9 ~ model | p + tau + prod_mu + prod_sigma,, data = df,
scales = list(y = "free", x = list(rot = 45)), main = file))
dev.off()

}

}

plotfun("2d.rda", doplot = FALSE)

plotfun("lognormal_2d.rda", doplot = FALSE)

plotfun("friedman.rda", doplot = FALSE)

plotfun("lognormal_friedman.rda", doplot = FALSE)

plotfun("timings.rda", doplot = FALSE)

x <- readLines("ALS_analysis.R")

idx <- 1:120

if(!dir.exists("rda")){dir.create("rda")}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Calculate individiual results.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

for (i in idx){
  f <- paste("rda/run_", i, ".R", sep = "")
  writeLines(gsub("INDEX", i, x), con = f)
  system(paste("R CMD BATCH", f, ifelse(i < max(idx), " & ", "")))
  print(i)
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Combine individiual results.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

out <- 0

for (i in idx) {
  
  f <- paste("rda/ALS_results_combined_", i, ".rda", sep = "")
  
  if (file.exists(f)) {
    load(f)
    out <- out + res
  } else {
    cat(f, "missing \n")
  }
  
}

res <- out

res[res == 0] <- NA

save(res, file = paste("ALS_results_combined.rda", sep = ""))

library("trtf")
library("party")
library("tram")
library("survival")
set.seed(29)

### joint parameters for all forest procecdures
## number of trees
ntree <- 250
## maximal depth of trees
nodedepth <- 10
## number of obs in terminal node
minbucket <- 20

### parameters for Bernstein-based forest procedures
## order of approximation
ord <- 5
## whather to make log-transformation before the approximation
logf <- TRUE

### parameters for party/partykit
ctrl_partykit <- partykit:::ctree_control(teststat = "quad", 
                                          testtype = "Univ", 
                                          mincriterion = 0,
                                          minbucket = minbucket,
                                          maxdepth = nodedepth)

### these are _HACKS_ to extract NN weights from various forests
### trtf::predict.traforest calls these methods to obtain the nearest
### neighbor weights and then proceeds to estimate the coefficients
## method for party::cforest
predict.mycf <- function(object, newdata = object$data, OOB = FALSE, type = "weights", ...)
{
    stopifnot(!OOB)
    do.call("cbind", object$cf@prediction_weights(newdata = newdata, OOB = FALSE))
}
### method for for trtf::traforest
predict.mytrtf <- function(object, newdata = object$data, OOB = FALSE, type = "weights", ...)
{
    stopifnot(!OOB)
    predict(object$trtf, newdata = newdata, OOB = FALSE, type = "weights")
}
## method for randomForestSRC::rfsrc
predict.myrfsrc <- function(object, newdata = object$data, OOB = FALSE, type = "weights", ...)
{
    stopifnot(!OOB)
    nd <- newdata[,names(newdata) != "y"]   
    t(predict(object$rfsrc, newdata = nd, forest.wt = TRUE)$forest.wt)
}
### method for ranger::ranger
predict.myranger <- function(object, newdata = object$data, OOB = FALSE, 
                             type = "weights", ...) {
    stopifnot(!OOB)
    inbag <- object$rg$inbag.counts
    nnewdata <- predict(object$rg, data = newdata, type = "terminalNodes")$predictions
    ndata <- predict(object$rg, data = object$data, type = "terminalNodes")$predictions

    w <- 0L
    for (b in 1:length(inbag)) {
        fdata <- ndata[,b]
        fnewdata <- nnewdata[,b]
        ids <- unique(fdata)
        tw <- inbag[[b]]
        pw <- sapply(ids, function(i) tw * (fdata == i))
        ret <- pw[, match(fnewdata, ids), drop = FALSE]
        if (OOB) ret[,tw > 0] <- 0
        w <- w + ret
    }
    return(w)
}
### method for L1SRC::rfsrc
predict.myL1 <- function(object, newdata = object$data, OOB = FALSE, 
                         type = "weights", ...) {
    stopifnot(!OOB)
    inbag <- object$weights
    nnewdata <- predict(object$L1, newdata = newdata, membership = TRUE)$membership
    ndata <- predict(object$L1, newdata = object$data, membership = TRUE)$membership

    w <- 0L
    for (b in 1:length(inbag)) {
        fdata <- ndata[,b]
        fnewdata <- nnewdata[,b]
        ids <- unique(fdata)
        tw <- inbag[[b]]
        pw <- sapply(ids, function(i) tw * (fdata == i))
        ret <- pw[, match(fnewdata, ids), drop = FALSE]
        if (OOB) ret[,tw > 0] <- 0
        w <- w + ret
    }
    return(w)
}

### Transfromation survival forest with Bernstein basis and general score
### Bs(theta)
mytraforest <- function(learn, test, mtry = mtry, order = ord, log_first = logf) {
    
    ### the largest values of time are not used for Bernstein approximation
    sup <- c(sqrt(.Machine$double.eps), quantile(learn$y[,1], .95))
    m0 <- as.mlt(Coxph(y ~ 1, 
                       data = learn, 
                       order = order, 
                       log_first = log_first, 
                       sup = sup))

    ### traforest generates the subsampling weights later to be used
    ### by other methods
    rf <- traforest(m0, 
                    formula = y ~ ., 
                    data = learn,
                    ntree = ntree, 
                    trace = FALSE, 
                    mtry = mtry,
                    control = ctrl_partykit)

    ### nearest neighbor predictions
    cf <- predict(rf, newdata = test, type = "coef")
    llNN <- logLik(rf, newdata = test, coef = cf)
	
    list(logLik_NN = llNN, m0 = m0, object = rf, 
         mtry = mtry)

}

### Weibull distributional survival forests with general score
### W(theta)
mytraforest_Seibold <- function(trtf, learn, test) {
    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    m0 <- as.mlt(Survreg(y ~ 1, data = learn))

    ### split wrt intercept and log(y)
    rf <- traforest(m0, 
                    formula = y ~ ., 
                    data = learn,
                    ntree = ntree, 
                    trace = FALSE, 
                    mtry = mtry,
                    control = ctrl_partykit, 
                    weights = initw)

    ### use hack to wire rf  weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(logLik_NN = llNN, m0 = m0, object = rf)
}


### Weibull distributional survival forests with log-rank score
### W(alpha)
mytraforest_W <- function(trtf, learn, test) {                                                                                                                                     
    
    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"
    
    m0 <- as.mlt(Survreg(y ~ 1, data = learn))

    ### split wrt to intercept only
    rf <- traforest(m0,
                    formula = y ~ ., 
                    data = learn, 
                    parm = "(Intercept)",
                    ntree = ntree, 
                    trace = FALSE, 
                    mtry = mtry,
                    control = ctrl_partykit, 
                    weights = initw)

    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(logLik_NN = llNN, m0 = m0, object = rf)

}

### Trnasformation survival forests with Bernstein basis and log-rank score
### Bs(alpha)
mytraforest_B <- function(trtf, learn, test, order = ord, log_first = logf) {
    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"
    
    ### There is no intercept in Bernstein-basis, therefore to perform
    ### log-rank splitting constant variable is added to the data implicitly
    learn$alpha <- 1
    test$alpha <- 1
    
    ### the largest values of time are not used for Bernstein approximation
    sup <- c(sqrt(.Machine$double.eps), quantile(learn$y[, 1], .95))
    m0 <- as.mlt(Coxph(y ~ alpha, 
                       data = learn, 
                       order = order, 
                       log_first = log_first,
                       support = sup, 
                       fixed = c("alpha" = 0)))

    ### split wrt to intercept ("log-rank scores") only
    rf <- traforest(m0, 
                    formula = y | alpha ~ ., 
                    data = learn,
                    parm = "alpha",
                    mltargs = list(fixed = c("alpha" = 0)),
                    ntree = ntree, 
                    trace = FALSE, 
                    mtry = mtry,
                    control = ctrl_partykit, 
                    weights = initw)

    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(logLik_NN = llNN, m0 = m0, object = rf)
}


### cforests for survival data
mycforest <- function(trtf, learn, test) {

    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ctrl_party <- party:::cforest_unbiased(ntree = ntree, 
                                           maxdepth = nodedepth, 
                                           mtry = mtry, minbucket = minbucket)

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    ### log-rank splitting
    rf <- party::cforest(y ~ ., data = learn, weights = initw, 
                         control = ctrl_party)

    ### use hack to wire rf weights into predict.traforest
    trtf$cf <- rf
    class(trtf) <- c(class(trtf)[1], "mycf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)
    
    list(logLik_NN = llNN, object = rf)

}

### random forests SRC
myrfsrc <- function(trtf, learn, test) {

    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    ### rfsrc needs Surv() in the formula, otherwise
    ### y is treated as multivariate response
    learn$time <- learn$y[,1]
    learn$event <- learn$y[,2]
    learn$y <- NULL
    ### use log-rank splitting
    rf <- randomForestSRC::rfsrc(Surv(time, event) ~ ., 
                                 data = learn, 
                                 samp = initw, 
                                 bootstrap = "by.user", 
                                 ntree = ntree, 
                                 nodedepth = nodedepth, 
                                 nodesize = minbucket, 
                                 mtry = mtry, 
                                 splitrule = "logrank")
    learn$y <- with(learn, Surv(time, event))

    ### use hack to wire rf weights into predict.traforest
    trtf$rfsrc <- rf
    class(trtf) <- c(class(trtf)[1], "myrfsrc", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(logLik_NN = llNN, object = rf)
}

### L1 splitting
myL1 <- function(trtf, learn, test) {

    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    ### L1 needs Surv() in the formula, otherwise
    ### y is treated as multivariate response
    learn$time <- learn$y[,1]
    learn$event <- learn$y[,2]
    learn$y <- NULL
    ### use log-rank splitting
    rf <- L1SRC::rfsrc(Surv(time, event) ~ ., 
                       data = learn, 
                       samp = initw, 
                       bootstrap = "by.user", 
                       ntree = ntree, 
                       nodedepth = nodedepth, 
                       nodesize = minbucket, 
                       mtry = mtry, 
                       splitrule = "custom2") ### custom2 is L1 by Denis!
    learn$y <- with(learn, Surv(time, event))

    ### use hack to wire rf weights into predict.traforest
    trtf$L1 <- rf
    class(trtf) <- c(class(trtf)[1], "myL1", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)
    
    list(logLik_NN = llNN, object = rf)
}

### ranger
myranger <- function(trtf, learn, test) {

    mtry <- trtf$mtry
    m0 <- trtf$m0
    trtf <- trtf$object

    ### mytraforest generated the subsampling weights 
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    ### cannot use subsampling weights from mytraforest, cannot restrict
    ### maxdepths (use approximation via min.node.size)
    rf <- ranger(y ~ ., 
                 data = learn, 
                 num.trees = ntree, 
                 mtry = mtry, 
                 replace = FALSE, 
                 splitrule = "logrank",  
                 min.node.size = max(minbucket, nrow(learn) / 2^nodedepth), 
                 keep.inbag = TRUE)

    ### use hack to wire rf weights into predict.traforest
    trtf$rg <- rf
    class(trtf) <- c(class(trtf)[1], "myranger", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(logLik_NN = llNN, object = rf)

}

###simple example
if (FALSE) {
data("GBSG2", package = "TH.data")
GBSG2$y <- with(GBSG2, Surv(time, cens))
GBSG2$time <- GBSG2$cens <- NULL

ntree <- 25
test <- sample(1:NROW(GBSG2), 25)

tf <- mytraforest(GBSG2[-test,], GBSG2[test,], mtry = 3)
tfS <- mytraforest_Seibold(tf, GBSG2[-test,], GBSG2[test,])
tfB <- mytraforest_B(tf, GBSG2[-test,], GBSG2[test,])
tfW <- mytraforest_W(tf, GBSG2[-test,], GBSG2[test,])
cf <- mycforest(tf, GBSG2[-test,], GBSG2[test,])
library("randomForestSRC")
sf <- myrfsrc(tf, GBSG2[-test,], GBSG2[test,])
library("ranger")
rg <- myranger(tf, GBSG2[-test,], GBSG2[test,])
##not public; private patch from Denis Laroque
detach(package:randomForestSRC)
library("L1SRC")
L1 <- myL1(tf, GBSG2[-test,], GBSG2[test,])
detach(package:L1SRC)

tf[1]
tfS[1]
tfB[1]
tfW[1]
cf[1]
sf[1]
rg[1]
L1[1]

}

library("trtf")
library("tram")
library("party")
library("parallel")
library("survival")
set.seed(7782)

### method for for trtf::traforest
predict.mytrtf <- function(object, newdata = object$data, OOB = FALSE, type = "weights", ...)
{
    stopifnot(!OOB)
    predict(object$trtf, newdata = newdata, OOB = FALSE, type = "weights")
}

### joint parameters for all forest procecdures
## number of trees
ntree <- 250

## maximal depth of trees
nodedepth <- 10
## number of obs in terminal node
minbucket <- 20
## number of variables to choose from in a node

### parameters of Bernstein approximation
## polynom order
ord <- 5
## whether approximation applied to the log-transformed time or not
logf <- TRUE

### parameters for party/partykit
ctrl_partykit <- partykit:::ctree_control(teststat = "quad",
                                          testtype = "Univ",
                                          mincriterion = 0,
                                          minbucket = minbucket,
                                          maxdepth = nodedepth)


### Weibull distributional survival forest 
### with log-rank prognosic and predictive splitting
### W(alpha, beta)
w_alpha_beta <- function(trtf, learn, test, strata){
    mtry <- trtf$mtry
    trtf <- trtf$object

    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"
    
    m0 <- Survreg(y ~ trt, data = learn)
    rf <- traforest(as.mlt(m0),
                    formula = y | trt ~ .,
                    data = learn,
                    parm = c("(Intercept)", "trt"),
                    mtry = mtry,
                    ntree = ntree,
                    strata = strata,
                    control = ctrl_partykit,
                    weights = initw)

    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(loglik = llNN, object = rf)
}

### Weibull distributional survival forest
### with general prognostic splitting and log-rank predictive splitting
### W(theta, beta)
w_theta_beta <- function(trtf, learn, test, strata){
    mtry <- trtf$mtry
    trtf <- trtf$object
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    m0 <- Survreg(y ~ trt, data = learn)
    rf <- traforest(as.mlt(m0),
                    formula = y | trt ~ .,
                    data = learn,
                    mtry = mtry,
                    ntree = ntree,
                    strata = strata,
                    control = ctrl_partykit,
                    weights = initw)
    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(loglik = llNN, object = rf)
}

### Weibull distributional survival forest
### with general prognostic and predictive splitting
### W(theta, theta)
w_theta_theta <- function(trtf, learn, test, strata){
    mtry <- trtf$mtry
    trtf <- trtf$object
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    m0 <- Survreg(y | trt ~ 1, data = learn)
    rf <- traforest(as.mlt(m0),
                    formula = y | trt ~ .,
                    data = learn,
                    mtry = mtry,
                    ntree = ntree,
                    strata = strata,
                    control = ctrl_partykit,
                    weights = initw)
    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(loglik = llNN, object = rf)
}

### Transformation survival forest with Bernstein approximation and
### log-rank prognostic and predictive splitting
### Bs(alpha, beta)
bs_alpha_beta <- function(trtf, learn, test, strata, order = ord, log_first = logf){

    mtry <- trtf$mtry
    trtf <- trtf$object
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    ## since bernstein basis does not include an intercept
    ## we add it to the data implicitly
    learn$int <- 1
    test$int <- 1

    ## the largest time values are not used in approximation
    sup <- c(sqrt(.Machine$double.eps), quantile(learn$y[,1], .95))
    m0 <- Coxph(y ~ int + trt, 
                data = learn, 
                fixed = c("int"), 
                order = order, 
                log_first = log_first, 
                support = sup)
    rf <- traforest(as.mlt(m0),
                    formula = y | int + trt ~ .,
                    data = learn,
                    parm = c("int", "trt"),
                    mltargs = list(fixed = c("int" = 0)),
                    mtry = mtry,
                    ntree = ntree,
                    strata = strata,
                    control = ctrl_partykit,
                    weigths = initw)
    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(loglik = llNN, object = rf)
}

### Transformation survival forest with Bernstein approximation
### wtih general prognostic and log-rank predictive splitting
### BS(theta, beta)
bs_theta_beta <- function(trtf, learn, test, strata, order = ord, log_first = logf){
    mtry <- trtf$mtry
    trtf <- trtf$object
    initw <- do.call("cbind", trtf$weights)
    storage.mode(initw) <- "double"

    ## the largest time values are not used in approximation                  
    sup <- c(sqrt(.Machine$double.eps), quantile(learn$y[,1], .95))
    m0 <- Coxph(y ~ trt, 
                data = learn, 
                order = order, 
                log_first = log_first, 
                support = sup)
    rf <- traforest(as.mlt(m0),
                    formula = y | trt ~ .,
                    data = learn,
                    mtry = mtry,
                    ntree = ntree,
                    strata = strata,
                    control = ctrl_partykit,
                    weights = initw)
    ### use hack to wire rf weights into predict.traforest
    trtf$trtf <- rf
    class(trtf) <- c(class(trtf)[1], "mytrtf", class(trtf)[-1])

    ### nearest neighbor predictions
    cf <- predict(trtf, newdata = test, type = "coef")
    llNN <- logLik(trtf, newdata = test, coef = cf)

    list(loglik = llNN, object = rf)
}

### Transformation survival forest with Bernstein approximation
### with general prognostic and predictive splitting
### BS(theta, theta)
bs_theta_theta <- function(learn, test, strata, mtry, order = ord, log_first = logf){
    ## we do not pass any learned forest to this function since we use it
    ## for learning the first forest which is used for learning other forests
    ## the largest time values are not used in approximation                  
    sup <- c(sqrt(.Machine$double.eps), quantile(learn$y[,1], .95))
    m0 <- Coxph(y | trt ~ 1, 
                data = learn, 
                order = order, 
                log_first = log_first, 
                support = sup)
    rf <- traforest(as.mlt(m0),
                    formula = y | trt ~ .,
                    data = learn,
                    mtry = mtry,
                    ntree = ntree,
                    strata = strata,
                    control = ctrl_partykit)
    ### nearest neighbor predictions
    cf <- predict(rf, newdata = test, type = "coef")
    llNN <- logLik(rf, newdata = test, coef = cf)

    list(loglik = llNN, object = rf, mtry = mtry)
}

### Cox model with treatment and treatment-variable interactions
tcoxph <- function(learn, test, order = ord, log_first = logf){
    ## make model with treatment and treatment-variable interactions
    fm <- names(learn)
    fm <- paste(fm[!(fm %in% c("y", "trt"))], collapse = "+")
    fmy <- as.formula(paste("y ~ trt +", fm, " + trt * (", fm, ")"))

    ## the largest time values are not used in approximation
    sup <- c(sqrt(.Machine$double.eps), quantile(learn$y[,1], .95))
    cph <- Coxph(fmy , learn, order = order, log_first = log_first, support = sup)
    list(loglik = logLik(cph, newdata = test), object = cph)
}

### Weibull model with treatment and treatment-varibale interactions
tweib <- function(learn, test){
    ## make model with treatment and treatment-variable interactions
    fm <- names(learn)
    fm <- paste(fm[!(fm %in% c("y", "trt"))], collapse = "+")
    fmy <- as.formula(paste("y ~ trt +", fm, " + trt * (", fm, ")"))

    tw <- Survreg(fmy, learn)
    list(loglik = logLik(tw, newdata = test), object = tw)
}

### same as friedman.R

library("trtf")

nsim <- 10
ntree <- 1000

f1 <- function(x1, x2, x3, x4, x5, scale = TRUE) {
    ret <- 10 * sin(pi * x1 * x2) + 20 * (x3 - 0.5)^2 + 10 * x4 + 5 * x5
    if (scale) {
        ret <- ret - min(ret) 
        ret <- ret / max(ret)
        ret <- 3 * ret - 1.5
    }
    ret
}

sigma <- function(d, prod_sigma = 1, ...)
    exp(with(d, f1(X1, X2, X3, X4, X5, scale = TRUE)) * prod_sigma)

mu <- function(d, prod_mu = 1, ...)
    with(d, f1(X6, X7, X8, X9, X10, scale = TRUE)) * prod_mu

med <- function(d, ...)
    mu(d, ...)

qu <- function(d, prob, prod_mu, prod_sigma, ...)
    do.call("cbind", lapply(prob, function(p) qnorm(p, mean = mu(d, prod_mu = prod_mu), 
                                                    sd = sigma(d, prod_sigma = prod_sigma))))

ry <- function(d, prod_mu, prod_sigma, ...)
    rnorm(nrow(d), mean = mu(d, prod_mu = prod_mu), sd = sigma(d, prod_sigma = prod_sigma))

dgp <- function(n = 250, p = 5, ...) {
    d <- data.frame(matrix(runif(10 * n), ncol = 10),
                    matrix(runif(n * p, min = 0, max = 1), nrow = n))
    d$y <- ry(d, ...)
    d	
}


loglik <- function(d, p, prod_mu, prod_sigma, ...)
    sum(dnorm(d$y, mean = mu(d, prod_mu = prod_mu), sd = sigma(d, prod_sigma = prod_sigma), log = TRUE))

parm <- expand.grid(nsim = nsim, n = c(250, 500, 1000, 2000, 4000), 
                    ntest = 10, p = 5, tau = 0, prod_sigma = 1, prod_mu = 1)
parm <- lapply(1:nrow(parm), function(i) parm[i,])

source("artificial_simfun.R")

ctrl <- ctree_control(teststat = "quad", testtype = "Univ", 
                      mincriterion = 0, saveinfo = FALSE, 
                      minsplit = 25, minbucket = 10, nmax = 25)

mc.reset.stream()
sf <- function(...) simfun(..., methods = c("ct", 
                                            "tt", 
                                            "ttExSplit",
                                            "ttM", 
                                            "ttBern", 
#                                            "ttBernExSplit",
                                            "tfbagg", 
                                            "tfrf", 
#                                            "tfrfExSplit" 
                                            "tfbaggBern", 
                                            "tfrfBern", 
                                            "cfbagg", 
                                            "cfrf", 
                                            "bagg", 
                                            "rf"
))
out <- mclapply(parm, function(x) do.call("sf", x), mc.cores = length(parm))

save(out, file = "timings_nmax.rda")

### reproduces https://doi.org/10.48550/arXiv.1701.02110 Version 1

options(warn = -1L)

###################################################
### code chunk number 1: setup
###################################################

library("trtf")
library("quantregForest")

nmax <- Inf

pdf("applications.pdf", width = 12, height = 8)

data("abalone", package = "AppliedPredictiveModeling")
response <- "Rings"
abalone[[response]] <- as.numeric(abalone[[response]])

ns <- 100
fm <- Rings ~ Type + LongestShell + Diameter + Height + WholeWeight + ShuckedWeight + 
        VisceraWeight + ShellWeight
mtry <- ceiling(length(all.vars(fm[[3]])) / 3)
var_m <- numeric_var("Rings", support = quantile(abalone[[response]], prob = c(.1, .9)), 
                     add = range(abalone[[response]]) - quantile(abalone[[response]], prob = c(.1, .9)), 
                     bounds = c(0, Inf))

B_m <- Bernstein_basis(var_m, order = 4, ui = "increasing")
uc_ctm_AB <- ctm(B_m, data = abalone, todistr = "Normal")
uc_mlt_AB <- mlt(uc_ctm_AB, data = abalone, scale = FALSE)

c_ctm_AB <- ctm(B_m, data = abalone, todistr = "Normal", shift = fm[c(1, 3)])
c_mlt_AB <- mlt(c_ctm_AB, data = abalone, scale = TRUE)

tt_AB <- trafotree(uc_ctm_AB, formula = fm, data = abalone, 
                control = ctree_control(mincriterion = .95, minsplit = ns*2, minbucket = ns, nmax = nmax),
                mltargs = list(scale = TRUE))

if (FALSE) {
tf_AB <- traforest(uc_ctm_AB, formula = fm, data = abalone, ntree = 100, trace = TRUE, 
                  mltargs = list(scale = TRUE),
                  control = ctree_control(mincriterion = 0, minsplit = ns*2, minbucket = ns, nmax = nmax),
                                          mtry = mtry)

qrf_AB <- quantregForest(x = abalone[, all.vars(fm[[3]])], y = abalone[, all.vars(fm[[2]])], nodesize = ns,
                                       mtry = mtry, ntree = 100, keep.inbag = TRUE)
}

data("BostonHousing2", package = "mlbench")
response <- "cmedv"
BostonHousing2[[response]] <- as.numeric(BostonHousing2[[response]])

ns <- 40
fm <- cmedv ~ crim + zn + indus + chas + nox + rm + age + dis + rad + tax + ptratio + b + lstat
mtry <- ceiling(length(all.vars(fm[[3]])) / 3)
var_m <- numeric_var("cmedv", support = quantile(BostonHousing2[[response]], prob = c(.1, .9)), 
                     add = range(BostonHousing2[[response]]) - quantile(BostonHousing2[[response]], prob = c(.1, .9)), 
                     bounds = c(0, Inf))

B_m <- Bernstein_basis(var_m, order = 4, ui = "increasing")
uc_ctm_BH <- ctm(B_m, data = BostonHousing2, todistr = "Normal")
uc_mlt_BH <- mlt(uc_ctm_BH, data = BostonHousing2, scale = FALSE)

c_ctm_BH <- ctm(B_m, data = BostonHousing2, todistr = "Normal", shift = fm[c(1, 3)])
c_mlt_BH <- mlt(c_ctm_BH, data = BostonHousing2, scale = TRUE)

tt_BH <- trafotree(uc_ctm_BH, formula = fm, data = BostonHousing2, 
                control = ctree_control(mincriterion = .95, minsplit = 2*ns, minbucket = ns, nmax = nmax),
                mltargs = list(scale = TRUE))

if (FALSE) {
tf_BH <- traforest(uc_ctm_BH, formula = fm, data = BostonHousing2, ntree = 100, 
                  control = ctree_control(mincriterion = 0, minsplit = 2*ns, minbucket = ns,
                                          nmax = nmax), mtry = mtry, 
                  trace = TRUE, mltargs = list(scale = TRUE))

qrf_BH <- quantregForest(x = BostonHousing2[, all.vars(fm[[3]])], y = BostonHousing2[, all.vars(fm[[2]])], 
                                       nodesize = ns, mtry = mtry, ntree = 100, keep.inbag = TRUE)
}

if (FALSE) { ### alr3 was archived on CRAN
data("BigMac2003", package = "alr3")
response <- "BigMac"
BigMac2003[[response]] <- as.numeric(BigMac2003[[response]])

ns <- 20
fm <- BigMac ~ Bread + Rice + FoodIndex + Bus + Apt + TeachGI + 
        TeachNI + TaxRate + TeachHours
mtry <- ceiling(length(all.vars(fm[[3]])) / 3)
var_m <- numeric_var("BigMac", support = quantile(BigMac2003[[response]], prob = c(.1, .9)),
                     add = range(BigMac2003[[response]]) - quantile(BigMac2003[[response]], prob = c(.1, .9)), 
                     bounds = c(0, Inf))

B_m <- Bernstein_basis(var_m, order = 4, ui = "increasing")
uc_ctm_BM <- ctm(B_m, data = BigMac2003, todistr = "Normal")
uc_mlt_BM <- mlt(uc_ctm_BM, data = BigMac2003, scale = FALSE)

c_ctm_BM <- ctm(B_m, data = BigMac2003, todistr = "Normal", shift = fm[c(1, 3)])
c_mlt_BM <- mlt(c_ctm_BM, data = BigMac2003, scale = TRUE)

tt_BM <- trafotree(uc_ctm_BM, formula = fm, data = BigMac2003, 
                control = ctree_control(mincriterion = .95, minsplit = 2 * ns, minbucket = ns, nmax = nmax),
                mltargs = list(scale = TRUE))

if (FALSE) {
tf_BM <- traforest(uc_ctm_BM, formula = fm, data = BigMac2003, ntree = 100, 
                  control = ctree_control(mincriterion = 0, minsplit = 2*ns, minbucket = ns, 
                  nmax = nmax), mltargs = list(scale = TRUE),
                  trace = TRUE, mtry = mtry)

qrf_BM <- quantregForest(x = BigMac2003[, all.vars(fm[[3]])], 
    y = BigMac2003[, all.vars(fm[[2]])], nodesize = ns, mtry = mtry, ntree = 100, keep.inbag = TRUE)
}

}
data("Ozone", package = "mlbench")
Ozone <- subset(Ozone, complete.cases(Ozone))
Ozone <- as.data.frame(lapply(Ozone, function(x) {
    x <- x[, drop = TRUE]
    if (is.factor(x)) return(as.ordered(x))
    x
}))
response <- "V4"
Ozone[[response]] <- as.numeric(Ozone[[response]])

ns <- 20
fm <- V4 ~ V1 + V2 + V3 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 + V13
mtry <- ceiling(length(all.vars(fm[[3]])) / 3)
var_m <- numeric_var("V4", support = quantile(Ozone[[response]], prob = c(.1, .9)), 
                     add = range(Ozone[[response]]) - quantile(Ozone[[response]], prob = c(.1, .9)), 
                     bounds = c(0, Inf))

B_m <- Bernstein_basis(var_m, order = 4, ui = "increasing")
uc_ctm_Ozone <- ctm(B_m, data = Ozone, todistr = "Normal")
uc_mlt_Ozone <- mlt(uc_ctm_Ozone, data = Ozone, scale = FALSE)

c_ctm_Ozone <- ctm(B_m, data = Ozone, todistr = "Normal", shift = fm[c(1, 3)])
c_mlt_Ozone <- mlt(c_ctm_Ozone, data = Ozone, scale = TRUE)

tt_Ozone <- trafotree(uc_ctm_Ozone, formula = fm, data = Ozone, 
                control = ctree_control(mincriterion = .95, minsplit = 2*ns, minbucket = ns, nmax = nmax),
                mltargs = list(scale = TRUE))

tf_Ozone <- traforest(uc_ctm_Ozone, formula = fm, data = Ozone, ntree = 100, 
                  control = ctree_control(mincriterion = 0, minsplit = 2*ns,
                  minbucket = ns, nmax = nmax), trace = TRUE, mtry = mtry, 
                  mltargs = list(scale = TRUE))

qrf_Ozone <- quantregForest(x = Ozone[, all.vars(fm[[3]])], y = Ozone[, all.vars(fm[[2]])], 
                                          nodesize = ns, mtry = mtry, ntree = 100, keep.inbag = TRUE)


library("lattice")
lwd <- 1.5
col <- rgb(.1, .1, .1, .05)
colR <- rgb(.75, 0, 0, .8)
colB <- rgb(0, 0, 0.5, .8)
trellis.par.set(list(plot.symbol = list(col="blue",pch=18, cex=0.5),
                     box.rectangle = list(col=1),
                     box.umbrella = list(lty=1, col=1),
                     strip.background = list(col = "white")))
ltheme <- canonical.theme(color = FALSE)     ## in-built B&W theme
ltheme$strip.background$col <- "transparent" ## change strip bg
lattice.options(default.theme = ltheme)
bpch <- 18
bfill <- FALSE

###################################################
### code chunk number 4: Ozone-uctm
###################################################
q <- mkgrid(uc_mlt_Ozone, n = 100)[[1]]
d <- predict(uc_mlt_Ozone, newdata = Ozone, q = q, type="density")
plot(d ~ q, type = "l", ylab = "Density", xlab = "Daily Maximum One-hour-average Ozone Reading")
rug(Ozone$V4, col = rgb(.1, .1, .1, .1), lwd = 2)


###################################################
### code chunk number 5: Ozone-ltm
###################################################
q <- mkgrid(uc_mlt_Ozone, n = 100)[[1]]

d <- predict(c_mlt_Ozone, newdata = Ozone, q = q, type="distribution")
lp <- c(predict(c_mlt_Ozone, newdata = Ozone, q = 1, terms = "bshift"))
nd <- expand.grid(q = q, lp = -lp)
nd$d <- c(d)
pfun <- function(x, y, z, subscripts, at, ...) {
    panel.contourplot(x, y, z, subscripts, at = 1:9/10, ...)
    panel.xyplot(x = -lp, y = Ozone$V4, pch = 19, cex = 1.5,
                 col = rgb(.1, .1, .1, .1), ...)
}
print(contourplot(d ~ lp + q, data = nd, panel = pfun, xlab = "Linear Predictor", 
                  ylab = "Observed", ylim = c(0, 40)))


###################################################
### code chunk number 6: Ozone-trtree
###################################################
plot(tt_Ozone, tp_args = list(type = "density", id = FALSE, ylines = 0, K = 100))


###################################################
### code chunk number 8: Ozone-forest
###################################################
nx <- lapply(Ozone, function(x) {
    if (is.factor(x))
        return(factor(levels(x)[which.max(table(x))], levels = levels(x), ordered = is.ordered(x)))
    return(median(x))
})

nx$V8 <- 20:100
nx <- expand.grid(nx)

q1 <- predict(qrf_Ozone, newdata = nx, what = c(.2, .5, .8))

q2 <- predict(tf_Ozone, newdata = nx, K = 100, type = "quantile", p = c(.2, .5,.8))
q2 <- do.call("rbind", q2)

nx <- rbind(cbind(nx["V8"], q = q1[,1], p = "0.2", model = "QRF", stringsAsFactors = FALSE),
            cbind(nx["V8"], q = q1[,2], p = "0.5", model = "QRF", stringsAsFactors = FALSE),
            cbind(nx["V8"], q = q1[,3], p = "0.8", model = "QRF", stringsAsFactors = FALSE),
            cbind(nx["V8"], q = q2[,1], p = "0.2", model = "Forest", stringsAsFactors = FALSE),
            cbind(nx["V8"], q = q2[,2], p = "0.5", model = "Forest", stringsAsFactors = FALSE),
            cbind(nx["V8"], q = q2[,3], p = "0.8", model = "Forest", stringsAsFactors = FALSE))
nx$model <- factor(as.character(nx$model), levels = c("QRF", "Forest"), labels = c("QRF", "Forest"))

print(xyplot(q ~ V8 | model, data = nx, groups = p, type = "l", lty = 1:3, col = "black",
             xlab = "Sandburg, CA, Temperature in Fahrenheit",
             ylab = "Daily Maximum One-hour-average Ozone Reading"))

varimp(tf_Ozone)

###################################################
### code chunk number 2: abalone-ltm
###################################################
q <- mkgrid(uc_mlt_AB, n = 100)[[1]]

d <- predict(c_mlt_AB, newdata = abalone, q = q, type="distribution")
lp <- c(predict(c_mlt_AB, newdata = abalone, q = 1, terms = "bshift"))
nd <- expand.grid(q = q, lp = -lp)
nd$d <- c(d)
pfun <- function(x, y, z, subscripts, at, ...) {
    panel.contourplot(x, y, z, subscripts, at = 1:9/10, ...)
    panel.xyplot(x = -lp, y = abalone$Rings, pch = 19, cex = 1.5,
                 col = rgb(.1, .1, .1, .1), ...)
}
print(contourplot(d ~ lp + q, data = nd, panel = pfun, xlab = "Linear Predictor",
                  ylab = "Observed"))


###################################################
### code chunk number 3: abalone-trtree
###################################################
plot(tt_AB, tp_args = list(type = "density", id = FALSE, ylines = 0, K = 100))


###################################################
### code chunk number 6: BigMac-ltm
###################################################
if (FALSE) {
q <- mkgrid(uc_mlt_BM, n = 100)[[1]]

d <- predict(c_mlt_BM, newdata = BigMac2003, q = q, type="distribution")
lp <- c(predict(c_mlt_BM, newdata = BigMac2003, q = 1, terms = "bshift"))
nd <- expand.grid(q = q, lp = -lp)
nd$d <- c(d)
pfun <- function(x, y, z, subscripts, at, ...) {
    panel.contourplot(x, y, z, subscripts, at = 1:9/10, ...)
    panel.xyplot(x = -lp, y = BigMac2003$BigMac, pch = 19, cex = 1.5,
                 col = rgb(.1, .1, .1, .1), ...)
}
print(contourplot(d ~ lp + q, data = nd, panel = pfun, xlab = "Linear Predictor",
                  ylab = "Observed"))


###################################################
### code chunk number 7: BigMac-trtree
###################################################
plot(tt_BM, tp_args = list(type = "density", id = FALSE, ylines = 0, K = 100))
}

###################################################
### code chunk number 10: BH-ltm
###################################################
q <- mkgrid(uc_mlt_BH, n = 100)[[1]]

d <- predict(c_mlt_BH, newdata = BostonHousing2, q = q, type="distribution")
lp <- c(predict(c_mlt_BH, newdata = BostonHousing2, q = 1, terms = "bshift"))
nd <- expand.grid(q = q, lp = -lp)
nd$d <- c(d)
pfun <- function(x, y, z, subscripts, at, ...) {
    panel.contourplot(x, y, z, subscripts, at = 1:9/10, ...)
    panel.xyplot(x = -lp, y = BostonHousing2$cmedv, pch = 19, cex = 1.5,
                 col = rgb(.1, .1, .1, .1), ...)
}
print(contourplot(d ~ lp + q, data = nd, panel = pfun, xlab = "Linear Predictor",
                  ylab = "Observed"))


###################################################
### code chunk number 11: BH-trtree
###################################################
plot(tt_BH, tp_args = list(type = "density", id = FALSE, ylines = 0, K = 100))

dev.off()

sessionInfo()

library("survival")
nsim <- 100
ntrain <- 250
ntest <- 500
set.seed(33)

### scale and shape parameters of Weibull distribution
wscale <- 1
wshape <- 1

### Friedman function
f1 <- function(x1, x2, x3, x4, x5, scale = TRUE) {
  ret <- 10 * sin(pi * x1 * x2) + 20 * (x3 - 0.5)^2 + 10 * x4 + 5 * x5
  if (scale) {
    ret <- ret - min(ret) 
    ret <- ret / max(ret)
    ret <- 3 * ret - 1.5
  }
  ret
}

### data-dependent shift parameter of the model
shift_f <- function(d, prod_shift = c(1, 1), ...){
  with(d, f1(X1, X2, X3, X4, X5, scale = TRUE)) * prod_shift[[1]] + 
          with(d, f1(X6, X7, X8, X9, X10, scale = TRUE)) * prod_shift[[2]] * d$trt}

### data-dependent scale parameter of the model
scale_f <- function(d, prod_scale = c(1, 1), ...){
  exp(with(d, f1(X11, X12, X13, X14, X15, scale = TRUE)) * prod_scale[[1]] + 
      with(d, f1(X16, X17, X18, X19, X20, scale = TRUE)) * prod_scale[[2]] * d$trt)}

### Model of the outcome
rmodel <- function(n = 100, shift = 0, scale = 1) {
  ### Gompertz distribution
  FZ <- mlt:::.MinExtrVal()   
  q0 <- function(u) qweibull(u, scale = wscale, shape = wshape)
  q1 <- function(u, shift, scale) q0(FZ$p((FZ$q(u) - shift) / scale))
  y <- q1(runif(n), shift = shift, scale = scale)

  inf_idx <- is.infinite(y)
  ninf <- sum(inf_idx)
  while (ninf > 0){
    y[inf_idx] <- q1(runif(ninf), shift = shift[inf_idx], 
                     scale = scale[inf_idx]) 
    inf_idx <- is.infinite(y) 
    ninf <- sum(inf_idx)
  }
  
  return(pmax(.Machine$double.eps, y))
}

ry <- function(d, prod_shift, prod_scale,...)
  rmodel(nrow(d),
         shift = shift_f(d, prod_shift = prod_shift), 
         scale = scale_f(d, prod_scale = prod_scale))


### data generation
dgp <- function(ntrain, ntest, p = 5, prod_shift = c(0, 0), prod_scale = c(0, 0), ...) {
  n <- ntrain + ntest
  d <- data.frame(matrix(runif(20 * n), ncol = 20),
                  matrix(runif(n * p, min = 0, max = 1), nrow = n))
  d$trt <- sample(rep(0:1, length = n))
  d$y <- ry(d, prod_shift=prod_shift, prod_scale=prod_scale)
  d$y <- with(d, Surv(y, rep(1, nrow(d))))
  list('train' = d[1:ntrain,], 'test' = d[(ntrain+1):n,],
       prod_shift = prod_shift, prod_scale = prod_scale)
}

### True model log-likelihood
mylogLik <- function(d){
  shift <- shift_f(d$test, prod_shift = d$prod_shift)
  scale <- scale_f(d$test, prod_scale = d$prod_scale)
  pw <- function(u) pweibull(u, scale=wscale, shape=wshape)
  FZ <- mlt:::.MinExtrVal()
  h <- function(u) FZ$q(pw(u))
  dh <- function(u) - dweibull(u, scale=wscale, shape=wshape) / ((1 - pw(u)) * log(1 - pw(u)))
  ty <- h(d$test$y[,1]) * scale + shift
  dhy <- dh(d$test$y[,1])
  dh_idx <- !is.na(dhy) 
  sum(ty[dh_idx] - exp(ty[dh_idx]) + log(scale[dh_idx]) + log(dhy[dh_idx]))
}

if (FALSE) {
### some basic checks
library("tram")
N <- 10000

d <- dgp(N, N, prod_shift = c(0, 0), prod_scale = c(0, 0))
m0 <- as.mlt(Survreg(y ~ 1, data = d$train))
logLik(m0, newdata = d$test)
mylogLik(d)
coef(m0)

d <- dgp(N, N, prod_shift = c(1, 0), prod_scale = c(0, 0))
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
m1 <- as.mlt(Survreg(y ~ sh, data = d$train, fixed = c("sh" = -1)))
logLik(m1, newdata = d$test, parm = coef(m1))
mylogLik(d)
coef(m1)

d <- dgp(N, N, prod_shift = c(0, 1), prod_scale = c(0, 0))
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
m2 <- as.mlt(Survreg(y ~ sh, data = d$train, fixed = c("sh" = -1)))
logLik(m2, newdata = d$test, parm = coef(m2))
mylogLik(d)
coef(m2)

d <- dgp(N, N, prod_shift = c(1, 1), prod_scale = c(0, 0))                                                                                                                                                  
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
m3 <- as.mlt(Survreg(y ~ sh, data = d$train, fixed = c("sh" = -1)))
logLik(m3, newdata = d$test, parm = coef(m3))
mylogLik(d)
coef(m3)

d <- dgp(N, N, prod_shift = c(0, 0), prod_scale = c(1, 0))                                                                                                                                                 
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$train$sc <- scale_f(d$train, prod_scale = d$prod_scale)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
d$test$sc <- scale_f(d$test, prod_scale = d$prod_scale)
m4 <- as.mlt(Survreg(y | sc + 0 ~ sh, data = d$train, fixed = c("sh" = 0)))                                                                                                                                 
logLik(m4, newdata = d$test, parm = coef(m4))
mylogLik(d)
coef(m4)

d <- dgp(N, N, prod_shift = c(0, 0), prod_scale = c(0, 1))                                                                                                                                                  
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$train$sc <- scale_f(d$train, prod_scale = d$prod_scale)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
d$test$sc <- scale_f(d$test, prod_scale = d$prod_scale)
m5 <- as.mlt(Survreg(y | sc + 0 ~ sh, data = d$train, fixed = c("sh" = 0)))                                                                                                                                 
logLik(m5, newdata = d$test, parm = coef(m5))                                                                                                                                                               
mylogLik(d)
coef(m5)

d <- dgp(N, N, prod_shift = c(0, 0), prod_scale = c(1, 1))
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$train$sc <- scale_f(d$train, prod_scale = d$prod_scale)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
d$test$sc <- scale_f(d$test, prod_scale = d$prod_scale)
m6 <- as.mlt(Survreg(y | sc + 0 ~ sh, data = d$train, fixed = c("sh" = 0)))
logLik(m6, newdata = d$test, parm = coef(m6))
mylogLik(d)
coef(m6)

d <- dgp(N, N, prod_shift = c(1, 1), prod_scale = c(1, 1))
d$train$sh <- shift_f(d$train, prod_shift = d$prod_shift)
d$train$sc <- scale_f(d$train, prod_scale = d$prod_scale)
d$test$sh <- shift_f(d$test, prod_shift = d$prod_shift)
d$test$sc <- scale_f(d$test, prod_scale = d$prod_scale)
m7 <- as.mlt(Survreg(y | sc + 0 ~ sh, data = d$train, fixed = c("sh" = -1)))
logLik(m7, newdata = d$test, parm = coef(m7))
mylogLik(d)
coef(m7)

}

args <- data.frame(p = rep(c(5, 50), times = 3),
                   prod_shift_int = rep(c(1, 0, 1), each = 2),
                   prod_shift_trt = rep(c(1, 0, 1), each = 2),
                   prod_scale_int = rep(c(0, 1, 1), each = 2),
                   prod_scale_trt = rep(c(0, 1, 1), each = 2))

simdat <- vector(mode = "list", length = nrow(args))
loglik <- vector(mode = "list", length = nrow(args))

for (i in 1:nrow(args)) {

    simdat[[i]] <- replicate(nsim, dgp(ntrain = ntrain,
                                       ntest = ntest,
                                       p = args[i, "p"],
                                       prod_shift = c(args[i, "prod_shift_int"], args[i, "prod_shift_trt"]),
                                       prod_scale = c(args[i, "prod_scale_int"], args[i, "prod_scale_trt"])),
                             simplify = FALSE)
    loglik[[i]] <- lapply(simdat[[i]], mylogLik)
}

### remove h, hprime etc because we don't want to store these environments
simdat <- lapply(simdat, function(x) lapply(x, function(y) y[1:2]))

save(args, simdat, loglik, file='tr_simulated_data.rda')

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Load packages --------------------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

library("trtf")
library("party")
library("partykit")
library("tram")
library("ranger")
library("ordinalForest")

library("psych")
# https://personality-project.org/r/psych/
# https://personality-project.org/r/psych-manual.pdf


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Define global variables ----------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

set.seed(12345)

nsim <- c(100)
ntree <- c(250, 2000)

sim_para <- expand.grid(nsim = nsim, ntree = ntree)
rownames(sim_para) <- paste("nsim", sim_para$nsim, "_",
                            "ntree", sim_para$ntree, sep = "")

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Kullback-Leibler-Divergence ------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KL_div <- function(p,      # orig probability density function of DGP
                   y,      # orig y realisations of DGP test data set
                   q,      # estimated probability density function of random forest algorithm
                   yhat){  # predicted y observations based on random forest algorithm
  
  # Approach I: KL(p(y_k|x) || q(y_k|x))
  # estimated probability density function (with original y observations)
  q_y <- q[cbind(1:nrow(q), unclass(y))]
  KL_I <- mean( (-1) * log(q_y) )
  # library("MASS")
  # truehist((-1) * log(q_y))
  
  # Approach II: KL(q(\hat{y_k}|x) || p(\hat{y_k}|x))
  # model based predicted y observations
  p_yhat <- p[cbind(1:nrow(p), unclass(yhat))]
  KL_II <- mean( (-1) * log(p_yhat) )
  # truehist((-1) * log(p_yhat))
  
  # Approach III: KL(p(#|x) || q(#|x))
  # comparing whole probability density functions
  KL_III <- mean( rowSums(p * log(p / q)) )
  # i = 2
  # j = 3
  # p[i, j] * log(p[i, j] / q[i, j])
  
  return(list(KL_I = KL_I,
              KL_II = KL_II,
              KL_III = KL_III))
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal Forest - perffunction = "equal" ------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ordforest_equal <- function(train, test, test_prb, mtry) {
  
  rf <- ordfor(depvar = "y", data = train,
               nsets = 1000,       # number of considered score sets
               ntreeperdiv = 100,  # number of trees considered per tried score
               ntreefinal = ntree, # number of trees for the forest
               perffunction = "equal",
               min.node.size = max(minbucket, nrow(train) / 2^nodedepth), 
               mtry = mtry, 
               keep.inbag = TRUE)
  
  # model-based probability density function and y predictions
  rf_prb <- predict(rf, newdata = test)$classfreqtree
  rf_y <- factor(predict(rf, newdata = test)$ypred,
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  
  # log-likelihood
  prb <- rf_prb[cbind(1:nrow(test), unclass(test$y))]
  of_eq_ll <- sum(log(pmax(sqrt(.Machine$double.eps), prb)))
  
  # Kullback-Leibler Divergence
  of_eq_KL_div <- KL_div(p = test_prb, y = test$y,
                         q = rf_prb, yhat = rf_y)
  
  return(list(of_eq_ll = of_eq_ll,
              of_eq_KL_I = of_eq_KL_div$KL_I,
              of_eq_KL_II = of_eq_KL_div$KL_II,
              of_eq_KL_III = of_eq_KL_div$KL_III))
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal Forest - perffunction = "proportional" -----------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ordforest_prop <- function(train, test, test_prb, mtry) {
  
  rf <- ordfor(depvar = "y", data = train,
               nsets = 1000,       # number of considered score sets
               ntreeperdiv = 100,  # number of trees considered per tried score
               ntreefinal = ntree, # number of trees for the forest
               perffunction = "proportional",
               min.node.size = max(minbucket, nrow(train) / 2^nodedepth), 
               mtry = mtry, 
               keep.inbag = TRUE)
  
  # model-based probability density function and y predictions
  rf_prb <- predict(rf, newdata = test)$classfreqtree
  rf_y <- factor(predict(rf, newdata = test)$ypred,
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  
  # log-likelihood
  prb <- rf_prb[cbind(1:nrow(test), unclass(test$y))]
  of_prop_ll <- sum(log(pmax(sqrt(.Machine$double.eps), prb)))
  
  # Kullback-Leibler Divergence
  of_prop_KL_div <- KL_div(p = test_prb, y = test$y,
                           q = rf_prb, yhat = rf_y)
  
  return(list(of_prop_ll = of_prop_ll,
              of_prop_KL_I = of_prop_KL_div$KL_I,
              of_prop_KL_II = of_prop_KL_div$KL_II,
              of_prop_KL_III = of_prop_KL_div$KL_III))
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal transfromation forest with Bernstein basis and general score -------
# Bs(theta) --> non-proportional odds deviations
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

traforest_theta <- function(train, test, test_prb, mtry) {
  
  m0 <- as.mlt(Polr(y ~ 1, data = train))
  rf <- traforest(m0, 
                  formula = y ~ ., 
                  data = train,
                  ntree = ntree, 
                  trace = FALSE, 
                  mtry = mtry,
                  control = ctrl_partykit)
  
  # model-based probability density function and y predictions
  rf_prb <- do.call("rbind", lapply(predict(rf, newdata = test, type = "density"), c))
  rf_prb <- data.frame(rf_prb)
  colnames(rf_prb) <- levels(test$y)
  rf_y <- factor(predict.cforest(rf, newdata = test, type = "response"),
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  names(rf_y) <- NULL
  
  # log-likelihood with “nearest neighbour” forest weights prediction
  cf <- predict(rf, newdata = test, type = "coef")
  # Warning message:
  #   In model.frame.default(object$predictf, data = newdata,
  #   na.action = na.pass,  : variable 'y' is not a factor
  tf_theta_ll <- logLik(rf, newdata = test, coef = cf)
  
  # Kullback-Leibler Divergence
  tf_theta_KL_div <- KL_div(p = test_prb, y = test$y,
                            q = rf_prb, yhat = rf_y)
  
  return(list(tf_theta_ll = tf_theta_ll,
              tf_theta_KL_I = tf_theta_KL_div$KL_I,
              tf_theta_KL_II = tf_theta_KL_div$KL_II,
              tf_theta_KL_III = tf_theta_KL_div$KL_III))
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal transformation forest with Bernstein basis and log-rank score ------
# Bs(alpha) --> proportional odds deviations from the model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

traforest_alpha <- function(train, test, test_prb, mtry) {
  
  ### There is no intercept in Bernstein-basis, therefor to perform
  ### log-rank splitting constant variable is added to the data implicitly
  train$alpha <- 1
  test$alpha <- 1
  m0 <- as.mlt(Polr(y ~ alpha, data = train, fixed = c("alpha" = 0)))
  ### split wrt to intercept ("log-rank scores") only
  rf <- traforest(m0, 
                  formula = y | alpha ~ ., 
                  data = train,
                  parm = "alpha",
                  mltargs = list(fixed = c("alpha" = 0)),
                  ntree = ntree, 
                  trace = FALSE, 
                  mtry = mtry,
                  control = ctrl_partykit)
  
  # model-based probability density function and y predictions
  rf_prb <- do.call("rbind", lapply(predict(rf, newdata = test,
                                            mnewdata = data.frame(alpha = 1),
                                            type = "density"), c))
  rf_prb <- data.frame(rf_prb)
  colnames(rf_prb) <- levels(test$y)
  rf_y <- factor(predict.cforest(rf, newdata = test, type = "response")[,"y"],
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  
  # log-likelihood with “nearest neighbour” forest weights prediction
  cf <- predict(rf, newdata = test, type = "coef")
  tf_alpha_ll <- logLik(rf, newdata = test, coef = cf)
  
  # Kullback-Leibler Divergence
  tf_alpha_KL_div <- KL_div(p = test_prb, y = test$y,
                            q = rf_prb, yhat = rf_y)
  
  return(list(tf_alpha_ll = tf_alpha_ll,
              tf_alpha_KL_I = tf_alpha_KL_div$KL_I,
              tf_alpha_KL_II = tf_alpha_KL_div$KL_II,
              tf_alpha_KL_III = tf_alpha_KL_div$KL_III))
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Conditional Inference Trees --> Forests ------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

mycforest <- function(train, test, test_prb, mtry) {
  
  ctrl_party <- party:::cforest_unbiased(ntree = ntree, 
                                         maxdepth = nodedepth, 
                                         mtry = mtry,
                                         minbucket = minbucket)
  
  ### log-rank splitting
  rf <- party::cforest(y ~ ., data = train, 
                       control = ctrl_party)
  
  # model-based probability density function and y predictions
  rf_prb <- predict(rf, newdata = test, type = "prob")
  rf_prb <- do.call("rbind", rf_prb)
  rf_y <- factor(predict(rf, newdata = test, type = "response"),
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  
  # log-likelihood
  prb <- rf_prb[cbind(1:nrow(test), unclass(test$y))]
  cf_ll <- sum(log(pmax(sqrt(.Machine$double.eps), prb)))
  
  # Kullback-Leibler Divergence
  cf_KL_div <- KL_div(p = test_prb, y = test$y,
                      q = rf_prb, yhat = rf_y)
  
  return(list(cf_ll = cf_ll,
              cf_KL_I = cf_KL_div$KL_I,
              cf_KL_II = cf_KL_div$KL_II,
              cf_KL_III = cf_KL_div$KL_III))
  
}

## ----libraries ,message=FALSE, size='scriptsize', cache=TRUE, cache.lazy=FALSE----
## load ALS data
## The data is available to registered users of https://nctu.partners.org/ProACT
## Register, download the data, and run
## R> library("TH.data"); demo("PROACT");
load("./ALS/ALSsurvdata.rda")
## attach packages
library("survival")
library("tram")
library("party")
library("trtf")
## to make results reproducible
set.seed(290875)

## ----data_preprocessing, size='scriptsize', cache=TRUE, cache.lazy=FALSE----
## remove items with missing values
ALScc <- ALSsurvdata[complete.cases(ALSsurvdata),]
## convert time variable from days to years
ALScc$time_onset_treatment <- ALScc$time_onset_treatment / 365
## too few levels for Coxph and Survreg
levels(ALScc$race) <- c("Caucasian", rep("Other", 3))

## stratification control for predictive models
grps <- with(ALScc, interaction(cens, Riluzole))

## output as Surv object
ALScc$y <- with(ALScc, Surv(survival.time, cens))
ALScc$cens <- ALScc$survival.time <- NULL
dfprog <- dfpred <- ALScc
dfprog$Riluzole <- NULL
## transform Riluzole variable to 0-1 variable
dfpred$Riluzole <- as.numeric(dfpred$Riluzole == 'Yes')

## ----settings, size='scriptsize', cahce=TRUE, cache.lazy=FALSE-----------
## number of randomly preselected variables for splitting
## set equal to the square root of the total number of variables
mtry <- ceiling(sqrt(ncol(ALScc) - 2))

### joint parameters for all forest procecdures
## number of trees
ntree <-250
## maximal depth of trees
nodedepth <- 10
## number of obs in a terminal node
minbucket <- 20

### Bernstein approximation parameters
## order of polynomials
ord <- 5
## indicator of whether log-transformation is applied
## to the time variable before the approximation
logf <- TRUE
## the largest time values are not used for approximation
## to keep it stable
sup <- c(sqrt(.Machine$double.eps), quantile(ALScc$y[,1], .95))

## set to integer value if what to learn forests with
## parallel computation
NCORES <- NULL
## Warning: when non-null value is used for NCORES it may
## offen cause issues. Non-parallel learning is stable.

### parameters for party/partykit
## with testype = "Univ" and mincriterion = 0 test-splitting
## procedure is free from early stops caused by p-values
ctrl_partykit <- partykit:::ctree_control(teststat = "quad",
                                          testtype = "Univ",
                                          mincriterion = 0,
                                          minbucket = minbucket,
                                          maxdepth = nodedepth)

## time points at which survival function is estimated
stime <- seq(from = 0, to = max(ALScc$y[,1]), length.out = 100)

## ----prognostic_cox_weibull_models, cache=TRUE, cache.lazy=FALSE, size='scriptsize'----
method <- c("Cox", "Weibull_alpha", "Cox_alpha", 
            "W_alpha", "Bs_alpha", "W_theta", "Bs_theta")
method <- factor(method, levels = method, labels = method)
distr <- expand.grid(time = stime, method = method)

## Cox PH model with constant PH
coxconst <- Coxph(y ~ 1, dfprog, order = ord, log_first = logf, support = sup)
logLik(coxconst)
distr$surv[distr$method == "Cox"] <- predict(coxconst, 
                                             newdata = data.frame(1), 
                                             q = stime, 
                                             type = "survivor")
## Cox PH model with prognostic PH
coxph <- Coxph(y ~ ., dfprog, order = ord, log_first = logf, support = sup)
logLik(coxph)
distr$surv[distr$method == "Cox_alpha"] <- predict(coxph, 
                                                   newdata = dfprog[1,], 
                                                   q = stime, 
                                                   type = "survivor")
## Weibull model
srg <- Survreg(y ~ ., dfprog, order = ord)
logLik(srg)
distr$surv[distr$method == "Weibull_alpha"] <- predict(srg, 
                                                       newdata = dfprog[1,], 
                                                       q = stime, 
                                                       type = "survivor")

## ----Bs_theta, cache=TRUE, cache.lazy=FALSE,size='scriptsize'------------
tf_B  <- traforest(as.mlt(Coxph(y ~ 1, 
                                data = dfprog, 
                                order = ord, 
                                log_first = logf, 
                                sup = sup)),
                   formula = y ~ ., 
                   data = dfprog,
                   ntree = ntree, 
                   mtry = mtry,
                   control = ctrl_partykit, 
                   cores = NCORES)
logLik(tf_B, OOB = TRUE)

distr$surv[distr$method == "Bs_theta"] <- predict(tf_B, 
                                                  dfprog[1,], 
                                                  type = 'survivor', 
                                                  q = stime)[[1]]
## The weights of the learned forest are used for learning other
## prognostic forests
initw <- do.call("cbind", tf_B$weights)
storage.mode(initw) <- "double"

## ----W_theta, cache=TRUE, cache.lazy=FALSE,size='scriptsize'-------------
tf_W <- traforest(as.mlt(Survreg(y ~ 1, data = dfprog)),
                  formula = y ~ ., 
                  data = dfprog,
                  ntree = ntree, 
                  mtry = mtry,
                  control = ctrl_partykit, 
                  weights = initw,
                  cores = NCORES)
logLik(tf_W, OOB = TRUE)

distr$surv[distr$method == "W_theta"] <- predict(tf_W, 
                                                 dfprog[1,], 
                                                 type = 'survivor', 
                                                 q = stime)[[1]]

## ----W_alpha, cache=TRUE, cache.lazy=FALSE,size='scriptsize'-------------
tf_W_alpha <- traforest(as.mlt(Survreg(y ~ 1, data = dfprog)),
                        formula = y ~ ., 
                        data = dfprog, 
                        parm = "(Intercept)",
                        ntree = ntree, 
                        mtry = mtry,
                        control = ctrl_partykit, 
                        weights = initw,
                        cores = NCORES)
logLik(tf_W_alpha, OOB = TRUE)

distr$surv[distr$method == "W_alpha"] <- predict(tf_W_alpha, 
                                                 dfprog[1,], 
                                                 type = 'survivor', 
                                                 q = stime)[[1]]

## ----Bs_alpha, cache=TRUE, cache.lazy=FALSE,size='scriptsize'------------
dfprog$alpha <- 1
tf_B_alpha <- traforest(m <- as.mlt(Coxph(y ~ alpha, 
                                          data = dfprog, 
                                          order = ord, 
                                          log_first = logf,
                                          support = sup, 
                                          fixed = c("alpha" = 0))),
                        formula = y | alpha ~ ., 
                        data = dfprog,
                        parm = "alpha",
                        mltargs = list(fixed = c("alpha" = 0)),
                        ntree = ntree, 
                        mtry = mtry,
                        control = ctrl_partykit,
                        weights = initw,
                        cores = NCORES)
logLik(tf_B_alpha, OOB = TRUE)

## a hack to make predictions
cf1 <- predict(tf_B_alpha, dfprog[1,], type = 'coef')[[1]]
cf <- coef(m, fixed = TRUE)
cf[names(cf1)] <- cf1
coef(m) <- cf
distr$surv[distr$method == "Bs_alpha"] <- predict(m, 
                                                  newdata = data.frame(alpha = 0), 
                                                  type = 'survivor', 
                                                  q = stime)

## ----prognostic_plot, fig.width=8, fig.height=6, size='scriptsize', cache=TRUE, cache.lazy=FALSE----
library("lattice")
xyplot(surv ~ time, data = distr, group = method, type = "l", auto.key = TRUE)

## ----predictive_cox_weibull_models, cache=TRUE, cache.lazy=FALSE, size='scriptsize'----
## a formula that states time-independent part of log-hazard as a linear combination
## of variables, treatmenent and treatment-variable interactionsn
fm <- names(dfpred)
fm <- paste(fm[!(fm %in% c("y", "Riluzole"))], collapse = "+")
fmy <- as.formula(paste("y ~ Riluzole +", fm, " + Riluzole * (", fm, ")"))

## grid methods vs time points
method <- c("Cox_alpha_beta", "Weibull_alpha_beta", 
            "Bs_theta_theta", "W_theta_theta", 
            "Bs_theta_beta", "W_theta_beta",
            "Bs_alpha_beta", "W_alpha_beta")
method <- factor(method, levels = method, labels = method)
distr <- expand.grid(time = stime, Riluzole = 0:1, method = method)

## for the first patient from ALS dataset prediction is made in
## the abcence and in the presence of treatment by Riluzole
nd <- dfpred[c(1, 1),]
nd$Riluzole <- 0:1

## predictive Cox PH model
tcoxph <- Coxph(fmy , dfpred, order = ord, log_first = logf, support = sup)
logLik(tcoxph)
distr$surv[distr$method == "Cox_alpha_beta"] <-  c(predict(tcoxph,
                                                           newdata = nd, 
                                                           q = stime, 
                                                           type = "survivor"))

## predictive Weibull PH model
tweib <- Survreg(fmy, dfpred)
logLik(tweib)
distr$surv[distr$method == "Weibull_alpha_beta"] <- c(predict(tweib,
                                                              newdata = nd, 
                                                              q = stime, 
                                                              type = "survivor"))

## ----Bs_theta_theta, cache=TRUE, cache.lazy=FALSE, size='scriptsize'-----
Bs_theta_theta <- traforest(as.mlt(Coxph(y | Riluzole ~ 1, 
                                         data = dfpred, 
                                         order = ord,
                                         log_first = logf, 
                                         support = sup)),
                            formula = y | Riluzole ~ ., 
                            data = dfpred,
                            mtry = mtry, 
                            ntree = ntree,
                            strata = grps, 
                            control = ctrl_partykit,
                            cores = NCORES)
logLik(Bs_theta_theta, OOB = TRUE)
distr$surv[distr$method == "Bs_theta_theta"] <- c(predict(Bs_theta_theta,
    newdata = dfpred[1,], mnewdata = data.frame(Riluzole = 0:1),
    type = 'survivor', q = stime)[[1]])

initw <- do.call("cbind", Bs_theta_theta$weights)
storage.mode(initw) <- "double"

## ----W_theta_theta, cache=TRUE, cache.lazy=FALSE, size='scriptsize'------
W_theta_theta <- traforest(as.mlt(Survreg(y | Riluzole ~ 1, data = dfpred)),
                           formula = y | Riluzole ~ ., 
                           data = dfpred,
                           mtry = mtry, 
                           ntree = ntree,
                           strata = grps, 
                           control = ctrl_partykit,
                           weights = initw, 
                           cores = NCORES)
logLik(W_theta_theta, OOB = TRUE)
distr$surv[distr$method == "W_theta_theta"] <- c(predict(W_theta_theta,
                                                         newdata = dfpred[1,], 
                                                         mnewdata = data.frame(Riluzole = 0:1),
                                                         type = 'survivor', 
                                                         q = stime)[[1]])

## ----Bs_theta_beta, cache=TRUE, cache.lazy=FALSE, size='scriptsize'------
Bs_theta_beta <- traforest(as.mlt(Coxph(y ~ Riluzole, 
                                        data = dfpred, 
                                        order = ord,
                                        log_first = logf, 
                                        support = sup)),
                           formula = y | Riluzole ~ ., 
                           data = dfpred,
                           mtry = mtry, 
                           ntree = ntree,
                           strata = grps, 
                           control = ctrl_partykit,
                           weights = initw, 
                           cores = NCORES)
logLik(Bs_theta_beta, OOB = TRUE)
distr$surv[distr$method == "Bs_theta_beta"] <- c(predict(Bs_theta_beta,
                                                         newdata = dfpred[1,], 
                                                         mnewdata = data.frame(Riluzole = 0:1),
                                                         type = 'survivor', 
                                                         q = stime)[[1]])

## ----W_theta_beta, cache=TRUE, cache.lazy=FALSE, size='scriptsize'-------
W_theta_beta <- traforest(as.mlt(Survreg(y ~ Riluzole, data = dfpred)),
                          formula = y | Riluzole ~ ., 
                          data = dfpred,
                          mtry = mtry, 
                          ntree = ntree,
                          strata = grps, 
                          control = ctrl_partykit,
                          weights = initw, 
                          cores = NCORES)
logLik(W_theta_beta, OOB = TRUE)
distr$surv[distr$method == "W_theta_beta"] <- c(predict(W_theta_beta,
                                                        newdata = dfpred[1,], 
                                                        mnewdata = data.frame(Riluzole = 0:1),
                                                        type = 'survivor', 
                                                        q = stime)[[1]])

## ----W_alpha_beta, cache=TRUE, cache.lazy=FALSE, size='scriptsize'-------
W_alpha_beta <- traforest(as.mlt(Survreg(y ~ Riluzole, data = dfpred)),
                          formula = y | Riluzole ~ ., 
                          data = dfpred,
                          parm = c("(Intercept)", "Riluzole"),
                          mtry = mtry, 
                          ntree = ntree,
                          strata = grps, 
                          control = ctrl_partykit,
                          weights = initw, 
                          cores = NCORES)
logLik(W_alpha_beta, OOB = TRUE)
distr$surv[distr$method == "W_alpha_beta"] <- c(predict(W_alpha_beta,
                                                        newdata = dfpred[1,], 
                                                        mnewdata = data.frame(Riluzole = 0:1),
                                                        type = 'survivor', 
                                                        q = stime)[[1]])

## ----Bs_alpha_beta, cache=TRUE, cache.lazy=FALSE, size='scriptsize'------
dfpred$int <- 1
Bs_alpha_beta <- traforest(m <- as.mlt(Coxph(y ~ int + Riluzole, 
                                             data = dfpred, 
                                             fixed = c("int" = 0), 
                                             order = ord,
                                             log_first = logf, 
                                             support = sup)),
                           formula = y | int + Riluzole ~ .,
                           data = dfpred, 
                           parm = c("int", "Riluzole"),
                           mltargs = list(fixed = c("int" = 0)), 
                           mtry = mtry,
                           ntree = ntree, 
                           strata = grps,
                           control = ctrl_partykit, 
                           weigths = initw,
                           cores = NCORES)
logLik(Bs_alpha_beta, OOB = TRUE)
cf1 <- predict(Bs_alpha_beta, dfprog[1,], type = 'coef')[[1]]
cf <- coef(m, fixed = TRUE)
cf[names(cf1)] <- cf1
coef(m) <- cf
distr$surv[distr$method == "Bs_alpha_beta"] <- c(predict(m, 
                                                         newdata = data.frame(int = 0, 
                                                                              Riluzole = 0:1), 
                                                         type = 'survivor', 
                                                         q = stime))

## ----predictive_plot, fig.width = 8, fig.height = 6, size='scriptsize', cache=TRUE, cache.lazy=FALSE----
library("lattice")
xyplot(surv ~ time | method, data = distr, group = Riluzole, type = "l", auto.key = TRUE)

### Reproducibility material (Figure 2)
###
###     Predictive Distribution Modelling Using Transformation Forests
###     by Torsten Hothorn & Achim Zeileis
###     doi: 10.1080/10618600.2021.1872581
###


library("quantregForest")
set.seed(290875)

CORES <- 6L

pdf("QRF.pdf", width = 12, height = 8)

ntree <- 1000
n <- 10000
p <- 10
x <- runif(n)
y <- rnorm(n, sd = 1 + (x > .5))
X <- matrix(runif(n * p), ncol = p)

qrf <- quantregForest(y = y, x = data.frame(x = x, X), mtry = p + 1, 
                      nodesize = 25, ntree = ntree, replace = FALSE)

xn <- 1:50 / 51
nX <- matrix(.5, nrow = length(xn), ncol = ncol(X))
nd <- data.frame(x = xn, nX)
qQRF <- predict(qrf, newdata = nd, what = c(.1, .9))

library("trtf")
var_y <- numeric_var("y", support = c(-5, 5))
B_y <- Bernstein_basis(var_y, order = 5, ui = "increasing")
m_y <- ctm(B_y)
trf <- traforest(m_y, formula = y ~ ., data = data.frame(y = y, x = x, X), 
                 ntree = ntree,
                 control = ctree_control(mincriterion = 0, 
                     minsplit = 25, minbucket = 10), 
                 mtry = p + 1, trace = TRUE, cores = CORES)

trt <- trafotree(m_y, formula = y ~ ., data = data.frame(y = y, x = x, X))

qTRT <- predict(trt, newdata = nd, type = "quantile", prob = c(.1, .9))
qTRT <- t(qTRT)

qTRF <- list(p1 = predict(trf, newdata = nd, type = "quantile", prob = .1),
             p2 = predict(trf, newdata = nd, type = "quantile", prob = .9))

lwd <- 1.5
col <- rgb(.1, .1, .1, .05)
colR <- rgb(.75, 0, 0, .8)
colRlight <- rgb(.75, 0, 0, .1)
colB <- rgb(0, 0, 0.75, .8)

q1 <- qnorm(.1, sd = 1 + (x > .5))
q9 <- qnorm(.9, sd = 1 + (x > .5))
plot(y ~ x, pch = 19, col = c(col, colRlight)[(y < q1 | y > q9) + 1], cex = .5)
lines(xn, qQRF[,1], lty = 1, lwd = lwd * 1.5, col = "black", type = "S")
lines(xn, qQRF[,2], lty = 1, lwd = lwd * 1.5, col = "black", type = "S")
lines(xn, qTRT[,1], lty = 2, lwd = lwd * 1.5, col = colB, type = "S")
lines(xn, qTRT[,2], lty = 2, lwd = lwd * 1.5, col = colB, type = "S")
lines(xn, unlist(qTRF[[1]]), lty = 1, lwd = lwd * 1.5, col = colB, type = "S")
lines(xn, unlist(qTRF[[2]]), lty = 1, lwd = lwd * 1.5, col = colB, type = "S")
legend("topleft", lty = c(1, 2, 1), lwd = c(lwd * 1.5, lwd * 1.5, lwd * 1.5),
       col = c("black", colB, colB),
       legend = c("Quantile Regression Forest", "Transformation Tree", "Transformation Forest"),
       bty = "n")

dev.off()

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Load packages --------------------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

library("trtf")
library("party")
library("partykit")
library("tram")
library("ranger")
library("ordinalForest")

library("psych")
# https://personality-project.org/r/psych/
# https://personality-project.org/r/psych-manual.pdf

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Cohen's Kappa --------------------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ckappa <- function(yobs1, yobs2) {
  
  x <- data.frame(yobs1 = yobs1, yobs2 = yobs2)
  
  # unweighted kappa
  ck <- tryCatch(cohen.kappa(x)$kappa,
                 error = function(e) NA)
  
  # weighted kappa
  J <- length(unique(c(x$yobs1, x$yobs2)))
  
  # linear weighted kappa
  w_lin <- matrix(0, ncol = J, nrow = J)
  w_lin[] <- abs((col(w_lin) - row(w_lin)))
  w_lin <- 1 - w_lin/(J - 1)
  lk <- tryCatch(cohen.kappa(x, w = w_lin)$weighted.kappa,
                 error = function(e) NA)
  
  # quadratic weighted kappa
  w_quad <- matrix(0, ncol = J, nrow = J)
  w_quad[] <- abs((col(w_quad) - row(w_quad)))^2
  w_quad <- 1 - w_quad/(J - 1)^2
  qk <- tryCatch(cohen.kappa(x, w = w_quad)$weighted.kappa,
                 error = function(e) NA)
  
  return(list(ck = ck,
              lk = lk,
              qk = qk))
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal Forest - perffunction = "equal" ------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ordforest_equal <- function(train, test, mtry) {
  
  rf <- ordfor(depvar = "y", data = train,
               nsets = 1000,       # number of considered score sets
               ntreeperdiv = 100,  # number of trees considered per tried score
               ntreefinal = ntree, # number of trees for the forest
               perffunction = "equal",
               min.node.size = max(minbucket, nrow(train) / 2^nodedepth), 
               mtry = mtry, 
               keep.inbag = TRUE)
  
  # model-based probability density function and y predictions
  rf_prb <- predict(rf, newdata = test)$classfreqtree
  
  # log-likelihood
  prb <- rf_prb[cbind(1:nrow(test), unclass(test$y))]
  of_eq_ll <- sum(log(pmax(sqrt(.Machine$double.eps), prb)))
  
  return(list(of_eq_ll = of_eq_ll))
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal Forest - perffunction = "proportional" -----------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ordforest_prop <- function(train, test, mtry) {
  
  rf <- ordfor(depvar = "y", data = train,
               nsets = 1000,       # number of considered score sets
               ntreeperdiv = 100,  # number of trees considered per tried score
               ntreefinal = ntree, # number of trees for the forest
               perffunction = "proportional",
               min.node.size = max(minbucket, nrow(train) / 2^nodedepth), 
               mtry = mtry, 
               keep.inbag = TRUE)
  
  # model-based probability density function and y predictions
  rf_prb <- predict(rf, newdata = test)$classfreqtree
  rf_y <- factor(predict(rf, newdata = test)$ypred,
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  
  # log-likelihood
  prb <- rf_prb[cbind(1:nrow(test), unclass(test$y))]
  of_prop_ll <- sum(log(pmax(sqrt(.Machine$double.eps), prb)))
  
  return(list(of_prop_ll = of_prop_ll,
              of_prop_y = rf_y))  # later needed for Cohen's Kappa
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal transfromation forest with Bernstein basis and general score -------
# Bs(theta) --> non-proportional odds deviations
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

traforest_theta <- function(train, test, mtry) {
  
  m0 <- as.mlt(Polr(y ~ 1, data = train))
  rf <- traforest(m0, 
                  formula = y ~ ., 
                  data = train,
                  ntree = ntree, 
                  trace = FALSE, 
                  mtry = mtry,
                  control = ctrl_partykit)
  
  # model-based probability density function and y predictions
  rf_prb <- do.call("rbind", lapply(predict(rf, newdata = test, type = "density"), c))
  rf_prb <- data.frame(rf_prb)
  colnames(rf_prb) <- levels(test$y)
  rf_y <- factor(predict.cforest(rf, newdata = test, type = "response"),
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  names(rf_y) <- NULL
  # yprob <- do.call("rbind", lapply(predict(rf, newdata = test, type = "density"), t))
  # yprob <- data.frame(yprob)
  # colnames(yprob) <- levels(test$y)
  # yhat <- factor(levels(test$y)[apply(yprob, 1, which.max)],
  #                levels = levels(test$y), order = TRUE)
  # table(yhat)
  
  # log-likelihood with “nearest neighbour” forest weights prediction
  cf <- predict(rf, newdata = test, type = "coef")
  tf_theta_ll <- logLik(rf, newdata = test, coef = cf)
  
  return(list(tf_theta_ll = tf_theta_ll,
              tf_theta_y = rf_y)) # later needed for Cohen's Kappa
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ordinal transformation forest with Bernstein basis and log-rank score ------
# Bs(alpha) --> proportional odds deviations from the model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

traforest_alpha <- function(train, test, mtry) {
  
  ### There is no intercept in Bernstein-basis, therefor to perform
  ### log-rank splitting constant variable is added to the data implicitly
  train$alpha <- 1
  test$alpha <- 1
  m0 <- as.mlt(Polr(y ~ alpha, data = train, fixed = c("alpha" = 0)))
  ### split wrt to intercept ("log-rank scores") only
  rf <- traforest(m0, 
                  formula = y | alpha ~ ., 
                  data = train,
                  parm = "alpha",
                  mltargs = list(fixed = c("alpha" = 0)),
                  ntree = ntree, 
                  trace = FALSE, 
                  mtry = mtry,
                  control = ctrl_partykit)
  
  # model-based probability density function and y predictions
  rf_prb <- do.call("rbind", lapply(predict(rf, newdata = test,
                                            mnewdata = data.frame(alpha = 1),
                                            type = "density"), c))
  rf_prb <- data.frame(rf_prb)
  colnames(rf_prb) <- levels(test$y)
  rf_y <- factor(predict.cforest(rf, newdata = test, type = "response")[,"y"],
                 levels = levels(train$y),
                 labels = levels(train$y),
                 ordered = TRUE)
  
  # log-likelihood with “nearest neighbour” forest weights prediction
  cf <- predict(rf, newdata = test, type = "coef")
  tf_alpha_ll <- logLik(rf, newdata = test, coef = cf)
  
  return(list(tf_alpha_ll = tf_alpha_ll))
  
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Conditional Inference Trees --> Forests ------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

mycforest <- function(train, test, mtry) {
  
  ctrl_party <- party:::cforest_unbiased(ntree = ntree, 
                                         maxdepth = nodedepth, 
                                         mtry = mtry,
                                         minbucket = minbucket)
  
  ### log-rank splitting
  rf <- party::cforest(y ~ ., data = train, 
                       control = ctrl_party)
  
  # model-based probability density function and y predictions
  rf_prb <- predict(rf, newdata = test, type = "prob")
  rf_prb <- do.call("rbind", rf_prb)
  
  # log-likelihood
  prb <- rf_prb[cbind(1:nrow(test), unclass(test$y))]
  cf_ll <- sum(log(pmax(sqrt(.Machine$double.eps), prb)))
  
  return(list(cf_ll = cf_ll))
  
}

x <- readLines("03_ALS_analysis.R")

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Calculate individiual results.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

idx <- 1:100

for (i in idx){
    f <- paste("run_", i, ".R", sep = "")
    writeLines(gsub("INDEX", i, x), con = f)
    system(paste("R CMD BATCH", f, ifelse(i < max(idx), " & ", "")))
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
### Combine individiual results.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

out <- 0

for (i in idx) {
  f <- paste("ALS_results_combined_", i, ".rda", sep = "")
  if (file.exists(f)) {
    load(f)
    out <- out + ret
  } else {
    cat(f, "missing \n")
  }
}

ret <- out

ret[ret == 0] <- NA

save(ret, file = "ALS_results_combined.rda")

rm(list=ls())

source("competitors.R")

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Functions used for the data generating process -----------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

### Friedman function
f1 <- function(x1, x2, x3, x4, x5, scale = TRUE) {
  ret <- 10 * sin(pi * x1 * x2) + 20 * (x3 - 0.5)^2 + 10 * x4 + 5 * x5
  if (scale) {
    ret <- ret - min(ret) 
    ret <- ret / max(ret)
    ret <- 5 * ret - 2.5 # OR
  }
  ret
}

### model scale function
## if there is no scale in the model, prod_scale = 0
scale_f <- function(d, prod_scale = 1, ...){
  ### roughly exp(-1.5, 1.5)
  exp(with(d, f1(X1, X2, X3, X4, X5, scale = TRUE) / (2.5 / 1.5)) * prod_scale)}

### model shift function
## if there is no shift in the model, prod_shift = 0
shift_f <- function(d, prod_shift = 1, ...){
  with(d, f1(X6, X7, X8, X9, X10, scale = TRUE)) * prod_shift}

rmodel <- function(n, shift, scale) {
  theta <- matrix(logit(c(.15, .5, .85)), nrow = n, ncol = 3, byrow = TRUE)
  lp <- scale * theta + shift
  cdf <- plogis(lp)
  prb <- cbind(cdf[,1], cdf[,2] - cdf[,1], cdf[,3] - cdf[,2], 1 - cdf[,3])
  y <- apply(prb, 1, function(r) which(rmultinom(1, size = 1, prob = r) > 0))
  y <- factor(y, levels = 1:4, labels = 1:4, ordered = TRUE)
  return(list(y = y, prb = prb))
}

ry <- function(d, prod_shift, prod_scale, ...){
  rmodel(n = nrow(d),
         shift = shift_f(d, prod_shift = prod_shift),
         scale = scale_f(d, prod_scale = prod_scale))
}

### data generation
dgp <- function(ntrain, ntest,
                p = 5,
                prod_shift = 0,
                prod_scale = 0, ...) {
  n <- ntrain + ntest
  dat <- data.frame(matrix(runif(10 * n), ncol = 10),
                  matrix(runif(n * p, min = 0, max = 1), nrow = n))
  ry_tmp <- ry(dat, prod_shift = prod_shift,
               prod_scale = prod_scale)
  dat$y <- ry_tmp$y
  prb <- ry_tmp$prb
  return(list('train' = dat[1:ntrain,], 'test' = dat[(ntrain+1):n,],
              'test_prb' = prb[(ntrain+1):n,],
              prod_shift = prod_shift, prod_scale = prod_scale))
}

### log-likelihood of the model
mylogLik <- function(d){
  shift <- shift_f(d$test, prod_shift = d$prod_shift)
  scale <- scale_f(d$test, prod_scale = d$prod_scale)
  FZ <- mlt:::.Logistic()
  # theta von rmodel(...)
  h <- c(-Inf, FZ$q(c(.15, .5, .85)), Inf)
  hu <- h[unclass(d$test$y) + 1]
  hl <- h[unclass(d$test$y)]
  sum(log(FZ$p(hu*scale + shift) - FZ$p(hl*scale + shift) ))
}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Data generating process -------------------------------------------------
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

set.seed(12345)

if(!dir.exists("rda")){dir.create("rda")}

for(nsim_i in seq_along(unique(sim_para[, "nsim"]))){
  
  # nsim_i=1
  # nsim_i=2
  
  ## number of repetitions
  nsim <- unique(sim_para[, "nsim"])[nsim_i]
  
  ## size of training sample
  ntrain <- 250
  
  ## size of validation sample
  ntest <- 500
  
  ## scale and shape parameters of Weibull distribution
  wscale <- 1
  wshape <- 1
  
  # data simulation process -------------------------------------------------
  
  args <- expand.grid(p = c(5, 50), 
                      prod_shift = c(0, 1), 
                      prod_scale = c(0, 1))
  
  simdat <- vector(mode = "list", length = nrow(args))
  loglik <- vector(mode = "list", length = nrow(args))
  
  for (d in 1:nrow(args)) {
    
    # d = 8
    
    simdat[[d]] <- replicate(unique(sim_para[, "nsim"])[nsim_i],
                             dgp(ntrain = ntrain,
                                 ntest = ntest,
                                 p = args[d, "p"],
                                 prod_shift = args[d, "prod_shift"],
                                 prod_scale = args[d, "prod_scale"]),
                             simplify = FALSE)
    loglik[[d]] <- lapply(simdat[[d]], mylogLik)
  }
  
  save(args, simdat, loglik,
       file = paste("rda/simulated_data_nsim",
                    unique(sim_para[, "nsim"])[nsim_i],
                    ".rda", sep = ""))
  
}

rm(list=ls())

set.seed(12345)

start_time <- Sys.time()

library("parallel")
RNGkind("L'Ecuyer-CMRG")

ncores <- detectCores()

source("competitors.R")

if(!dir.exists("rda")){dir.create("rda")}

for(sp in seq_len(nrow(sim_para))){          # sp = 1

  load(paste("rda/simulated_data_nsim", sim_para[sp, "nsim"], ".rda", sep = ""))
  
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Define global parameters ------------------------------------------------
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
  ### joint parameters for all forest procedures
  
  ### number of repetitions
  nsim <- sim_para[sp, "nsim"]
  
  ## number of trees
  ntree <- sim_para[sp, "ntree"]
  
  ## maximal depth of trees
  nodedepth <- 10
  
  ## number of obs in terminal node
  minbucket <- 20
  
  ### parameters for party/partykit
  ctrl_partykit <- partykit:::ctree_control(teststat = "quad", 
                                            testtype = "Univ", 
                                            mincriterion = 0,
                                            minbucket = minbucket,
                                            maxdepth = nodedepth)
  
  res <- args[rep(1:nrow(args), each = sim_para[sp, "nsim"]),]
  simdat <- do.call("c", simdat)
  
  # true log-likelihood
  res$loglik <- unlist(do.call("c", loglik))
  
  # Ordinal Forests
  res$of_eq_ll <- res$of_eq_KL_I <- res$of_eq_KL_II <- res$of_eq_KL_III <- 0
  res$of_prop_ll <- res$of_prop_KL_I <- res$of_prop_KL_II <- res$of_prop_KL_III <- 0
  # Ordinal Transformation Forests
  res$tf_theta_ll <- res$tf_theta_KL_I <- res$tf_theta_KL_II <- res$tf_theta_KL_III <- 0
  res$tf_alpha_ll <- res$tf_alpha_KL_I <- res$tf_alpha_KL_II <- res$tf_alpha_KL_III <- 0
  res$cf_ll <- res$cf_KL_I <- res$cf_KL_II <- res$cf_KL_III <- 0
  
  # filename to save estimated likelihoods
  filename <- paste("rda/results_empeval_", rownames(sim_para)[sp], ".rda", sep = "")
  
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Ordinal Forest (equal) --------------------------------------------------
  # ### perffunction = "equal"
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
  of_equal_fun <- function(d) {
    mtry <- ifelse(res[d, "p"] <= 5, 
                   ncol(simdat[[d]]$train) - 1, 
                   ceiling(sqrt(ncol(simdat[[d]]$train) - 1)))
    rf <- ordforest_equal(train = simdat[[d]]$train,
                          test = simdat[[d]]$test,
                          test_prb = simdat[[d]]$test_prb,
                          mtry = mtry)
    return(list(of_eq_ll = rf$of_eq_ll,
                of_eq_KL_I = rf$of_eq_KL_I,
                of_eq_KL_II = rf$of_eq_KL_II,
                of_eq_KL_III = rf$of_eq_KL_III))
  }
  
  tmp <- runif(10)
  
  ### train and test for each repetition of the simulated data
  of_eq_all <- mclapply(1:length(simdat), of_equal_fun, mc.cores = ncores)
  ### summarize and results
  of_eq_colnam <- c("of_eq_ll", "of_eq_KL_I", "of_eq_KL_II", "of_eq_KL_III")
  for(i in seq_along(of_eq_colnam)){
    res[, of_eq_colnam[i]] <- as.vector( do.call("rbind", lapply(of_eq_all, function(l) l[[of_eq_colnam[i]]])) )
  }
  print("ordinalForest (equal) is done!")
  save(res, file = filename)
  
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Ordinal Forest (proportional) -------------------------------------------
  # ### perffunction = "proportional"
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
  of_prop_fun <- function(d) {
    mtry <- ifelse(res[d, "p"] <= 5, 
                   ncol(simdat[[d]]$train) - 1, 
                   ceiling(sqrt(ncol(simdat[[d]]$train) - 1)))
    rf <- ordforest_prop(train = simdat[[d]]$train,
                         test = simdat[[d]]$test,
                         test_prb = simdat[[d]]$test_prb,
                         mtry = mtry)
    return(list(of_prop_ll = rf$of_prop_ll,
                of_prop_y = rf$of_prop_y,
                of_prop_KL_I = rf$of_prop_KL_I,
                of_prop_KL_II = rf$of_prop_KL_II,
                of_prop_KL_III = rf$of_prop_KL_III))
  }
  
  tmp <- runif(10)
  
  ### train and test for each repetition of the simulated data
  of_prop_all <- mclapply(1:length(simdat), of_prop_fun, mc.cores = ncores)
  ### summarize and results
  of_prop_colnam <- c("of_prop_ll", "of_prop_KL_I", "of_prop_KL_II", "of_prop_KL_III")
  for(i in seq_along(of_prop_colnam)){
    res[, of_prop_colnam[i]] <- as.vector( do.call("rbind", lapply(of_prop_all, function(l) l[[of_prop_colnam[i]]])) )
  }
  print("ordinalForest (proportional) is done!")
  save(res, file = filename)
  
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Ordinal transformation forest with Bernstein basis and general score --------
  # Bs(theta)
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
  trtf_theta_fun <- function(d) {
    ## number of randomly preselected variables to perform a split
    ## for low-dimensional data it is set to the number of variables
    ## for high-dimensional data it is set to the square root of the number of variables
    mtry <- ifelse(res[d, "p"] <= 5,
                   ncol(simdat[[d]]$train) - 1,
                   ceiling(sqrt(ncol(simdat[[d]]$train) - 1)))
    rf <- traforest_theta(train = simdat[[d]]$train,
                          test = simdat[[d]]$test,
                          test_prb = simdat[[d]]$test_prb,
                          mtry = mtry)
    return(list(tf_theta_ll = rf$tf_theta_ll,
                tf_theta_KL_I = rf$tf_theta_KL_I,
                tf_theta_KL_II = rf$tf_theta_KL_II,
                tf_theta_KL_III = rf$tf_theta_KL_III))
  }
  
  tmp <- runif(10)
  
  ### train and test for each repetition of the simulated data
  tf_theta_all <- mclapply(1:length(simdat), trtf_theta_fun, mc.cores = ncores)
  ### summarize and results
  tf_theta_colnam <- c("tf_theta_ll", "tf_theta_KL_I", "tf_theta_KL_II", "tf_theta_KL_III")
  for(i in seq_along(tf_theta_colnam)){
    res[, tf_theta_colnam[i]] <- as.vector( do.call("rbind", lapply(tf_theta_all, function(l) l[[tf_theta_colnam[i]]])) )
  }
  print("trtf_theta is done!")
  save(res, file = filename)
  
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Ordinal transformation forest with Bernstein basis and log-rank score ----------
  # Bs(alpha)
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
  trtf_alpha_fun <- function(d) {
    mtry <- ifelse(res[d, "p"] <= 5,
                   ncol(simdat[[d]]$train) - 1,
                   ceiling(sqrt(ncol(simdat[[d]]$train) - 1)))
    rf <- traforest_alpha(train = simdat[[d]]$train,
                          test = simdat[[d]]$test,
                          test_prb = simdat[[d]]$test_prb,
                          mtry = mtry)
    return(list(tf_alpha_ll = rf$tf_alpha_ll,
                tf_alpha_KL_I = rf$tf_alpha_KL_I,
                tf_alpha_KL_II = rf$tf_alpha_KL_II,
                tf_alpha_KL_III = rf$tf_alpha_KL_III))
  }
  
  tmp <- runif(10)
  
  ### train and test for each repetition of the simulated data
  tf_alpha_all <- mclapply(1:length(simdat), trtf_alpha_fun, mc.cores = ncores)
  ### summarize and results
  tf_alpha_colnam <- c("tf_alpha_ll", "tf_alpha_KL_I", "tf_alpha_KL_II", "tf_alpha_KL_III")
  for(i in seq_along(tf_alpha_colnam)){
    res[, tf_alpha_colnam[i]] <- as.vector( do.call("rbind", lapply(tf_alpha_all, function(l) l[[tf_alpha_colnam[i]]])) )
  }
  print("trtf_alpha is done!")
  save(res, file = filename)
  
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Conditional Inference Forests -------------------------------------------
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
  cf_fun <- function(d){
    mtry <- ifelse(res[d, "p"] <= 5,
                   ncol(simdat[[d]]$train) - 1,
                   ceiling(sqrt(ncol(simdat[[d]]$train) - 1)))
    rf <- mycforest(train = simdat[[d]]$train,
                    test = simdat[[d]]$test,
                    test_prb = simdat[[d]]$test_prb,
                    mtry = mtry)
    return(list(cf_ll = rf$cf_ll,
                cf_KL_I = rf$cf_KL_I,
                cf_KL_II = rf$cf_KL_II,
                cf_KL_III = rf$cf_KL_III))
  }
  
  tmp <- runif(10)
  
  ### train and test for each repetition of the simulated data
  cf_all <- mclapply(1:length(simdat), cf_fun, mc.cores = ncores)
  ### summarize and results
  cf_colnam <- c("cf_ll", "cf_KL_I", "cf_KL_II", "cf_KL_III")
  for(i in seq_along(cf_colnam)){
    res[, cf_colnam[i]] <- as.vector( do.call("rbind", lapply(cf_all, function(l) l[[cf_colnam[i]]])) )
  }
  print("mycforest is done!")
  save(res, file = filename)
  
  ### delete no longer used variables
  rm(list = c("args", "simdat", "loglik",
              "of_eq_all", "of_prop_all",
              "tf_theta_all", "tf_alpha_all", "cf_all"))

}

end_time <- Sys.time()
print(end_time - start_time)

sessionInfo()

### run simulations
if (!file.exists("2d.rda"))
    source("2d.R")
if (!file.exists("lognormal_2d.rda"))
    source("lognormal_2d.R")
if (!file.exists("friedman.rda"))
    source("friedman.R")
if (!file.exists("lognormal_friedman.rda"))
    source("lognormal_friedman.R")
if (!file.exists("timings.rda"))
    source("timings.R")

### generate plots
pdf("plots.pdf", width = 6, height = 10)
source("plots.R")
dev.off()

library("lattice")
library("grid")
lwd <- 1.5
col <- rgb(.1, .1, .1, .05)
colR <- rgb(.75, 0, 0, .8)
colRlight <- rgb(.75, 0, 0, .1)
colB <- rgb(0, 0, 0.75, .8)
trellis.par.set(list(plot.symbol = list(col="black",pch=18, cex=0.75),
                     box.rectangle = list(col=1),
                     box.umbrella = list(lty=1, col=1),
                     strip.background = list(col = "white")))
ltheme <- canonical.theme(color = FALSE)     ## in-built B&W theme
ltheme$strip.background$col <- "transparent" ## change strip bg
lattice.options(default.theme = ltheme)

fn <- function(n) formatC(n, big.mark = "{,}")

plotfun <- function(file, models = c("gt", "tt"), labels = models,
                    wplot = c("ll", "L.1", "L.5", "L.9"), 
                    hypo = c("H2", "H2", "H2", "H0"), main = file) {

  wplot <- match.arg(wplot, several.ok = TRUE)

  load(file)

  df0 <- subset(df, tau < 0.1 & model %in% models)
  df0$model <- df0$model[,drop = TRUE]
  df0$model <- factor(as.character(df0$model), levels = models, labels = labels)

  df0$p <- factor(df0$p, levels = sort(unique(df0$p)),
                  labels = c("Low dimensional", "High dimensional"))
  df0$prod_mu <- factor(df0$prod_mu, levels = 0:1, labels = c("Constant mean", "Varying mean"))
  df0$prod_mu <- factor(as.character(df0$prod_mu), 
                        levels = rev(levels(df0$prod_mu)), 
                        labels = rev(levels(df0$prod_mu)))
  df0$prod_sigma <- factor(df0$prod_sigma, levels = 0:1, 
                           labels = c("Constant variance", "Varying variance"))
  df0$prod_sigma <- factor(as.character(df0$prod_sigma), 
                           levels = rev(levels(df0$prod_sigma)), 
                           labels = rev(levels(df0$prod_sigma)))

  df0gt <- subset(df0, model == labels[models == "gt"])
  nm <- names(df0gt)
  nm[match(c("ll", "L.1", "L.5", "L.9"), nm)] <- paste0("gt", c("ll", "L.1", "L.5", "L.9"))
  names(df0gt) <- nm
  df0gt$model <- NULL
  df0 <- merge(df0, df0gt, by = c("id", "p", "prod_mu", "prod_sigma"))
  df0 <- subset(df0, model != labels[models == "gt"])
  df0$model <- df0$model[, drop = TRUE]
  df0$rightlabel <- with(df0, interaction(prod_mu, prod_sigma))
  levels(df0$rightlabel) <- hypo

  mypanel <- function(x, y, groups, subscripts, ...) {
    lim <- current.panel.limits(unit = "native")
    at <- mean(ylim <- lim$ylim)
    lrect(lim$xlim[1], lim$ylim[1], lim$xlim[2], lim$ylim[2], col = "white")
    panel.bwplot(x = x, y = y, ...)
#    panel.abline(h = median(y[x == "Truth"]), lty = 3)
    panel.abline(h = 0, lty = 3)
    lab <- as.character(unique(df0$rightlabel[subscripts]))
    if (unique(df0$p[subscripts]) == "High dimensional")
        panel.axis("right", at = at, labels = lab, 
                   ticks = FALSE, outside = TRUE)
    tapply(1:length(y), groups[subscripts], function(i) {
        llines(x = 1:nlevels(x), y = y[i][order(x[i])], 
               col = rgb(0.1, 0.1, 0.1, 0.1))
    })
    lrect(lim$xlim[1], lim$ylim[2], lim$xlim[2], lim$ylim[2] * 100, 
          col = "white", border = "white")
  }

  myprepanel <- function(df, response) {
      df$response <- response
      function(x, y, groups, subscripts, ...) {
          ps <- df[subscripts, "prod_sigma"]
          pm <- df[subscripts, "prod_mu"]
          tdf <- subset(df, prod_sigma == ps & prod_mu == pm)
          bp <- boxplot(split(tdf[["response"]], tdf$model), plot = FALSE)
          ylim <- range(bp$stats) * c(1, 1.2)
#          ylim <- pmin(500, range(bp$stats) * c(.995, 1))
          if (ylim[1] > 0) ylim[1] <- 0
          list(ylim = ylim) 
      }
  }

### we need clipping off for outside panel.axis but this
### draws outside the box. 
ps <-  list(clip = list(panel = "off", strip = "on"),
                    layout.widths = list(right.padding = 5))
#                    layout.heights = list(top.padding = 0,
#                        main = 0, main.key.padding = 0, key.top = 0))

sc <- function(which.given, ..., factor.levels) {
    if (which.given == 1) {
        strip.default(which.given = which.given, ..., factor.levels)
    } else {
        if (which.given == 2) {
            strip.default(which.given = 2, ..., 
               factor.levels = c(expression(paste("Varying ", mu)), expression(paste("Constant ", mu))))
        } else {
            strip.default(which.given = 3, ..., 
               factor.levels = c(expression(paste("Varying ", sigma)), expression(paste("Constant ", sigma))))
        }
   }
}

  if ("ll" %in% wplot)
  print(bwplot(I(- ll + gtll) ~ model | p + prod_mu + prod_sigma, data = df0, 
        panel = mypanel, groups = id, do.out = FALSE,
        prepanel = myprepanel(df0, with(df0, - ll + gtll)),
        par.settings = ps, strip = sc,
        page = function(...) grid.text(main, x = .5, y = .980, 
                                       gp = gpar(cex = 1.25)),
        ylab = "Negative Log-likelihood difference", scales = list(y =
        "free", x = list(rot = 45)), layout = c(2, 4)))

  if ("L.1" %in% wplot)
  print(bwplot(L.1 - gtL.1 ~ model | p + prod_mu + prod_sigma, data = df0, 
        panel = mypanel, groups = id, do.out = FALSE,
        prepanel = myprepanel(df0, with(df0, L.1 - gtL.1)),
        par.settings = ps, strip = sc,
        page = function(...) grid.text(main, x = .5, y = .975, 
                                       gp = gpar(cex = 1.25)),
        ylab = "10% quantile risk difference", scales = list(y =
        "free", x = list(rot = 45)), layout = c(2, 4)))

  if ("L.5" %in% wplot)
  print(bwplot(L.5 - gtL.5 ~ model | p + prod_mu + prod_sigma, data = df0, 
        panel = mypanel, groups = id, do.out = FALSE,
        prepanel = myprepanel(df0, with(df0, L.5 - gtL.5)),
        par.settings = ps, strip = sc,
        page = function(...) grid.text(main, x = .5, y = .975, 
                                       gp = gpar(cex = 1.25)),
        ylab = "Absolute error difference", scales = list(y =
        "free", x = list(rot = 45)), layout = c(2, 4)))

  if ("L.9" %in% wplot)
  print(bwplot(L.9 - gtL.9 ~ model | p + prod_mu + prod_sigma, data = df0, 
        panel = mypanel, groups = id, do.out = FALSE,
        prepanel = myprepanel(df0, with(df0, L.9 - gtL.9)),
        par.settings = ps, strip = sc,
        page = function(...) grid.text(main, x = .5, y = .975, 
                                       gp = gpar(cex = 1.25)),
        ylab = "90% quantile risk difference", scales = list(y =
        "free", x = list(rot = 45)), layout = c(2, 4)))
}

forests <- c("gt", "tfbagg", "tfrf", "tfbaggBern", "tfrfBern", "cfbagg", "cfrf", "bagg", "rf")
trees <- c("gt", "ct", "tt", "ttExSplit", "ttM", "ttBern", "ttBernExSplit")
models <- c("gt", "ct", "tt", "ttBern", # "tfbagg", 
            "cfrf", 
            "rf", "rfBern",
            "tfrf", # "tfbaggBern", 
            "tfrfBern") # "cfbagg", 
            # "bagg", 
labels <- c("Truth", "CTree (P = 2)", "TTree (P = 2)", "TTree (P = 6)", 
            #"TBagging (lin)", 
            "CForest (P = 2)", 
            #"Bagging", 
            "RForest (P = 2)", "RForest (P = 6)",
            "TForest (P = 2)", 
            #"TBagging (nonlin)", 
            "TForest (P = 6)") 
            # "CBagging", 
logmodels <- c(
            "gt", 
            #"ct", 
            # "tt", 
            "ttBern", # "tfbagg", 
            "rfBern",
            # "tfrf", # "tfbaggBern", 
            "tfrfBern") # "cfbagg", 
            # "cfrf", 
            # "bagg", 
            #"rf", 
loglabels <- c(
            "Truth", 
            #"CTree (P = 2)", "TTree (P = 2)", 
            "TTree (P = 6)", 
            #"TBagging (lin)", 
            #"TForest (P = 2)", 
            #"TBagging (nonlin)", 
            "RForest (P = 6)",

            "TForest (P = 6)") 
            # "CBagging", 
            #"CForest (P = 2)", 
            #"Bagging", 
            #"RForest (P = 2)", 



### R code from vignette source '/home/thothorn/transforest/paper/empeval.Rnw'

###################################################
### code chunk number 2: 2d-ll
###################################################
plotfun("2d.rda_out.rda", models = models, labels = labels, wplot = "ll",
        main = "Tree-Structured Normal (H1a)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 3: ln-2d-ll
###################################################
plotfun("lognormal_2d.rda_out.rda", models = logmodels, labels = loglabels, wplot = "ll",
        main = "Tree-Structured Log-Normal (H1a)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 4: fm-ll
###################################################
plotfun("friedman.rda_out.rda", models = models, labels = labels, wplot = "ll",
        main = "Non-Linear Normal (H1b)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 5: ln-fm-ll
###################################################
plotfun("lognormal_friedman.rda_out.rda", models = logmodels, labels = loglabels, wplot = "ll",
         main = "Non-Linear Log-Normal (H1b)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 2: 2d-L1
###################################################
plotfun("2d.rda_out.rda", models = models, labels = labels, wplot = "L.1",
        main = "Tree-Structured Normal (H1a)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 3: 2d-L5
###################################################
plotfun("2d.rda_out.rda", models = models, labels = labels, wplot = "L.5",
        main = "Tree-Structured Normal (H1a)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 4: 2d-L9
###################################################
plotfun("2d.rda_out.rda", models = models, labels = labels, wplot = "L.9",
        main = "Tree-Structured Normal (H1a)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 5: ln-2d-L1
###################################################
plotfun("lognormal_2d.rda_out.rda", models = logmodels, labels = loglabels, wplot = "L.1",
        main = "Tree-Structured Log-Normal (H1a)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 6: ln-2d-L5
###################################################
plotfun("lognormal_2d.rda_out.rda", models = logmodels, labels = loglabels, wplot = "L.5",
        main = "Tree-Structured Log-Normal (H1a)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 7: ln-2d-L9
###################################################
plotfun("lognormal_2d.rda_out.rda", models = logmodels, labels = loglabels, wplot = "L.9",
        main = "Tree-Structured Log-Normal (H1a)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 8: fm-L1
###################################################
plotfun("friedman.rda_out.rda", models = models, labels = labels, wplot = "L.1",
        main = "Non-Linear Normal (H1b)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 9: fm-L5
###################################################
plotfun("friedman.rda_out.rda", models = models, labels = labels, wplot = "L.5",
        main = "Non-Linear Normal (H1b)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 10: fm-L9
###################################################
plotfun("friedman.rda_out.rda", models = models, labels = labels, wplot = "L.9",
        main = "Non-Linear Normal (H1b)", hypo = c("H2c", "H2c", "H2b", "H2a"))


###################################################
### code chunk number 11: ln-fm-L1
###################################################
plotfun("lognormal_friedman.rda_out.rda", models = logmodels, labels = loglabels, wplot = "L.1",
         main = "Non-Linear Log-Normal (H1b)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 12: ln-fm-L5
###################################################
plotfun("lognormal_friedman.rda_out.rda", models = logmodels, labels = loglabels, wplot = "L.5",
         main = "Non-Linear Log-Normal (H1b)", hypo = c("H2d", "H2d", "H2b", "H2a"))


###################################################
### code chunk number 13: ln-fm-L9
###################################################
plotfun("lognormal_friedman.rda_out.rda", models = logmodels, labels = loglabels, wplot = "L.9",
         main = "Non-Linear Log-Normal (H1b)", hypo = c("H2d", "H2d", "H2b", "H2a"))

### same as friedman.R

library("trtf")

nsim <- 10
ntree <- 1000

f1 <- function(x1, x2, x3, x4, x5, scale = TRUE) {
    ret <- 10 * sin(pi * x1 * x2) + 20 * (x3 - 0.5)^2 + 10 * x4 + 5 * x5
    if (scale) {
        ret <- ret - min(ret) 
        ret <- ret / max(ret)
        ret <- 3 * ret - 1.5
    }
    ret
}

sigma <- function(d, prod_sigma = 1, ...)
    exp(with(d, f1(X1, X2, X3, X4, X5, scale = TRUE)) * prod_sigma)

mu <- function(d, prod_mu = 1, ...)
    with(d, f1(X6, X7, X8, X9, X10, scale = TRUE)) * prod_mu

med <- function(d, ...)
    mu(d, ...)

qu <- function(d, prob, prod_mu, prod_sigma, ...)
    do.call("cbind", lapply(prob, function(p) qnorm(p, mean = mu(d, prod_mu = prod_mu), 
                                                    sd = sigma(d, prod_sigma = prod_sigma))))

ry <- function(d, prod_mu, prod_sigma, ...)
    rnorm(nrow(d), mean = mu(d, prod_mu = prod_mu), sd = sigma(d, prod_sigma = prod_sigma))

dgp <- function(n = 250, p = 5, ...) {
    d <- data.frame(matrix(runif(10 * n), ncol = 10),
                    matrix(runif(n * p, min = 0, max = 1), nrow = n))
    d$y <- ry(d, ...)
    d	
}


loglik <- function(d, p, prod_mu, prod_sigma, ...)
    sum(dnorm(d$y, mean = mu(d, prod_mu = prod_mu), sd = sigma(d, prod_sigma = prod_sigma), log = TRUE))

parm <- expand.grid(nsim = nsim, n = c(250, 500, 1000, 2000, 4000), 
                    ntest = 10, p = 5, tau = 0, prod_sigma = 1, prod_mu = 1)
parm <- lapply(1:nrow(parm), function(i) parm[i,])

source("artificial_simfun.R")

ctrl <- ctree_control(teststat = "quad", testtype = "Univ", 
                      mincriterion = 0, saveinfo = FALSE, 
                      minsplit = 25, minbucket = 10)

mc.reset.stream()
sf <- function(...) simfun(..., methods = c("ct", 
                                            "tt", 
                                            "ttExSplit",
                                            "ttM", 
                                            "ttBern", 
                                            "ttBernExSplit",
                                            "tfbagg", 
                                            "tfrf", 
#                                            "tfrfExSplit" 
                                            "tfbaggBern", 
                                            "tfrfBern", 
                                            "cfbagg", 
                                            "cfrf", 
                                            "bagg", 
                                            "rf"
))

parm

out <- mclapply(parm, function(x) do.call("sf", x), mc.cores = length(parm))

save(out, file = "timings.rda")

load('results_empeval.rda')

for (name in names(ret)){
    ret[, name] <- as.numeric(ret[, name])
}

meth <- c('rsf', 'rg', 'cf', 'tf_W_alpha', 'tf_B_alpha', 'tf_W', 'tf_B', 'L1')
tick_meth <- c('RSF NP(\U0001D6FC)', 
               'Ranger NP(\U0001D6FC)', 
               'CForest NP(\U0001D6FC)', 
               'TF W(\U0001D6FC)', 
               'TF Bs(\U0001D6FC)', 
               'TF W(\U0001D6DD)',
               'TF Bs(\U0001D6DD)',
               'L1') 

### expression(W(bold(vartheta)))


args <- unique(ret[, c('p', 'prod_shift', 'prod_scale')])

LogLikDiff <- ret[, meth] - ret$loglik


cairo_pdf(file = 'empeval_boxplot.pdf')
par(mfrow=c(4,2), oma = c(10, 6, 4, 8),
    mar=c(0, 0, 0, 0))


for (i in 1:nrow(args)){
  arg <- args[i,]
  idx <- (ret$prod_shift == arg$prod_shift) & (ret$prod_scale == arg$prod_scale) & (ret$p == arg$p)
  boxplot(LogLikDiff[idx, ], las=2, axes=FALSE, frame.plot=TRUE, ylim=c(-320, 15), at = c(1:3, 4.5, 5.5, 7, 8, 9.5))
  if (i > 6) {
    axis(1, col = "grey40", col.axis = "grey20", at = c(1:3, 4.5, 5.5, 7, 8, 9.5), labels=tick_meth, las = 2, cex.axis = 1.5)
  }
  if (i %in% c(1, 3, 5, 7)){
    axis(2, col = "grey40", col.axis = "grey20", at=c(-300, -250, -200, -150, -100, -50, 0))
  }
  if (i == 1){ mtext("low", side=3, line=1, cex = 1.2)}
  if (i == 2) {
    mtext('high', side=3, line=1, cex = 1.2)
    mtext('No', side=4, line=1, las=2, cex = 1.2)}
  if (i == 4) {
    mtext('PH', side=4, line=1, las=2, cex = 1.2)
  }
  if (i == 6){
    mtext('Non-PH', side=4, line=1, las=2, cex = 1.2)
  }
  if(i == 8){
    mtext('Combined', side=4, line=1, las=2, cex = 1.2)
  }
  if(i == 3){
    mtext('log-likelihood difference', outer=TRUE, side=2, line=3, cex = 1.2)
  }
}
dev.off()

start_time <- Sys.time()
library("trtf")
library("tram")
library("party")
library("parallel")
library("survival")
set.seed(9235)

source("tr_competitors.R")
### load data from tr_simulated.rda generated by running tr_data_simulation.R
load('tr_simulated_data.rda')

### Number of repetitions
NSim <- length(simdat[[1]])
ret <- args[rep(1:nrow(args), each = NSim),]
simdat <- do.call("c", simdat)
idx <- 1:nrow(ret)

### Log-likelihood estimates will be stored in ret
ret$tweib <- ret$tcoxph <- 0
ret$w_alpha_beta <- ret$w_theta_beta <- ret$w_theta_theta <- 0
ret$bs_alpha_beta <- ret$bs_theta_beta <- ret$bs_theta_theta <- 0
ret$loglik <- unlist(do.call("c", loglik))

### a filename to save results
res_filename <- 'tr_results_empeval.rda'

### number of cores for parallel computing
ncores <- 15

### number of randomly preselected variables for splitting
### for low-dimentional data it is set to the number of variables
### for high-dimensional data it is set to the square root of the number of variables
mtry <- lapply(idx, function(i) {
                         ifelse(ret[i, "p"] <= 5,
                         ncol(simdat[[i]]$train) - 2,
                         ceiling(sqrt(ncol(simdat[[i]]$train) - 2)))})

### folder to save Bs(theta, theta) model for each somulated dataset
tf_folder <- 'results_tf'
dir.create(tf_folder)
tf_dir <- paste(getwd(), tf_folder, sep='/')

### Bs(theta, theta)
bstt <- function(i){
    if (i %% NSim == 0){
         print(i)
     }

    tf <- bs_theta_theta(simdat[[i]]$train,
                         simdat[[i]]$test,
                         strata = simdat[[i]]$train$trt,
                         mtry = mtry[[i]])
    ## save the learned forest
    save(tf, file=paste0(tf_dir, '/tf', i, '.rda'))

    return(unclass(tf$loglik))
}
### apply in parallel to all simulated datasets
f0 <- mclapply(idx, bstt, mc.cores = ncores)
ret$bs_theta_theta <- unlist(f0)
print('Model Bs_theta_theta is done!')
save(ret, file = res_filename)

### Bs(alpha, beta)
bsab <- function(i){
    if (i %% NSim == 0){
         print(i)
     }
    ## load the corresponding Bs(theta,theta) forest
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## tf is the loaded forest
    unclass(bs_alpha_beta(tf, 
                          simdat[[i]]$train,
                          simdat[[i]]$test,
                          strata = simdat[[i]]$train$trt)$loglik)
}
### apply in parallel to all simulated datasets
f0 <- mclapply(idx, bsab, mc.cores = ncores)
ret$bs_alpha_beta <- unlist(f0)
print('Model Bs_alpha_beta is done!')
save(ret, file = res_filename)


### Bs(theta, beta)
bstb <- function(i){
    if (i %% NSim == 0){
         print(i)
     }

    ## load the corresponding Bs(theta,theta) forest
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## tf is the loaded forest
    unclass(bs_theta_beta(tf, 
                          simdat[[i]]$train,
                          simdat[[i]]$test,
                          strata = simdat[[i]]$train$trt)$loglik)
}
### apply in parallel to all simulated datasets
f0 <- mclapply(idx, bstb, mc.cores = ncores)
ret$bs_theta_beta <- unlist(f0)
print('Model Bs_theta_beta is done!')
save(ret, file = res_filename)


### W(alpha, beta)
wab <- function(i){
    if (i %% NSim == 0){
        print(i)
    }
    ## load the corresponding Bs(theta,theta) forest
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## tf is the loaded forest
    unclass(w_alpha_beta(tf, 
                         simdat[[i]]$train,
                         simdat[[i]]$test,
                         strata = simdat[[i]]$train$trt)$loglik)
}
### apply in parallel to all simulated datasets
f0 <- mclapply(idx, wab, mc.cores = ncores)
ret$w_alpha_beta <- unlist(f0)
print('Model W_alpha_beta is done!')
save(ret, file = res_filename)

### W(theta, beta)
wtb <- function(i){
    if (i %% NSim == 0){
         print(i)
     }
    ## load the corresponding Bs(theta,theta) forest
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## tf is the loaded forest
    unclass(w_theta_beta(tf, 
                         simdat[[i]]$train,
                         simdat[[i]]$test,
                         strata = simdat[[i]]$train$trt)$loglik)
}
### apply in parallel to all simulated datasets
f0 <- mclapply(idx, wtb, mc.cores = ncores)
ret$w_theta_beta <- unlist(f0)
print('Model W_theta_beta is done!')
save(ret, file = res_filename)

### W(theta, theta)
wtt <- function(i){
    if (i %% NSim == 0){
         print(i)
     }
    ## load the corresponding Bs(theta,theta) forest
    load(paste0(tf_dir, '/tf', i, '.rda'))
    ## tf is the loaded forest
    unclass(w_theta_theta(tf, 
                          simdat[[i]]$train,
                          simdat[[i]]$test,
                          strata = simdat[[i]]$train$trt)$loglik)
}
### apply in parallel to all simulated datasets
f0 <- mclapply(idx, wtt, mc.cores = ncores)
ret$w_theta_theta <- unlist(f0)
print('Model W_theta_theta is done!')
save(ret, file = res_filename)

end_time <- Sys.time()
print(end_time - start_time)

sessionInfo()
